package org.artifactory.ph;

import lombok.extern.slf4j.Slf4j;
import org.artifactory.common.repo.PackageType;
import org.artifactory.common.repo.RepoInfo;
import com.jfrog.ph.acl.exception.PackageException;
import com.jfrog.ph.acl.exception.PackageRuntimeException;
import com.jfrog.ph.acl.service.PackageSpecificationAdapter;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.repo.cache.expirable.IndexChecker;

import javax.annotation.Nonnull;

import static org.artifactory.ph.PackageRepoTypeMapper.packageType;

@Slf4j
public class PackageSpecificationImpl extends IndexChecker implements PackageSpecification {
    private final PackageSpecificationAdapter packageSpecificationAdapter;

    public PackageSpecificationImpl(PackageSpecificationAdapter packageSpecificationAdapter) {
        this.packageSpecificationAdapter = packageSpecificationAdapter;
    }

    @Override
    public boolean isExpirable(RepoType repoType, String repoKey, @Nonnull String path) {
        try {
            PackageType packageType = packageType(repoType);
            if (packageType == null || !packageSpecificationAdapter.isSupportedPackage(packageType)) {
                return false;
            }
            return packageSpecificationAdapter.isExpirable(packageType, path);
        } catch (PackageException e) {
            log.error("Failed to check if artifact is expired, reason:{}", e.getInternalReason());
            throw new PackageRuntimeException(e.getStatus(), e.getPublicReason(), e.getInternalReason(), e);
        }
    }

    @Override
    public boolean isLocalGenerated(@Nonnull RepoType repoType, @Nonnull String repoKey, @Nonnull String path) {
        try {
            PackageType packageType = packageType(repoType);
            if (packageType == null || !packageSpecificationAdapter.isSupportedPackage(packageType)) {
                return false;
            }
            return packageSpecificationAdapter.isLocalGenerated(packageType, path);
        } catch (PackageException e) {
            log.error("Failed to check if artifact is locally generated, reason:{}", e.getInternalReason());
            throw new PackageRuntimeException(e.getStatus(), e.getPublicReason(), e.getInternalReason(), e);
        }
    }

    @Override
    public boolean shouldExcludePropertyFromMirroring(@Nonnull RepoType repoType, @Nonnull String propertyKey) {
        try {
            PackageType packageType = packageType(repoType);
            if (packageType == null || !packageSpecificationAdapter.isSupportedPackage(packageType)) {
                return false;
            }
            return packageSpecificationAdapter.isLocallyGeneratedProperty(packageType, propertyKey);
        } catch (PackageException e) {
            log.error("Failed to check if property should be excluded from mirroring, reason:{}",
                    e.getInternalReason());
            throw new PackageRuntimeException(e.getStatus(), e.getPublicReason(), e.getInternalReason(), e);
        }
    }

    @Override
    public boolean isPackageArtifact(@Nonnull String repoKey, @Nonnull String path) {
        try {
            if (!packageSpecificationAdapter.isSupportedPackage(repoKey)) {
                return false;
            }
            return packageSpecificationAdapter.isPackageArtifact(repoKey, path);
        } catch (PackageException e) {
            log.error("Failed to check if artifact is package artifact, reason:{}", e.getInternalReason());
            throw new PackageRuntimeException(e.getStatus(), e.getPublicReason(), e.getInternalReason(), e);
        }
    }

    @Override
    public boolean hasReindex(org.artifactory.common.repo.RepoType repoType, RepoType packageType) {
        return packageSpecificationAdapter.hasReindex(getRepoInfo(repoType, packageType));
    }

    @Override
    public boolean isViewable(@Nonnull String repoKey, @Nonnull String path, RepoType repoPkgType) {
        // special cases, remove once package is moved to the PH
        if (repoPkgType == RepoType.Cargo && path.startsWith("index/") && !path.endsWith("/")) {
            return true;
        }
        if (repoPkgType == RepoType.Debian && path.endsWith("/Packages")) {
            return true;
        }
        return packageSpecificationAdapter.isViewable(repoKey, path);
    }

    @Override
    public boolean isPackageArtifact(@Nonnull String repoKey, @Nonnull String path, @Nonnull RepoType repoType) {
        try {
            if (packageSpecificationAdapter.isSupportedPackage(repoKey)) {
                return isPackageArtifact(repoKey, path);
            }
            if (RepoType.NuGet.equals(repoType) && path.endsWith(".nupkg")) {
                return true;
            }
            if (RepoType.Npm.equals(repoType) && path.endsWith(".tgz")) {
                return true;
            }
        } catch (PackageException e) {
            log.error("Failed to check if artifact is package artifact, reason:{}", e.getInternalReason());
            throw new PackageRuntimeException(e.getStatus(), e.getPublicReason(), e.getInternalReason(), e);
        }
        return false;
    }

    private RepoInfo getRepoInfo(org.artifactory.common.repo.RepoType repoType, RepoType packageType) {
        PackageType packageT = PackageRepoTypeMapper.packageType(packageType);
        return switch (repoType) {
            case REMOTE -> new RepoInfo(org.artifactory.common.repo.RepoType.REMOTE, packageT);
            case VIRTUAL -> new RepoInfo(org.artifactory.common.repo.RepoType.VIRTUAL, packageT);
            case LOCAL -> new RepoInfo(org.artifactory.common.repo.RepoType.LOCAL, packageT);
        };
    }

    @Override
    public RepoType[] getTypes() {
        // will always be evaluated
        return RepoType.values();
    }
}
