package org.artifactory.probes.service;

import lombok.extern.slf4j.Slf4j;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.probes.jobs.ReadinessJob;
import org.artifactory.probes.model.ReadinessStatus;
import org.artifactory.schedule.Task;
import org.artifactory.schedule.TaskService;
import org.artifactory.schedule.TaskUtils;
import org.artifactory.security.access.AccessService;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.state.ArtifactoryServerState;
import org.artifactory.storage.StorageService;
import org.artifactory.storage.db.InternalDbService;
import org.artifactory.storage.db.servers.service.ArtifactoryServersCommonService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.ws.rs.ServiceUnavailableException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;

/**
 * @author Andrey Tashlitsky
 */
@Slf4j
@Service
public class ProbesServiceImpl implements ProbesService, ContextCreationListener {

    private static final String HEALTH_CHECK_INTERVAL_KEY_NAME = "artifactory.healthCheck.intervalSecs";
    private static final int DEFAULT_HEALTH_CHECK_INTERVAL = 15;
    private static final int MIN_HEALTH_CHECK_INTERVAL = 3;
    private static final String LIVENESS_FAIL_READINESS_ENABLED_KEY_NAME =
            "artifactory.probes.liveness.failOnLongFailingReadiness.enabled";
    private static final boolean DEFAULT_LIVENESS_FAIL_READINESS_ENABLED = true;
    private static final String LIVENESS_FAIL_READINESS_DURATION_KEY_NAME =
            "artifactory.probes.liveness.failOnLongFailingReadiness.failureDurationSecs";
    private static final int DEFAULT_LIVENESS_FAIL_READINESS_DURATION = 60;
    private static final int MIN_LIVENESS_FAIL_READINESS_DURATION = 0;

    private final ArtifactoryServersCommonService artifactoryServers;
    private final StorageService storageService;
    private final InternalDbService internalDbService;
    private final AccessService accessService;
    private final TaskService taskService;
    private final ReentrantLock readinessLock = new ReentrantLock();

    private ReadinessStatus readinessStatus;
    private long lastSuccessfulReadiness = 0;

    @Autowired
    public ProbesServiceImpl(ArtifactoryServersCommonService artifactoryServers, StorageService storageService,
            InternalDbService internalDbService, AccessService accessService, TaskService taskService) {
        this.artifactoryServers = artifactoryServers;
        this.storageService = storageService;
        this.internalDbService = internalDbService;
        this.accessService = accessService;
        this.taskService = taskService;
    }

    @Override
    public void triggerReadinessCheck() {
        if (readinessLock.tryLock()) {
            try {
                log.trace("Readiness check has started");
                readinessStatus = ReadinessStatus.builder()
                        .running(checkIsRunning())
                        .offline(checkIsOffline())
                        .canWrite(checkCanWrite())
                        .storageServiceReady(checkStorageServiceReady())
                        .internalDbServiceHealthy(checkInternalDbServiceHealthy())
                        .accessServiceReady(checkAccessServiceReady())
                        .build();
                if (readinessStatus.isReady()) {
                    lastSuccessfulReadiness = System.currentTimeMillis();
                }
                log.trace("Readiness check has finished");
            } finally {
                readinessLock.unlock();
            }
        }
    }

    @Override
    public void checkReadinessStatus(boolean noCache) {
        log.trace("Checking readiness status");
        if (noCache) {
            triggerReadinessCheck();
        }
        if (readinessStatus != null && readinessStatus.isReady()) {
            log.trace("Artifactory is ready");
            return;
        }
        throw new ServiceUnavailableException("One or more of the dependent services/resources have failed");
    }

    @Override
    public void checkLivenessStatus() {
        log.trace("Checking liveness status");
        if (!isLivenessDependentOnReadiness()) {
            return;
        }
        if (readinessStatus == null) {
            String message = "Liveness probe cannot test readiness since readiness was not triggered";
            log.warn(message);
            throw new ServiceUnavailableException(message);
        }
        if (!readinessStatus.isReady()) {
            int failDuration = getReadinessFailureLimit();
            long secondsSinceLastSuccessfulReadiness =
                    TimeUnit.MILLISECONDS.toSeconds(System.currentTimeMillis() - lastSuccessfulReadiness);
            if (secondsSinceLastSuccessfulReadiness > failDuration) {
                throw new ServiceUnavailableException(
                        "Liveness failed: readiness state failed for more than " + failDuration + " seconds");
            }
        }
        log.trace("Artifactory is alive");
    }

    @Override
    public void onContextCreated() {
        int intervalSecs = getReadinessHealthCheckInterval();
        long intervalMillies = TimeUnit.SECONDS.toMillis(intervalSecs);
        Task t = TaskUtils.createRepeatingTask(ReadinessJob.class, intervalMillies, 0L);
        log.debug("Scheduling ReadinessJob to run every {} seconds", intervalSecs);
        taskService.startTask(t, false);
        log.debug("Scheduled ReadinessJob to run every {} seconds", intervalSecs);
    }

    /**
     * @return Health check interval (seconds).
     */
    private int getReadinessHealthCheckInterval() {
        return Math.max(ArtifactoryHome.get().getSysConfig()
                .getInt(HEALTH_CHECK_INTERVAL_KEY_NAME, DEFAULT_HEALTH_CHECK_INTERVAL), MIN_HEALTH_CHECK_INTERVAL);
    }

    /**
     * @return True if liveness check is dependent on readiness check. False otherwise.
     */
    private boolean isLivenessDependentOnReadiness() {
        return ArtifactoryHome.get().getSysConfig()
                .getBoolean(LIVENESS_FAIL_READINESS_ENABLED_KEY_NAME, DEFAULT_LIVENESS_FAIL_READINESS_ENABLED);
    }

    /**
     * @return Readiness failure limit (seconds).
     */
    private int getReadinessFailureLimit() {
        return Math.max(ArtifactoryHome.get().getSysConfig()
                        .getInt(LIVENESS_FAIL_READINESS_DURATION_KEY_NAME, DEFAULT_LIVENESS_FAIL_READINESS_DURATION),
                MIN_LIVENESS_FAIL_READINESS_DURATION);
    }

    /**
     * Check whether Artifactory is in a RUNNING state. Only in this state, Artifactory can be considered "ready". Other
     * states such as CONVERTING, STARTING, STOPPING, etc. are considered as "not ready".
     *
     * @return True if Artifactory is in a RUNNING state. False otherwise.
     */
    private boolean checkIsRunning() {
        try {
            boolean running = artifactoryServers.getCurrentMember().getServerState() == ArtifactoryServerState.RUNNING;
            if (running) {
                log.debug("Readiness running test success");
            } else {
                log.warn("Readiness running test failed");
            }
            return running;
        } catch (Exception e) {
            log.warn("Readiness running test failed with the following error: {}", e.getMessage());
            return false;
        }
    }

    /**
     * Checks whether Artifactory application context is in an "offline" state.
     *
     * @return True if Artifactory is in an "offline" state. False otherwise.
     */
    private boolean checkIsOffline() {
        try {
            boolean offline = ContextHelper.get().isOffline();
            if (offline) {
                log.warn("Readiness online test failed");
            } else {
                log.debug("Readiness online test success");
            }
            return offline;
        } catch (Exception e) {
            log.warn("Readiness online test failed with the following error: {}", e.getMessage());
            return true;
        }
    }

    /**
     * Checks whether Artifactory can write to data and log directories.
     *
     * @return True if Artifactory can write. False otherwise.
     */
    private boolean checkCanWrite() {
        try {
            boolean canWrite =
                    ArtifactoryHome.get().getDataDir().canWrite() && ArtifactoryHome.get().getLogDir().canWrite();
            if (canWrite) {
                log.debug("Readiness can write test success");
            } else {
                log.warn("Readiness can write test failed");
            }
            return canWrite;
        } catch (Exception e) {
            log.warn("Readiness can write test failed with the following error: {}", e.getMessage());
            return false;
        }
    }

    /**
     * Checks whether Artifactory's storage service is ready.
     *
     * @return True if Artifactory's storage service is ready. False otherwise.
     */
    private boolean checkStorageServiceReady() {
        try {
            storageService.ping(true);
            log.debug("Readiness storage service ping success");
            return true;
        } catch (Exception e) {
            log.warn("Readiness storage service ping test failed with the following error: {}", e.getMessage());
            return false;
        }
    }

    /**
     * Checks whether Artifactory's DB service is ready.
     *
     * @return True if Artifactory's DB service is ready. False otherwise.
     */
    private boolean checkInternalDbServiceHealthy() {
        try {
            internalDbService.health();
            log.debug("Readiness internal DB service health success");
            return true;
        } catch (Exception e) {
            log.warn("Readiness internal DB service health test failed with the following error: {}", e.getMessage());
            return false;
        }
    }

    /**
     * Checks whether Artifactory's Access service is ready.
     *
     * @return True if Artifactory's Access service is ready. False otherwise.
     */
    private boolean checkAccessServiceReady() {
        try {
            accessService.ping();
            log.debug("Readiness Access service ping success");
            return true;
        } catch (Exception e) {
            log.warn("Readiness Access service ping test failed with the following error: {}", e.getMessage());
            return false;
        }
    }
}
