package org.artifactory.repo.service.validation;

import com.google.common.collect.Lists;
import lombok.experimental.UtilityClass;
import org.apache.commons.lang.StringUtils;
import org.artifactory.api.config.RepositoryConfigService;
import org.artifactory.util.UiRequestUtils;

import java.util.List;
import java.util.Optional;

import static org.apache.commons.lang.StringUtils.isBlank;
import static org.artifactory.repo.system.SystemRepos.SystemRepoConfig.isSystemRepo;
import static org.jdom2.Verifier.isXMLNameCharacter;

@UtilityClass
public class RepositoryNameValidator {
    private static final int REPO_KEY_MAX_LENGTH = 64;
    private static final List<String> forbiddenNames = List.of(".", "..", "&");
    private static final List<Character> forbiddenChars = Lists
            .newArrayList('/', '\\', ':', '|', '?', '*', '"', '<', '>');
    private static final List<String> blacklistPrefixes = List.of("jfrog-system-reserved", "jfrog-artifactory-system");

    /**
     * @throws IllegalArgumentException if repo key is blank
     */
    public static void isKeyBlank(String repoKey) throws IllegalArgumentException {
        if (isBlank(repoKey)) {
            throw new IllegalArgumentException("Repository key cannot be empty");
        }
    }

    /**
     * @throws IllegalArgumentException if repo key exceeds the max length
     */
    public static void isMaxLengthExceed(String repoKey, String repoType) {
        //Local repos limited at 64, remote repos append '-cache' to cache repos which is 6 chars.
        int maxLength = "remote".equals(repoType) ? (REPO_KEY_MAX_LENGTH - 6) : REPO_KEY_MAX_LENGTH;
        if (StringUtils.length(repoKey) > maxLength) {
            throw new IllegalArgumentException("Repository key exceeds maximum length (" + maxLength + ") chars.");
        }
    }

    /**
     * @throws IllegalArgumentException if repo key is a reserved name
     */
    public static void isReservedName(String repoKey) {
        if (UiRequestUtils.isReservedName(repoKey) || isSystemRepo(repoKey)) {
            throw new IllegalArgumentException("Repository key '" + repoKey + "' is a reserved name");
        }
    }

    /**
     * @throws IllegalArgumentException if repo key contains a blacklisted prefix
     */
    public static void isBlacklistedPrefix(String repoKey) {
        Optional<String> prefix = blacklistPrefixes.stream().filter(repoKey::startsWith).findFirst();
        if (prefix.isPresent()) {
            throw new IllegalArgumentException(
                    "Repository key '" + repoKey + "' contains blacklisted prefix '" + prefix.get() + "'");
        }
    }

    /**
     * @throws IllegalArgumentException if repo key is a forbidden name
     */
    public static void isForbiddenName(String repoKey) {
        if (forbiddenNames.stream().anyMatch(repoKey::equals)) {
            throw new IllegalArgumentException("Invalid Repository key");
        }
    }

    /**
     * @throws IllegalArgumentException if repo key contains a forbidden character
     */
    public static void isForbiddenCharacter(String repoKey) {
        for (int i = 0; i < repoKey.length(); ++i) {
            if (forbiddenChars.contains(repoKey.charAt(i))) {
                throw new IllegalArgumentException("Illegal Repository key : '/,\\,:,|,?,<,>,*,\"' is not allowed");
            }
            if (!isXMLNameCharacter(repoKey.charAt(i))) {
                throw new IllegalArgumentException("Illegal Repository key : '" + repoKey.charAt(i) + "'");
            }
        }
    }

    /**
     * @throws IllegalArgumentException if repo key already exists in repository config service (meaning the key is
     *                                  used by other repository)
     */
    public static void isKeyAvailableInRepositoryConfigService(String repoKey,
            RepositoryConfigService repositoryConfigService) {
        if (repositoryConfigService.existsRepositoryIgnoreCase(repoKey)) {
            throw new IllegalArgumentException("Case insensitive repository key already exists");
        }
    }
}
