/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.config;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.SerializationUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.CoreAddonsImpl;
import org.artifactory.addon.HaAddon;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.api.config.RepositoryConfigService;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.property.PredefinedValue;
import org.artifactory.descriptor.property.Property;
import org.artifactory.descriptor.property.PropertySet;
import org.artifactory.descriptor.property.SystemPropertyDescriptor;
import org.artifactory.descriptor.replication.LocalReplicationDescriptor;
import org.artifactory.descriptor.replication.RemoteReplicationDescriptor;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.ProxyDescriptor;
import org.artifactory.descriptor.repo.RepoLayout;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.descriptor.security.SecurityDescriptor;
import org.artifactory.exception.DocumentBuildException;
import org.artifactory.opentracing.ArtifactoryTracer;
import org.artifactory.platform.PlatformConfigServiceImpl;
import org.artifactory.projects.ProjectAdminValidation;
import org.artifactory.repo.confighandler.manager.ConfigChangeEventHandlerManager;
import org.artifactory.schedule.CachedThreadPoolTaskExecutor;
import org.artifactory.security.access.AccessService;
import org.artifactory.service.ConfigValidationService;
import org.artifactory.spring.ArtifactoryApplicationContext;
import org.artifactory.storage.db.fs.service.ConfigsServiceImpl;
import org.artifactory.storage.db.security.service.VersioningCacheImpl;
import org.artifactory.storage.fs.service.ConfigsService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.test.ArtifactoryHomeStub;
import org.artifactory.version.ArtifactoryServiceVersion;
import org.artifactory.version.ArtifactoryServiceVersionProvider;
import org.artifactory.version.CompoundVersionDetails;
import org.easymock.EasyMock;
import org.easymock.IAnswer;
import org.jetbrains.annotations.NotNull;
import org.jfrog.access.client.auth.ldap.AuthSettingsClient;
import org.jfrog.common.ResourceUtils;
import org.jfrog.common.opentracing.OpenTracerBuilder;
import org.jfrog.common.platform.subscription.SubscriptionType;
import org.springframework.test.util.ReflectionTestUtils;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static org.easymock.EasyMock.anyLong;
import static org.easymock.EasyMock.anyObject;
import static org.easymock.EasyMock.anyString;
import static org.easymock.EasyMock.createMock;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.expectLastCall;
import static org.easymock.EasyMock.getCurrentArguments;
import static org.easymock.EasyMock.replay;
import static org.mockito.Mockito.any;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoInteractions;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNotEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertTrue;
import static org.testng.Assert.expectThrows;
import static org.testng.AssertJUnit.assertNull;

/**
 * Unit tests for the CentralConfigServiceImpl.
 *
 * @author Yossi Shaul
 */
public class CentralConfigServiceImplTest extends ArtifactoryHomeBoundTest {

    private static final CompoundVersionDetails earlyVersion = new CompoundVersionDetails(
            ArtifactoryServiceVersionProvider.v500.get(),
            "EARLY_TEST", 0
    );
    private static final CompoundVersionDetails recentVersion = new CompoundVersionDetails(
            ArtifactoryServiceVersionProvider.v522m001.get(),
            "RECENT_TEST", 0
    );
    private CentralConfigServiceImpl centralConfigService;
    private Map<String, String> configs;
    private ConfigsService configsService;
    private ArtifactoryHomeStub artifactoryHome;
    private AccessService accessService;
    private AuthSettingsClient authSettingsClient;

    private RepositoryConfigService repositoryConfigService;

    @DataProvider
    public static Object[][] orderProvider() {
        return new Object[][] {
                {true},
                {false}
        };
    }

    @BeforeMethod
    public void setup() throws IOException {
        configs = new HashMap<>();
        InputStream is = getClass().getResourceAsStream("/META-INF/default/artifactory.config.xml");
        String defaultConfig = IOUtils.toString(is, Charset.defaultCharset());
        configs.put(ArtifactoryHome.ARTIFACTORY_CONFIG_FILE, defaultConfig);

        centralConfigService = new CentralConfigServiceImpl();
        AuthorizationService authService = EasyMock.createMock(AuthorizationService.class);
        AddonsManager addonsManager = EasyMock.createMock(AddonsManager.class);
        HaCommonAddon haCommonAddon = EasyMock.createMock(HaCommonAddon.class);
        ConfigValidationService configValidationService = mock(ConfigValidationService.class);
        ConfigChangeEventHandlerManager configChangeEventHandlerManager = mock(ConfigChangeEventHandlerManager.class);
        ProjectAdminValidation projectAdminValidation = new ProjectAdminValidation(authService);
        HaAddon haAddon = EasyMock.createMock(HaAddon.class);
        ConfigurationChangesInterceptors interceptors = createMock(ConfigurationChangesInterceptors.class);

        configsService = EasyMock.createMock(ConfigsService.class);
        configsService.backupConfig(anyObject(), anyLong());
        expectLastCall().anyTimes();

        ArtifactoryApplicationContext context = mock(ArtifactoryApplicationContext.class);

        when(context.isReady()).thenReturn(true);
        when(context.beanForType(ArtifactoryTracer.class))
                .thenReturn(new ArtifactoryTracer(OpenTracerBuilder.getQuietJaegerTracer("Artifactory")));
        when(context.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        when(context.beanForType(ConfigChangeEventHandlerManager.class)).thenReturn(configChangeEventHandlerManager);
        expect(authService.isAdmin()).andReturn(true).anyTimes();
        expect(authService.isAnyProjectAdmin()).andReturn(true).anyTimes();
        expect(haCommonAddon.isHaEnabled()).andReturn(false).anyTimes();
        expect(addonsManager.addonByType(HaCommonAddon.class)).andReturn(haCommonAddon).anyTimes();
        expect(addonsManager.addonByType(HaAddon.class)).andReturn(haAddon).anyTimes();
        interceptors.onBeforeSave(anyObject());
        expectLastCall().anyTimes();
        interceptors.onAfterSave(anyObject(), anyObject());
        expectLastCall().anyTimes();
        IAnswer<Boolean> answer = initConfigsService();
        expect(configsService.updateConfigIfLastModificationMatch(anyObject(), anyObject(), anyLong(), anyLong()))
                .andAnswer(answer).anyTimes();
        accessService = mock(AccessService.class);
        authSettingsClient = mock(AuthSettingsClient.class);
        when(accessService.getAccessAuthSettingsClient()).thenReturn(authSettingsClient);
        doNothing().when(accessService).updateConfigurationModel(AccessConfigHolder.getInstance().drainConfig());
        ReflectionTestUtils.setField(centralConfigService, "authService", authService);
        ReflectionTestUtils.setField(centralConfigService, "addonsManager", addonsManager);
        ReflectionTestUtils.setField(centralConfigService, "interceptors", interceptors);
        ReflectionTestUtils.setField(centralConfigService, "configsService", configsService);
        ReflectionTestUtils.setField(centralConfigService, "accessService", accessService);
        ReflectionTestUtils
                .setField(centralConfigService, "centralConfigDiffValidation", projectAdminValidation);
        ReflectionTestUtils.setField(centralConfigService, "configValidationService", configValidationService);
        ReflectionTestUtils.setField(centralConfigService, "subConverters", Set.of());
        ReflectionTestUtils.setField(centralConfigService, "platformConfigService",
                mock(PlatformConfigServiceImpl.class));

        artifactoryHome = new ArtifactoryHomeStub(getClass().getName());
        when(context.getArtifactoryHome()).thenReturn(artifactoryHome);
        RepositoryService repositoryService = EasyMock.createMock(RepositoryService.class);
        when(context.getBean(RepositoryService.class)).thenReturn(repositoryService);
        repositoryConfigService = EasyMock.createMock(RepositoryConfigService.class);
        when(context.beanForType(RepositoryConfigService.class)).thenReturn(repositoryConfigService);
        centralConfigService.setRepositoryConfigService(repositoryConfigService);
        repositoryConfigService.bulkInsertRepositoryConfigs(anyObject(), EasyMock.anyBoolean(), EasyMock.anyBoolean(),
                EasyMock.anyBoolean());
        repositoryConfigService.importRepoConfigs(anyObject(), EasyMock.anyBoolean(), EasyMock.anyBoolean());
        expectLastCall().anyTimes();
        repositoryConfigService.reloadRepoConfigurations();
        expectLastCall().anyTimes();

        expect(repositoryConfigService.countAll()).andReturn(2).anyTimes();
        expect(repositoryConfigService.deleteAll(EasyMock.anyBoolean(), EasyMock.anyBoolean())).andReturn(2).anyTimes();
        expect(repositoryConfigService.findRepositoriesByType(anyObject())).andReturn(List.of()).anyTimes();
        expect(repositoryConfigService.getAllRepositoryConfigs()).andReturn(List.of()).anyTimes();
        expect(repositoryConfigService.exists("local")).andReturn(true).anyTimes();
        expect(repositoryConfigService.exists("remote-repo")).andReturn(true).anyTimes();
        expect(repositoryConfigService.exists("noRemote")).andReturn(false).anyTimes();
        expect(repositoryConfigService.exists("missingRepo")).andReturn(false).anyTimes();

        replay(authService, addonsManager, haCommonAddon, interceptors, repositoryConfigService, configsService);
        ArtifactoryHome.bind(artifactoryHome);
        ArtifactoryContextThreadBinder.bind(context);

        CachedThreadPoolTaskExecutor cachedThreadPoolTaskExecutor = new CachedThreadPoolTaskExecutor();
        ReflectionTestUtils
                .setField(centralConfigService, "cachedThreadPoolTaskExecutor", cachedThreadPoolTaskExecutor);

        centralConfigService.init();

        verifyNoInteractions(configChangeEventHandlerManager);
    }

    @AfterMethod
    public void cleanup() {
        ArtifactoryContextThreadBinder.unbind();
    }

    private IAnswer<Boolean> initConfigsService() {
        expect(configsService.getConfig(anyString())).andAnswer(() -> configs.get(getCurrentArguments()[0].toString()))
                .anyTimes();
        expect(configsService.hasConfig(anyObject()))
                .andAnswer(() -> configs.containsKey(getCurrentArguments()[0].toString())).anyTimes();
        IAnswer<Boolean> answer = () -> {
            Object[] args = getCurrentArguments();
            configs.put((String) args[0], (String) args[1]);
            return true;
        };
        expect(configsService.addOrUpdateConfig(anyString(), anyObject(), anyLong())).andAnswer(answer).anyTimes();
        configsService.addConfig(anyObject(), anyObject(), anyLong());
        expectLastCall().andAnswer(answer).anyTimes();
        configsService.updateConfig(anyObject(), anyObject(), anyLong());
        expectLastCall().andAnswer(answer).anyTimes();
        return answer;
    }

    @Test
    public void keyStorePasswordMovesToDescriptor() throws Exception {
        String PASSWORD_CONFIG_KEY = "keystore:password";
        String PASSWORD = "123456";

        // Create mocks
        ConfigsServiceImpl configsService = mock(ConfigsServiceImpl.class);
        ReflectionTestUtils.setField(centralConfigService, "configsService", configsService);
        when(configsService.getConfig(PASSWORD_CONFIG_KEY)).thenReturn(PASSWORD);
        when(configsService.hasConfig(eq(ArtifactoryHome.ARTIFACTORY_CONFIG_FILE))).thenReturn(false);

        AddonsManager addonsManager = mock(AddonsManager.class);
        ReflectionTestUtils.setField(centralConfigService, "addonsManager", addonsManager);
        when(addonsManager.addonByType(any())).thenReturn(new CoreAddonsImpl());

        ConfigurationChangesInterceptors interceptors = mock(ConfigurationChangesInterceptors.class);
        ReflectionTestUtils.setField(centralConfigService, "interceptors", interceptors);

        // Create descriptor
        MutableCentralConfigDescriptor configDescriptor = getConfigDescriptor();
        configDescriptor.setSecurity(new SecurityDescriptor());
        initCache(centralConfigService, configDescriptor);

        centralConfigService.moveKeyStorePasswordToConfig(earlyVersion, recentVersion, configDescriptor);

        assertTrue(PASSWORD.equals(configDescriptor.getSecurity().getSigningKeysSettings().getKeyStorePassword()));
    }

    @Test
    public void testMergeOfOneChange() {
        initConfig();

        MutableCentralConfigDescriptor newDescriptor = centralConfigService.getMutableDescriptor();
        String newBaseUrl = "https://newUrlbase:8080";
        newDescriptor.setUrlBase(newBaseUrl);
        MutableCentralConfigDescriptor clone = SerializationUtils.clone(newDescriptor);
        centralConfigService.saveEditedDescriptorAndReload(newDescriptor);
        assertNotEquals(centralConfigService.getDescriptor().getRevision(), clone.getRevision());
        assertEquals(centralConfigService.getDescriptor().getUrlBase(), newBaseUrl);
    }

    @Test
    public void testValidationCalled() {
        ConfigValidationService configValidationService = mock(ConfigValidationService.class);
        ReflectionTestUtils
                .setField(centralConfigService, "configValidationService", configValidationService);
        MutableCentralConfigDescriptor newDescriptor = centralConfigService.getMutableDescriptor();
        RepoLayout repoLayout = new RepoLayout();
        repoLayout.setName("new");
        centralConfigService.saveEditedDescriptorAndReload(newDescriptor, repoLayout);
        verify(configValidationService, times(1)).validate(repoLayout);
    }

    @Test
    public void testSaveConfigAtFirstTime() throws InterruptedException {
        EasyMock.reset(configsService);
        initConfigsService();
        EasyMock.replay(configsService);

        File latestFile = artifactoryHome.getArtifactoryConfigLatestFile();
        latestFile.delete();
        assertFalse(latestFile.exists());

        MutableCentralConfigDescriptor config = centralConfigService.getMutableDescriptor();
        centralConfigService.saveEditedDescriptorAndReload(config);
        assertNotEquals(centralConfigService.getDescriptor().getRevision(), 0);

        Thread.sleep(200); // File is written in background
        assertTrue(latestFile.exists());
        EasyMock.verify(configsService);
    }

    @Test(expectedExceptions = RuntimeException.class, expectedExceptionsMessageRegExp = "Failed to read object from " +
            "stream")
    public void testSaveInvalidXmlWithImport() throws IOException {
        initConfig();

        String invalidConfig = IOUtils
                .toString(getClass().getResourceAsStream("artifactory.config.invalid.xml"));
        FileUtils.write(ArtifactoryHome.get().getArtifactoryConfigImportFile(), invalidConfig);

        centralConfigService.init();
    }

    @Test
    public void testSaveValidXmlWithImport() throws IOException {
        initConfig();

        String validConfig = IOUtils
                .toString(getClass().getResourceAsStream("artifactory.config.valid.xml"));
        FileUtils.write(ArtifactoryHome.get().getArtifactoryConfigImportFile(), validConfig);

        centralConfigService.init();
    }

    @Test(dataProvider = "orderProvider")
    public void testMergeOfTwoChangesWithoutConflicts(boolean inOrder) {
        initConfig();

        MutableCentralConfigDescriptor newDescriptor = centralConfigService.getMutableDescriptor();
        String newBaseUrl = "https://newUrlbase:8080";
        newDescriptor.setUrlBase(newBaseUrl);

        MutableCentralConfigDescriptor newDescriptor2 = centralConfigService.getMutableDescriptor();
        String newServerName = "https://newUrlbase:8080";
        newDescriptor2.setServerName(newServerName);

        saveWithOrder(inOrder, newDescriptor, newDescriptor2);

        assertTrue(centralConfigService.getDescriptor().getRevision() > newDescriptor.getRevision());
        assertTrue(centralConfigService.getDescriptor().getRevision() > newDescriptor2.getRevision());
        assertEquals(centralConfigService.getDescriptor().getUrlBase(), newBaseUrl);
        assertEquals(centralConfigService.getDescriptor().getServerName(), newServerName);
    }

    @Test(dataProvider = "orderProvider")
    public void testMergeOfTwoChangesWithConflict(boolean inOrder) {
        initConfig();

        MutableCentralConfigDescriptor newDescriptor = centralConfigService.getMutableDescriptor();
        newDescriptor.setUrlBase("https://newUrlbase:8080");

        MutableCentralConfigDescriptor newDescriptor2 = centralConfigService.getMutableDescriptor();
        newDescriptor2.setUrlBase("https://newUrlbase222:8080");

        String toValidate;
        if (inOrder) {
            centralConfigService.saveEditedDescriptorAndReload(newDescriptor);
            centralConfigService.saveEditedDescriptorAndReload(newDescriptor2);
            toValidate = newDescriptor2.getUrlBase();
        } else {
            centralConfigService.saveEditedDescriptorAndReload(newDescriptor2);
            centralConfigService.saveEditedDescriptorAndReload(newDescriptor);
            toValidate = newDescriptor.getUrlBase();
        }

        assertTrue(centralConfigService.getDescriptor().getRevision() > newDescriptor.getRevision());
        assertTrue(centralConfigService.getDescriptor().getRevision() > newDescriptor2.getRevision());
        assertEquals(centralConfigService.getDescriptor().getUrlBase(), toValidate);
    }

    @Test(dataProvider = "orderProvider")
    public void testMergeOfMapObjects(boolean inOrder) {
        initConfig();

        MutableCentralConfigDescriptor newDescriptor = centralConfigService.getMutableDescriptor();
        LocalReplicationDescriptor newReplication = new LocalReplicationDescriptor();
        newReplication.setReplicationKey("hello-world");
        newReplication.setRepoKey("repo-key");
        newDescriptor.addLocalReplication(newReplication);

        MutableCentralConfigDescriptor newDescriptor2 = centralConfigService.getMutableDescriptor();
        LocalReplicationDescriptor newReplication2 = new LocalReplicationDescriptor();
        newReplication2.setReplicationKey("hello-world2");
        newReplication2.setRepoKey("repo-key2");
        newDescriptor2.addLocalReplication(newReplication2);

        saveWithOrder(inOrder, newDescriptor, newDescriptor2);

        assertTrue(centralConfigService.getDescriptor().getRevision() > newDescriptor.getRevision());
        assertTrue(centralConfigService.getDescriptor().getRevision() > newDescriptor2.getRevision());
        LocalReplicationDescriptor localReplication =
                centralConfigService.getDescriptor().getLocalReplication("repo-key" + (inOrder ? "2" : ""));
        assertNotNull(localReplication);
        assertEquals(localReplication.getReplicationKey(), "hello-world" + (inOrder ? "2" : ""));
    }

    @Test(dataProvider = "orderProvider")
    public void testMergeListOfObjects(boolean inOrder) {
        initConfig();

        MutableCentralConfigDescriptor newDescriptor = centralConfigService.getMutableDescriptor();
        PropertySet propertySet = new PropertySet();
        propertySet.setName("name");
        newDescriptor.addPropertySet(propertySet);

        MutableCentralConfigDescriptor newDescriptor2 = centralConfigService.getMutableDescriptor();
        PropertySet propertySet2 = new PropertySet();
        propertySet2.setName("name2");
        newDescriptor2.addPropertySet(propertySet2);

        saveWithOrder(inOrder, newDescriptor, newDescriptor2);

        assertTrue(centralConfigService.getDescriptor().getRevision() > newDescriptor.getRevision());
        assertTrue(centralConfigService.getDescriptor().getRevision() > newDescriptor2.getRevision());
        assertEquals(centralConfigService.getDescriptor().getPropertySets().get(0).getName(),
                "name" + (inOrder ? "" : "2"));
        assertEquals(centralConfigService.getDescriptor().getPropertySets().get(1).getName(),
                "name" + (inOrder ? "2" : ""));
    }

    @Test(dataProvider = "orderProvider")
    public void testMergeListWithConflictOnKey(boolean inOrder) {
        initConfig();

        MutableCentralConfigDescriptor newDescriptor = centralConfigService.getMutableDescriptor();
        PropertySet propertySet = new PropertySet();
        propertySet.setName("name");
        propertySet.setVisible(false);
        newDescriptor.addPropertySet(propertySet);

        MutableCentralConfigDescriptor newDescriptor2 = centralConfigService.getMutableDescriptor();
        PropertySet propertySet2 = new PropertySet();
        propertySet2.setName("name");
        propertySet2.setVisible(true);
        newDescriptor2.addPropertySet(propertySet2);

        saveWithOrder(inOrder, newDescriptor, newDescriptor2);

        assertTrue(centralConfigService.getDescriptor().getRevision() > newDescriptor.getRevision());
        assertTrue(centralConfigService.getDescriptor().getRevision() > newDescriptor2.getRevision());
        assertEquals(centralConfigService.getDescriptor().getPropertySets().size(), 1);
        assertEquals(centralConfigService.getDescriptor().getPropertySets().get(0).isVisible(), inOrder);
    }

    @Test(dataProvider = "orderProvider")
    public void testMergeMapWithConflictOnKey(boolean inOrder) {
        initConfig();

        MutableCentralConfigDescriptor newDescriptor = centralConfigService.getMutableDescriptor();
        LocalReplicationDescriptor newReplication = new LocalReplicationDescriptor();
        newReplication.setReplicationKey("hello-world");
        newReplication.setRepoKey("repo-key");
        newDescriptor.addLocalReplication(newReplication);

        MutableCentralConfigDescriptor newDescriptor2 = centralConfigService.getMutableDescriptor();
        LocalReplicationDescriptor newReplication2 = new LocalReplicationDescriptor();
        newReplication2.setReplicationKey("hello-world");
        newReplication2.setRepoKey("repo-key2");
        newDescriptor2.addLocalReplication(newReplication2);

        saveWithOrder(inOrder, newDescriptor, newDescriptor2);

        assertTrue(centralConfigService.getDescriptor().getRevision() > newDescriptor.getRevision());
        assertTrue(centralConfigService.getDescriptor().getRevision() > newDescriptor2.getRevision());
        LocalReplicationDescriptor localReplication =
                centralConfigService.getDescriptor().getLocalReplication("repo-key" + (inOrder ? "2" : ""));
        assertNotNull(localReplication);
        assertEquals(localReplication.getReplicationKey(), "hello-world");
    }

    @Test
    public void testGetJfrogBaseUrlNoPath() {
        initConfig();

        MutableCentralConfigDescriptor newDescriptor = centralConfigService.getMutableDescriptor();
        String newBaseUrl = "https://newUrlbase:8080";
        newDescriptor.setUrlBase(newBaseUrl);
        centralConfigService.saveEditedDescriptorAndReload(newDescriptor);
        assertEquals(centralConfigService.getDescriptor().getJFrogUrlBase(8081), newBaseUrl);
    }

    @Test
    public void testGetJfrogBaseUrlWIthPortSwitch() {
        initConfig();

        MutableCentralConfigDescriptor newDescriptor = centralConfigService.getMutableDescriptor();
        String newBaseUrl = "https://newUrlbase:8080";
        newDescriptor.setUrlBase(newBaseUrl);
        centralConfigService.saveEditedDescriptorAndReload(newDescriptor);
        assertEquals(centralConfigService.getDescriptor().getJFrogUrlBase(8082), newBaseUrl);
    }

    @Test
    public void testGetJfrogBaseUrlWithPath() {
        initConfig();

        MutableCentralConfigDescriptor newDescriptor = centralConfigService.getMutableDescriptor();
        String jfrogBaseUrl = "https://newUrlbase:8080";
        String newBaseUrl = jfrogBaseUrl + "/artifactory";
        newDescriptor.setUrlBase(newBaseUrl);
        centralConfigService.saveEditedDescriptorAndReload(newDescriptor);
        assertEquals(centralConfigService.getDescriptor().getJFrogUrlBase(8081), jfrogBaseUrl);
        assertEquals(centralConfigService.getDescriptor().getUrlBase(), newBaseUrl);
    }

    @Test
    public void centralConfigProcessingWithDtdSuccessful() {
        final String configWithDtd = ResourceUtils.getResourceAsString("/config/config-with-doctype_v2.1.0.xml");

        centralConfigService.setConfigXml(configWithDtd);

        // no exception for non UI/API request.
        centralConfigService.setConfigXml(configWithDtd, false);

        // exception for UI/API request.
        final Throwable throwable = expectThrows(
                Throwable.class,
                () -> centralConfigService.setConfigXml(configWithDtd, true)
        );
        Assert.assertNotNull(throwable);
        Assert.assertTrue(throwable instanceof DocumentBuildException);
        Assert.assertEquals(throwable.getMessage(),
                "Error on line 2: DOCTYPE is disallowed when the feature " +
                        "\"http://apache.org/xml/features/disallow-doctype-decl\" set to true."
        );
    }

    @Test
    public void getStringBasedOnEnumTest() {
        assertNull(
                centralConfigService
                        .getReloadablePropertyByEnum(SubscriptionType.class, ConstantValues.aolSubscriptionType));

        artifactoryHome.getArtifactoryProperties().setProperty(ConstantValues.aolSubscriptionType.getPropertyName(),
                SubscriptionType.FREE_XRAY.getString());
        SubscriptionType stringBasedEnum =
                centralConfigService
                        .getReloadablePropertyByEnum(SubscriptionType.class, ConstantValues.aolSubscriptionType);
        assertEquals(stringBasedEnum, SubscriptionType.FREE_XRAY);

        centralConfigService.getDescriptor().getSystemProperties()
                .add(new SystemPropertyDescriptor(ConstantValues.aolSubscriptionType.getPropertyName(),
                        SubscriptionType.FREE.getString()));
        stringBasedEnum =
                centralConfigService
                        .getReloadablePropertyByEnum(SubscriptionType.class, ConstantValues.aolSubscriptionType);
        assertEquals(stringBasedEnum, SubscriptionType.FREE);
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = "No enum value.*")
    public void getStringBasedOnEnumWithExceptionFromConstantValuesTest() {
        centralConfigService.getDescriptor().getSystemProperties()
                .add(new SystemPropertyDescriptor(ConstantValues.aolSubscriptionType.getPropertyName(),
                        "wrong subscription"));
        centralConfigService.getReloadablePropertyByEnum(SubscriptionType.class, ConstantValues.aolSubscriptionType);
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = "No enum value.*")
    public void getStringBasedOnEnumWithExceptionFromConfigTest() {
        artifactoryHome.getArtifactoryProperties().setProperty(ConstantValues.aolSubscriptionType.getPropertyName(),
                "no subscription");
        centralConfigService.getReloadablePropertyByEnum(SubscriptionType.class, ConstantValues.aolSubscriptionType);
    }

    @Test
    public void testRemoveDuplicatedReplicationsNoNeedToRun() throws IOException {
        initConfig();
        CompoundVersionDetails source = createVersion();
        CompoundVersionDetails target = createVersion();
        MutableCentralConfigDescriptor mutableDescriptor = spy(centralConfigService.getMutableDescriptor());
        centralConfigService.removeDuplicatedReplications(source, target, mutableDescriptor);
        verify(mutableDescriptor, times(0)).getLocalReplicationsMap();
    }

    @Test
    public void testRemoveDuplicatedReplicationsRunsWithMarker() throws IOException {
        initConfig();
        CompoundVersionDetails source = createVersion();
        CompoundVersionDetails target = createVersion();
        MutableCentralConfigDescriptor mutableDescriptor = spy(centralConfigService.getMutableDescriptor());
        File forceDuplicateReplicationConvertFile = artifactoryHome.getForceDuplicateReplicationConvertFile();
        if (forceDuplicateReplicationConvertFile.exists()) {
            forceDuplicateReplicationConvertFile.delete();
        }
        assertTrue(forceDuplicateReplicationConvertFile.createNewFile());
        centralConfigService.removeDuplicatedReplications(source, target, mutableDescriptor);
        verify(mutableDescriptor, times(1)).getLocalReplicationsMap();
        assertFalse(forceDuplicateReplicationConvertFile.exists());
    }

    @Test
    public void testRemoveDuplicatedReplications() {
        initConfig();
        CompoundVersionDetails source =
                new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v7140m001.get(), "build", 9);
        CompoundVersionDetails target = createVersion();

        MutableCentralConfigDescriptor mutableDescriptor = centralConfigService.getMutableDescriptor();
        LocalRepoDescriptor localRepoDescriptor = createLocalRepo(mutableDescriptor);

        LocalReplicationDescriptor localReplicationDescriptor =
                createLocalReplication(mutableDescriptor, localRepoDescriptor.getKey(), "https://url.com",
                        " 0 23 5 * * ?");

        LocalReplicationDescriptor localReplicationDescriptor2 =
                createLocalReplication(mutableDescriptor, localRepoDescriptor.getKey(), "https://url2.com",
                        " 0 23 5 * * ?");

        LocalReplicationDescriptor localReplicationDescriptor3 =
                createLocalReplication(mutableDescriptor, localRepoDescriptor.getKey(), "https://url.com",
                        " 0 13 5 * * ?");
        LocalReplicationDescriptor localReplicationDescriptor4 =
                createLocalReplication(mutableDescriptor, "missingRepo", "https://url.com",
                        " 0 13 5 * * ?");

        HttpRepoDescriptor httpRepoDescriptor = createRemoteRepo(mutableDescriptor);

        createRemoteReplication(mutableDescriptor, httpRepoDescriptor.getKey(), " 0 23 5 * * ?", "replica");
        RemoteReplicationDescriptor remoteReplicationDescriptor2 =
                createRemoteReplication(mutableDescriptor, httpRepoDescriptor.getKey(), " 0 1 5 * * ?", "replica1");
        createRemoteReplication(mutableDescriptor, "noRemote", " 0 1 5 * * ?", "replica2");
        centralConfigService.removeDuplicatedReplications(source, target, mutableDescriptor);
        assertLocalReplications(mutableDescriptor, localRepoDescriptor, localReplicationDescriptor,
                localReplicationDescriptor2, localReplicationDescriptor3, localReplicationDescriptor4);

        assertRemoteReplications(mutableDescriptor, remoteReplicationDescriptor2);
    }

    @Test
    public void testIsInterested() throws IOException {
        CompoundVersionDetails target = createVersion();
        CompoundVersionDetails source = createVersion();

        /* this check is added to avoid running the test on release version. Only snapshot and milestone versions
         need this test to be run.
         */
        if (source.isCurrent()) {return;}

        assertTrue(centralConfigService.isInterested(source, target));
        source = new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "build", 9);

        assertFalse(centralConfigService.isInterested(source, target));
        File forceDuplicateReplicationConvertFile = artifactoryHome.getForceDuplicateReplicationConvertFile();
        if (!forceDuplicateReplicationConvertFile.exists()) {
            assertTrue(forceDuplicateReplicationConvertFile.createNewFile());
        }
        forceDuplicateReplicationConvertFile.deleteOnExit();
        assertTrue(centralConfigService.isInterested(source, target));
    }

    private void assertRemoteReplications(MutableCentralConfigDescriptor mutableDescriptor,
            RemoteReplicationDescriptor remoteReplicationDescriptor2) {
        List<RemoteReplicationDescriptor> remoteReplications = mutableDescriptor.getRemoteReplications();
        assertEquals(remoteReplications.size(), 1);
        assertEquals(remoteReplications.get(0), remoteReplicationDescriptor2);
    }

    private void assertLocalReplications(MutableCentralConfigDescriptor mutableDescriptor,
            LocalRepoDescriptor localRepoDescriptor,
            LocalReplicationDescriptor localReplicationDescriptor,
            LocalReplicationDescriptor localReplicationDescriptor2,
            LocalReplicationDescriptor localReplicationDescriptor3,
            LocalReplicationDescriptor localReplicationDescriptor4) {
        Map<String, Map<String, LocalReplicationDescriptor>> localReplicationsMap =
                mutableDescriptor.getLocalReplicationsMap();
        assertEquals(localReplicationsMap.size(), 1);
        assertNull(localReplicationsMap.get(localReplicationDescriptor4.getRepoKey()));
        Map<String, LocalReplicationDescriptor> replicationPerUrl =
                localReplicationsMap.get(localRepoDescriptor.getKey());
        assertEquals(replicationPerUrl.size(), 2);
        assertEquals(replicationPerUrl.get(localReplicationDescriptor2.getUrl()), localReplicationDescriptor2);
        assertEquals(replicationPerUrl.get(localReplicationDescriptor3.getUrl()), localReplicationDescriptor3);
        assertNotEquals(replicationPerUrl.get(localReplicationDescriptor.getUrl()), localReplicationDescriptor);
    }

    @NotNull
    private HttpRepoDescriptor createRemoteRepo(MutableCentralConfigDescriptor mutableDescriptor) {
        HttpRepoDescriptor httpRepoDescriptor = new HttpRepoDescriptor();
        httpRepoDescriptor.setUrl("https://remote.com");
        httpRepoDescriptor.setType(RepoType.Docker);
        httpRepoDescriptor.setKey("remote-repo");
        mutableDescriptor.addRemoteRepository(httpRepoDescriptor);
        return httpRepoDescriptor;
    }

    @NotNull
    private LocalRepoDescriptor createLocalRepo(MutableCentralConfigDescriptor mutableDescriptor) {
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setKey("local");
        localRepoDescriptor.setType(RepoType.Docker);
        mutableDescriptor.addLocalRepository(localRepoDescriptor);
        return localRepoDescriptor;
    }

    @NotNull
    private RemoteReplicationDescriptor createRemoteReplication(MutableCentralConfigDescriptor mutableDescriptor,
            String key, String s, String replica1) {
        RemoteReplicationDescriptor remoteReplicationDescriptor2 = new RemoteReplicationDescriptor();
        remoteReplicationDescriptor2.setRepoKey(key);
        remoteReplicationDescriptor2.setCronExp(s);
        remoteReplicationDescriptor2.setReplicationKey(replica1);
        mutableDescriptor.addRemoteReplication(remoteReplicationDescriptor2);
        return remoteReplicationDescriptor2;
    }

    @NotNull
    private LocalReplicationDescriptor createLocalReplication(MutableCentralConfigDescriptor mutableDescriptor,
            String key, String url, String cron) {
        LocalReplicationDescriptor localReplicationDescriptor = new LocalReplicationDescriptor();
        localReplicationDescriptor.setUrl(url);
        localReplicationDescriptor.setCronExp(cron);
        localReplicationDescriptor.setRepoKey(key);
        localReplicationDescriptor.setReplicationKey(key + url + cron);
        mutableDescriptor.addLocalReplication(localReplicationDescriptor);
        return localReplicationDescriptor;
    }

    @NotNull
    private CompoundVersionDetails createVersion() {
        return new CompoundVersionDetails(
                ArtifactoryServiceVersionProvider.values()[ArtifactoryServiceVersionProvider.values().length - 1].get(),
                "build", 9);
    }

    private void initConfig() {
        MutableCentralConfigDescriptor config = centralConfigService.getMutableDescriptor();
        centralConfigService.saveEditedDescriptorAndReload(config);
        assertNotEquals(centralConfigService.getDescriptor().getRevision(), 0);

        config = centralConfigService.getMutableDescriptor();
        centralConfigService.saveEditedDescriptorAndReload(config);
        assertNotEquals(centralConfigService.getDescriptor().getRevision() > config.getRevision(),
                config.getRevision() + " " + centralConfigService.getDescriptor().getRevision());
        assertNotEquals(centralConfigService.getDescriptor().getRevision(), 1);
    }

    private void saveWithOrder(boolean inOrder, MutableCentralConfigDescriptor newDescriptor,
            MutableCentralConfigDescriptor newDescriptor2) {
        if (inOrder) {
            centralConfigService.saveEditedDescriptorAndReload(newDescriptor);
            centralConfigService.saveEditedDescriptorAndReload(newDescriptor2);
        } else {
            centralConfigService.saveEditedDescriptorAndReload(newDescriptor2);
            centralConfigService.saveEditedDescriptorAndReload(newDescriptor);
        }
    }

    private void initCache(CentralConfigServiceImpl centralConfigService,
            MutableCentralConfigDescriptor configDescriptor) {
        VersioningCacheImpl<CentralConfigServiceImpl.CentralConfigDescriptorCache> cache =
                new VersioningCacheImpl<>(3000,
                        () -> new CentralConfigServiceImpl.CentralConfigDescriptorCache(configDescriptor,
                                configDescriptor.getServerName()));
        ReflectionTestUtils.setField(centralConfigService, "descriptorCache", cache);
    }

    private MutableCentralConfigDescriptor getConfigDescriptor() {
        MutableCentralConfigDescriptor cc = new CentralConfigDescriptorImpl();
        cc.setServerName("mymy");

        ProxyDescriptor proxy = new ProxyDescriptor();
        proxy.setHost("localhost");
        proxy.setKey("proxy");
        proxy.setPort(8987);
        cc.setProxies(Arrays.asList(proxy));

        // property sets
        PropertySet propSet = new PropertySet();
        propSet.setName("propSet1");
        Property prop = new Property();
        prop.setName("prop1");
        PredefinedValue value1 = new PredefinedValue();
        value1.setValue("value1");
        prop.addPredefinedValue(value1);
        PredefinedValue value2 = new PredefinedValue();
        value2.setValue("value2");
        prop.addPredefinedValue(value2);
        propSet.addProperty(prop);
        cc.addPropertySet(propSet);

        return cc;
    }
}
