package org.artifactory.hqc.service;

import com.fasterxml.jackson.databind.JsonNode;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.hqc.config.HeavyQueryCacheConfig;
import org.artifactory.hqc.enums.HeavyQueryCacheType;
import org.artifactory.hqc.exception.HeavyQueryCacheException;
import org.artifactory.schedule.TaskService;
import org.artifactory.storage.db.DbService;
import org.artifactory.storage.db.hqc.entity.HeavyQueryCacheEntry;
import org.artifactory.test.ArtifactoryHomeBoundJUnitTest;
import org.artifactory.test.ArtifactoryHomeStub;
import org.awaitility.Awaitility;
import org.jfrog.common.JsonUtils;
import org.jfrog.common.concurrency.ConflictGuard;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.Mock;
import org.mockito.Spy;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.artifactory.common.ConstantValues.dbOperationsBulkSize;
import static org.artifactory.common.ConstantValues.hqcCleanupJobCron;
import static org.artifactory.common.ConstantValues.hqcEnabled;
import static org.artifactory.common.ConstantValues.hqcExpiryFactor;
import static org.artifactory.common.ConstantValues.hqcMinCallableDurationMillis;
import static org.artifactory.hqc.enums.HeavyQueryCacheType.TOTAL_ARTIFACT_COUNT;
import static org.artifactory.hqc.enums.HeavyQueryCacheType.XRAY_INDEXES;
import static org.artifactory.hqc.service.HeavyQueryCacheServiceImpl.CLEANUP_JOB_DEFAULT_CRON;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertSame;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyList;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.clearInvocations;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;

/**
 * @author Tomer Slutsky
 * @author Andrey Tashlitsky
 */
class HeavyQueryCacheServiceImplTest extends ArtifactoryHomeBoundJUnitTest {

    private static final long DELAY_FOR_CACHE = 110L;
    private static final long DELAY_FOR_NO_CACHE = 30L;
    private static final String CACHE_KEY = TOTAL_ARTIFACT_COUNT + "_repo";
    private final ArtifactoryHomeStub artifactoryHomeStub = getOrCreateArtifactoryHomeStub();
    @Mock
    private DbService dbService;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private TaskService taskService;
    @Mock
    private HaCommonAddon haCommonAddon;
    @Mock
    private ConflictsGuard conflictsGuard;
    @Mock
    private ConflictGuard conflictGuard;
    @Spy
    private HeavyQueryCacheDaoMock hqcDaoSpy;
    private HeavyQueryCacheServiceImpl hqcService;
    private AutoCloseable autoCloseable;

    @BeforeEach
    public void beforeEach() throws InterruptedException {
        autoCloseable = openMocks(this);
        when(haCommonAddon.getConflictsGuard(anyString())).thenReturn(conflictsGuard);
        when(conflictsGuard.getLock(anyString())).thenReturn(conflictGuard);
        when(conflictGuard.tryToLock(anyLong(), any(TimeUnit.class))).thenReturn(true);
        when(addonsManager.addonByType(HaCommonAddon.class)).thenReturn(haCommonAddon);
        hqcService = new HeavyQueryCacheServiceImpl(hqcDaoSpy, dbService, addonsManager, taskService);
        artifactoryHomeStub.setProperty(hqcMinCallableDurationMillis.getPropertyName(), "100");
        artifactoryHomeStub.setProperty(hqcExpiryFactor.getPropertyName(), hqcExpiryFactor.getDefValue());
        artifactoryHomeStub.setProperty(hqcEnabled.getPropertyName(), hqcEnabled.getDefValue());
        clearInvocations(hqcDaoSpy);
    }

    @AfterEach
    public void afterEach() throws Exception {
        autoCloseable.close();
    }

    @ParameterizedTest
    @MethodSource("getDataProvider")
    <T> void getTest(Callable<T> callable, Class<T> clazz, HeavyQueryCacheType hqcType, boolean useCache, T expected) {
        HeavyQueryCacheConfig hqcDefaultConfig = HeavyQueryCacheConfig.builder().minCallableDurationMillis(100).build();
        String cacheKey = hqcType.toString() + "_repo";
        if (useCache) {
            assertNull(hqcDaoSpy.getCachedResponse(cacheKey, 1L));
            hqcService.get(callable, clazz, "repo", hqcType, hqcDefaultConfig);
            assertNotNull(hqcDaoSpy.getCachedResponse(cacheKey, 1L));
            clearInvocations(hqcDaoSpy);
            T actual = hqcService.get(callable, clazz, "repo", hqcType, hqcDefaultConfig);
            verify(hqcDaoSpy, times(1)).getCachedResponse(eq(cacheKey), anyLong());
            assertEquals(expected, actual);
        } else {
            hqcService.get(callable, clazz, "repo", hqcType, hqcDefaultConfig);
            assertNotNull(hqcDaoSpy.getCachedResponse(cacheKey, 1L));
        }
        byte[] data = hqcDaoSpy.getCachedResponse(cacheKey, 1L).getData();
        assertEquals(expected, JsonUtils.getInstance().readValue(data, clazz));
    }

    @Test
    void getFeatureDisabledTest() {
        artifactoryHomeStub.setProperty(hqcEnabled.getPropertyName(), "false");
        Object o = new Object();
        Object result =
                hqcService.get(() -> o, Object.class, "repo", TOTAL_ARTIFACT_COUNT,
                        HeavyQueryCacheConfig.builder().build());
        assertSame(o, result);
    }

    @Test
    void getFeatureDisabledExceptionTest() {
        artifactoryHomeStub.setProperty(hqcEnabled.getPropertyName(), "false");
        Callable<Object> c = () -> {
            throw new Exception("cool");
        };
        Object result =
                hqcService.get(c, Object.class, "repo", TOTAL_ARTIFACT_COUNT, HeavyQueryCacheConfig.builder().build());
        assertNull(result);
    }

    @Test
    void getTestNoCacheDueToTime() {
        HeavyQueryCacheConfig hqcConfig = HeavyQueryCacheConfig.builder().build();
        assertNull(hqcDaoSpy.getCachedResponse(CACHE_KEY, 1L));
        hqcService.get(getInfo(1, DELAY_FOR_NO_CACHE), Integer.class, "repo", TOTAL_ARTIFACT_COUNT, hqcConfig);
        assertNull(hqcDaoSpy.getCachedResponse(CACHE_KEY, 1L));
        verify(hqcDaoSpy, times(4)).getCachedResponse(eq(CACHE_KEY), anyLong());
    }

    @Test
    void getTestNoCacheDueToSize() {
        HeavyQueryCacheConfig hqcConfig = HeavyQueryCacheConfig.builder().build();
        assertNull(hqcDaoSpy.getCachedResponse(CACHE_KEY, 1L));
        hqcService.get(getInfo(new byte[1024 * 1024 + 1], DELAY_FOR_CACHE), byte[].class, "repo", TOTAL_ARTIFACT_COUNT,
                hqcConfig);
        assertNull(hqcDaoSpy.getCachedResponse(CACHE_KEY, 1L));
        verify(hqcDaoSpy, times(4)).getCachedResponse(eq(CACHE_KEY), anyLong());
    }

    @Test
    void getTestRequestTimeout() throws Exception {
        HeavyQueryCacheConfig hqcConfig = HeavyQueryCacheConfig.builder().build();
        when(conflictGuard.tryToLock(anyLong(), any(TimeUnit.class))).thenReturn(false);
        HeavyQueryCacheException e = assertThrows(HeavyQueryCacheException.class,
                () -> hqcService.get(() -> new byte[1024], byte[].class, "repo", TOTAL_ARTIFACT_COUNT, hqcConfig));
        assertTrue(e.getMessage()
                .contains("The query is in progress, but a timeout has occurred for the current request"));
    }

    @Test
    void getTestDeleteOldEntry() {
        artifactoryHomeStub.setProperty(hqcExpiryFactor.getPropertyName(), "0");
        HeavyQueryCacheConfig hqcConfig = HeavyQueryCacheConfig.builder().build();
        assertNull(hqcDaoSpy.getCachedResponse(CACHE_KEY, 1L));
        hqcService.get(getInfo(new byte[1024], DELAY_FOR_CACHE), byte[].class, "repo", TOTAL_ARTIFACT_COUNT, hqcConfig);
        assertNull(hqcDaoSpy.getCachedResponse(CACHE_KEY, System.currentTimeMillis()));
        hqcService.get(getInfo(new byte[1024], DELAY_FOR_CACHE), byte[].class, "repo", TOTAL_ARTIFACT_COUNT, hqcConfig);
        assertNull(hqcDaoSpy.getCachedResponse(CACHE_KEY, System.currentTimeMillis()));
        hqcService.get(getInfo(new byte[1024], DELAY_FOR_CACHE), byte[].class, "repo", TOTAL_ARTIFACT_COUNT, hqcConfig);
        verify(hqcDaoSpy, times(9)).getCachedResponse(eq(CACHE_KEY), anyLong());
        assertEquals(1, hqcDaoSpy.getHqcMap().size()); // previous entries are deleted when caching new entry
    }

    @Test
    void runCleanupTest() {
        artifactoryHomeStub.setProperty(dbOperationsBulkSize.getPropertyName(), "2");
        long currentTimeMillis = System.currentTimeMillis();
        List.of(
                HeavyQueryCacheEntry.builder().id(1).cacheKey("1").expiryTime(currentTimeMillis - 10_000).build(),
                HeavyQueryCacheEntry.builder().id(2).cacheKey("2").expiryTime(currentTimeMillis + 10_000).build(),
                HeavyQueryCacheEntry.builder().id(3).cacheKey("3").expiryTime(currentTimeMillis - 20_000).build(),
                HeavyQueryCacheEntry.builder().id(4).cacheKey("4").expiryTime(currentTimeMillis + 20_000).build(),
                HeavyQueryCacheEntry.builder().id(5).cacheKey("5").expiryTime(currentTimeMillis - 30_000).build(),
                HeavyQueryCacheEntry.builder().id(6).cacheKey("6").expiryTime(currentTimeMillis + 30_000).build(),
                HeavyQueryCacheEntry.builder().id(7).cacheKey("7").expiryTime(currentTimeMillis - 40_000).build(),
                HeavyQueryCacheEntry.builder().id(8).cacheKey("8").expiryTime(currentTimeMillis + 40_000).build(),
                HeavyQueryCacheEntry.builder().id(9).cacheKey("9").expiryTime(currentTimeMillis - 50_000).build(),
                HeavyQueryCacheEntry.builder().id(0).cacheKey("0").expiryTime(currentTimeMillis + 50_000).build()
        ).forEach(e -> hqcDaoSpy.create(e));
        int count = hqcService.runCleanup();
        assertTrue(count > 0);
        assertEquals(5, count);
        Map<Long, HeavyQueryCacheEntry> cache = hqcDaoSpy.getHqcMap();
        assertEquals(5, cache.values().size());
        Set<Long> ids = cache.values().stream()
                .map(HeavyQueryCacheEntry::getId)
                .collect(Collectors.toSet());
        assertEquals(Set.of(2L, 4L, 6L, 8L, 0L), ids);
        verify(hqcDaoSpy, times(10)).create(any(HeavyQueryCacheEntry.class));
        verify(hqcDaoSpy, times(4)).getExpiredEntries(anyLong()); // 3 batches, 1 empty (2,2,1,0)
        verify(hqcDaoSpy, times(3)).delete(anyList());
        verify(hqcDaoSpy, times(1)).getHqcMap();
        verifyNoMoreInteractions(hqcDaoSpy);
    }

    @ParameterizedTest
    @MethodSource("cronDataProvider")
    void getCronExpressionTest(String providedCronExpression, String expectedCron) {
        artifactoryHomeStub.setProperty(hqcCleanupJobCron.getPropertyName(), providedCronExpression);
        String cronExpression = hqcService.getCronExpression();
        assertEquals(expectedCron, cronExpression);
    }

    @ParameterizedTest
    @MethodSource("shouldCacheDataProvider")
    void shouldCacheTest(long callbackDuration, int blobLength, boolean expected) {
        HeavyQueryCacheConfig hqcConfig = HeavyQueryCacheConfig.builder().build();
        boolean shouldCache = hqcService.shouldCache(callbackDuration, blobLength, hqcConfig);
        assertEquals(expected, shouldCache);
    }

    @ParameterizedTest
    @MethodSource("getExpiryTimeDataProvider")
    void getExpiryTimeTest(long cacheTime, long callableDuration, int expiryFactor, int expiryLowerBound,
            int expiryUpperBound, long expectedCachePeriod) {
        HeavyQueryCacheConfig hqcConfig = HeavyQueryCacheConfig.builder()
                .expiryFactor(expiryFactor)
                .expiryLowerBound(expiryLowerBound)
                .expiryUpperBound(expiryUpperBound)
                .build();
        long expiryTime = hqcService.getExpiryTime(cacheTime, callableDuration, hqcConfig);
        assertEquals(cacheTime + expectedCachePeriod, expiryTime);
    }

    private static Stream<Arguments> getDataProvider() {
        return Stream.of(
                Arguments.of(getInfo(1, DELAY_FOR_CACHE), Integer.class, XRAY_INDEXES, false, 1),
                Arguments.of(getInfo("bla", DELAY_FOR_CACHE), String.class, XRAY_INDEXES, false, "bla"),
                Arguments.of(getInfo(JsonUtils.getInstance().readTree("{\"repo\":\"something\",\n " +
                                "\"another\": \"else\"}"), DELAY_FOR_CACHE), JsonNode.class, XRAY_INDEXES, false,
                        JsonUtils.getInstance().readTree("{\"repo\":\"something\"," +
                                "\"another\":\"else\"}")),
                Arguments.of(getInfo(1, DELAY_FOR_CACHE), Integer.class, TOTAL_ARTIFACT_COUNT, false, 1),
                Arguments.of(getInfo("bla", DELAY_FOR_CACHE), String.class, TOTAL_ARTIFACT_COUNT, false, "bla"),
                Arguments.of(getInfo(JsonUtils.getInstance().readTree("{\"repo\":\"something\",\n " +
                                "\"another\": \"else\"}"), DELAY_FOR_CACHE), JsonNode.class, TOTAL_ARTIFACT_COUNT,
                        false,
                        JsonUtils.getInstance().readTree("{\"repo\":\"something\"," +
                                "\"another\":\"else\"}")),
                Arguments.of(getInfo(1, DELAY_FOR_CACHE), Integer.class, XRAY_INDEXES, true, 1),
                Arguments.of(getInfo("bla", DELAY_FOR_CACHE), String.class, XRAY_INDEXES, true, "bla"),
                Arguments.of(getInfo(JsonUtils.getInstance().readTree("{\"repo\":\"something\",\n " +
                                "\"another\": \"else\"}"), DELAY_FOR_CACHE), JsonNode.class, XRAY_INDEXES, true,
                        JsonUtils.getInstance().readTree("{\"repo\":\"something\"," +
                                "\"another\":\"else\"}")),
                Arguments.of(getInfo(1, DELAY_FOR_CACHE), Integer.class, TOTAL_ARTIFACT_COUNT, true, 1),
                Arguments.of(getInfo("bla", DELAY_FOR_CACHE), String.class, TOTAL_ARTIFACT_COUNT, true, "bla"),
                Arguments.of(getInfo(JsonUtils.getInstance().readTree("{\"repo\":\"something\",\n " +
                                "\"another\": \"else\"}"), DELAY_FOR_CACHE), JsonNode.class, TOTAL_ARTIFACT_COUNT, true,
                        JsonUtils.getInstance().readTree("{\"repo\":\"something\"," +
                                "\"another\":\"else\"}"))
        );
    }

    private static <T> Callable<T> getInfo(T data, long delay) {
        return () -> {
            Awaitility.await().pollDelay(delay, TimeUnit.MILLISECONDS).untilTrue(new AtomicBoolean(true));
            return data;
        };
    }

    private static Stream<Arguments> cronDataProvider() {
        return Stream.of(
                Arguments.of(CLEANUP_JOB_DEFAULT_CRON, CLEANUP_JOB_DEFAULT_CRON),
                Arguments.of("2 30 * * * *", CLEANUP_JOB_DEFAULT_CRON), // negative, invalid cron expression
                Arguments.of("0 0 0/1 1/1 * ? *", CLEANUP_JOB_DEFAULT_CRON), // negative, every hour
                Arguments.of("0 0 12 1 1/1 ? *", CLEANUP_JOB_DEFAULT_CRON), // negative, every month
                Arguments.of("0 0 12 ? * */8 *", CLEANUP_JOB_DEFAULT_CRON), // negative, every 8 days
                Arguments.of("0 0 12 ? * */2 *", "0 0 12 ? * */2 *"), // every 2 days
                Arguments.of(null, CLEANUP_JOB_DEFAULT_CRON)
        );
    }

    private static Stream<Arguments> shouldCacheDataProvider() {
        return Stream.of(
                Arguments.of(500, 1024, true), // 1KB, valid size & duration
                Arguments.of(500, 1024 * 1024, false), // 1MB, invalid size, valid duration
                Arguments.of(500, 1024 * 1024 - 1, true), // valid size & duration
                Arguments.of(50, 1024, false) // valid size, invalid duration
        );
    }

    private static Stream<Arguments> getExpiryTimeDataProvider() {
        long currentTimeMillis = System.currentTimeMillis();
        return Stream.of(
                Arguments.of(currentTimeMillis, 100, 10, 1, 10_000, 1_000), // in bounds
                Arguments.of(currentTimeMillis, 100, 10, 1_100, 10_000, 1_100), // too low, returning lower
                Arguments.of(currentTimeMillis, 100, 10, 1, 900, 900) // too high, returning higher
        );
    }
}
