package org.artifactory.maven.index;

import org.apache.lucene.store.FSDirectory;
import org.apache.maven.index.context.IndexingContext;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.mime.MavenNaming;
import org.artifactory.mime.MimeTypes;
import org.artifactory.model.xstream.fs.FileInfoImpl;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.StoringRepo;
import org.artifactory.resource.FileResource;
import org.artifactory.resource.ResourceStreamHandle;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.test.util.ReflectionTestUtils;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.nio.file.Files;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

import static org.artifactory.mime.MimeType.def;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class RepoIndexerTest {

    @Mock
    private ArtifactoryHome artifactoryHome;
    private AutoCloseable mocks;

    @BeforeEach
    void setup() {
        mocks = MockitoAnnotations.openMocks(this);
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        MimeTypes mimeTypes = mock(MimeTypes.class);
        when(mimeTypes.getByExtension(any())).thenReturn(def);
        when(artifactoryHome.getMimeTypes()).thenReturn(mimeTypes);
        when(context.getArtifactoryHome()).thenReturn(artifactoryHome);
        ArtifactoryHome.bind(artifactoryHome);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterEach
    void tearDown() throws Exception {
        ArtifactoryContextThreadBinder.unbind();
        ArtifactoryHome.unbind();
        mocks.close();
    }

    @Test
    void testMergeIntoNoLeftovers() throws Exception {
        StoringRepo virtualRepo = mock(StoringRepo.class);
        RepoIndexer repoIndexer = new RepoIndexer(virtualRepo);
        StoringRepo localRepo = mock(StoringRepo.class);
        RepoPath nexusIndexFile = RepoPathFactory.create("l", MavenNaming.NEXUS_INDEX_GZ_PATH);
        when(localRepo.getRepoPath(MavenNaming.NEXUS_INDEX_GZ_PATH)).thenReturn(
                nexusIndexFile);
        FileResource fileResource = new FileResource(new FileInfoImpl(nexusIndexFile));
        when(localRepo.getInfo(any())).thenReturn(fileResource);
        Map<StoringRepo, FSDirectory> extractedRepoIndexes = new HashMap<>();
        ResourceStreamHandle resourceStreamHandle = mock(ResourceStreamHandle.class);
        ByteArrayInputStream inputStream = new ByteArrayInputStream(new byte[]{1, 2, 3, 4, 5, 6, 7, 8, 0});
        when(resourceStreamHandle.getInputStream()).thenReturn(inputStream);
        when(localRepo.getResourceStreamHandle(any(), any())).thenReturn(resourceStreamHandle);

        File tempDirectory = Files.createTempDirectory(getClass().getName()).toFile();
        when(artifactoryHome.getTempWorkDir()).thenReturn(tempDirectory);
        try {
            IndexingContext indexingContext = mock(IndexingContext.class);
            ReflectionTestUtils.setField(repoIndexer, "context", indexingContext);
            repoIndexer.mergeInto(localRepo, extractedRepoIndexes);
            assertEquals(0, Objects.requireNonNull(tempDirectory.list()).length);
        } finally {
            boolean delete = tempDirectory.delete();
            if (!delete) {
                tempDirectory.deleteOnExit();
            }
        }

    }
}