package org.artifactory.ph;

import lombok.NonNull;
import org.apache.commons.io.IOUtils;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpRequestBase;
import org.artifactory.api.repo.exception.RepoRejectException;
import org.artifactory.api.request.ArtifactoryHeaders;
import org.artifactory.api.request.ArtifactoryResponse;
import org.artifactory.api.request.DownloadService;
import org.artifactory.api.request.InternalArtifactoryRequest;
import org.artifactory.common.resource.BaseResource;
import com.jfrog.ph.acl.exception.PackageException;
import com.jfrog.ph.acl.model.PackageArtifact;
import com.jfrog.ph.acl.model.PackageDownloadContext;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.fs.FileInfo;
import org.artifactory.fs.RepoResource;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.xstream.fs.PropertiesImpl;
import org.artifactory.repo.HttpRepo;
import org.artifactory.repo.LocalCacheRepo;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RemoteRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.repo.virtual.VirtualRepo;
import org.artifactory.request.InternalRequestContext;
import org.artifactory.resource.MutableRepoResourceInfo;
import org.artifactory.resource.RemoteRepoResource;
import org.artifactory.resource.ResourceStreamHandle;
import org.artifactory.resource.UnfoundRepoResource;
import org.artifactory.rest.exception.RestException;
import org.artifactory.sapi.fs.VfsFile;
import org.artifactory.util.StringInputStream;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.Mockito;

import javax.ws.rs.core.Response;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.stream.Stream;

import static org.apache.http.HttpStatus.SC_NOT_FOUND;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;

public class PackageArtifactServiceImplTest {

    @Mock
    private DownloadService downloadService;
    @Mock
    private InternalRepositoryService repositoryService;
    @Mock
    private VirtualRepo virtualRepo;

    private PackageArtifactServiceImpl packageDownloadService;

    @BeforeEach
    public void setUp() {
        openMocks(this);
        packageDownloadService = new PackageArtifactServiceImpl(downloadService, repositoryService);
    }

    @Test
    void testDownload() throws PackageException, IOException {
        PackageDownloadContext context = new PackageDownloadContext("repoKey", "path");
        context.header("key", "val");
        Response response = packageDownloadService.download(context);

        ArgumentCaptor<InternalArtifactoryRequest> captor = ArgumentCaptor.forClass(InternalArtifactoryRequest.class);
        verify(downloadService).process(captor.capture(), any(ArtifactoryResponse.class));
        final List<InternalArtifactoryRequest> params = captor.getAllValues();
        Assertions.assertEquals(200, response.getStatus());
        Assertions.assertEquals("val", params.get(0).getHeaderValues("key").nextElement());
    }

    @Test
    void testDownloadFormRemoteCache() throws PackageException, IOException {
        LocalCacheRepo localRepo = mock(LocalCacheRepo.class);
        when(localRepo.getKey()).thenReturn("repoKey-cache");
        when(repositoryService.localFederatedOrCachedRepositoryByKey("repoKey")).thenReturn(localRepo);
        PackageDownloadContext context = new PackageDownloadContext("repoKey", "path");
        context.header("key", "val");
        Response response = packageDownloadService.downloadFromRemoteCache(context);

        ArgumentCaptor<InternalArtifactoryRequest> captor = ArgumentCaptor.forClass(InternalArtifactoryRequest.class);
        verify(downloadService).process(captor.capture(), any(ArtifactoryResponse.class));
        final List<InternalArtifactoryRequest> params = captor.getAllValues();
        Assertions.assertEquals(200, response.getStatus());
        Assertions.assertEquals("val", params.get(0).getHeaderValues("key").nextElement());
    }

    @Test
    void testDownloadForbiddenRestException() throws IOException {
        PackageDownloadContext context = new PackageDownloadContext("repoKey", "path");
        context.header("key", "val");
        doThrow(new RestException(HttpStatus.SC_FORBIDDEN, "Forbidden")).when(downloadService).process(any(), any());
        PackageException exception =
                Assertions.assertThrows(PackageException.class, () -> packageDownloadService.download(context));
        Assertions.assertEquals(403, exception.getStatus());
        Assertions.assertEquals("Forbidden", exception.getInternalReason());
        Assertions.assertEquals("Forbidden", exception.getPublicReason());
    }

    @Test
    void testDownloadFromRemoteCacheRepoNotExist() {
        PackageDownloadContext context = new PackageDownloadContext("repoKey", "path");
        try {
            packageDownloadService.downloadFromRemoteCache(context);
            Assertions.fail("Expected exception");
        } catch (PackageException e) {
            Assertions.assertEquals(e.getStatus(), HttpStatus.SC_BAD_REQUEST);
            Assertions.assertEquals(e.getInternalReason(),
                    "Repo 'repoKey' does not exist or storing artifacts locally is disabled");
        }
    }

    @Test
    void testDownloadFrommRemoteCacheBadInput() {
        PackageDownloadContext context = new PackageDownloadContext("repoKey-cache", "path");
        try {
            packageDownloadService.downloadFromRemoteCache(context);
            Assertions.fail("Expected exception");
        } catch (PackageException e) {
            Assertions.assertEquals(e.getStatus(), HttpStatus.SC_BAD_REQUEST);
            Assertions.assertEquals(e.getInternalReason(),
                    "Download request from remote-cache repo should not include '-cache' in the repo key");
        }
    }

    @Test
    void testGetArtifactFromVirtual() throws PackageException {
        String virtualRepoKey = "virtual";
        String localRepoKey = "local";
        String path = "path";
        RepoPath virtualRepoPath = RepoPathFactory.create(virtualRepoKey, path);
        RepoPath localRepoPath = RepoPathFactory.create(localRepoKey, path);
        @SuppressWarnings("unchecked")
        LocalRepo<LocalRepoDescriptor, LocalRepoConfig> localRepo = Mockito.mock(LocalRepo.class);
        VirtualRepo virtualRepo = Mockito.mock(VirtualRepo.class);
        Mockito.when(virtualRepo.getKey()).thenReturn(virtualRepoKey);
        RepoResource repoResource = info(localRepoKey, path);
        Mockito.when(virtualRepo.getInfo(any())).thenReturn(repoResource);
        FileInfo fileInfo = mock(FileInfo.class);
        VfsFile<?> file = mock(VfsFile.class);
        Mockito.when(fileInfo.getRepoKey()).thenReturn(localRepoKey);
        Mockito.when(fileInfo.getRepoPath()).thenReturn(virtualRepoPath);
        Mockito.when(fileInfo.getRelPath()).thenReturn(path);
        Mockito.when(file.getRepoKey()).thenReturn(localRepoKey);
        Mockito.when(file.getRepoPath()).thenReturn(localRepoPath);
        Mockito.when(file.getPath()).thenReturn(path);
        Mockito.when(file.getProperties()).thenReturn(new PropertiesImpl());
        Mockito.when(localRepo.getKey()).thenReturn(localRepoKey);
        Mockito.when(localRepo.getImmutableFile(localRepoPath)).thenReturn(file);
        LocalRepoDescriptor localDescriptor = mock(LocalRepoDescriptor.class);
        VirtualRepoDescriptor virtualDescriptor = mock(VirtualRepoDescriptor.class);
        Mockito.when(localDescriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(virtualDescriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(localRepo.getDescriptor()).thenReturn(localDescriptor);
        Mockito.when(virtualRepo.getDescriptor()).thenReturn(virtualDescriptor);
        Mockito.when(repositoryService.localFederatedOrCachedRepositoryByKey(localRepoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.repositoryByKey(localRepoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.repositoryByKey(virtualRepoKey)).thenReturn(virtualRepo);
        Mockito.when(repositoryService.virtualRepositoryByKey(virtualRepoKey)).thenReturn(virtualRepo);
        Mockito.when(repositoryService.getVirtualFileInfo(virtualRepoPath)).thenReturn(fileInfo);
        PackageArtifact artifact = packageDownloadService.getArtifact(virtualRepoKey, path);
        Assertions.assertEquals(localRepoKey, artifact.getRepoKey());
        Assertions.assertEquals(path, artifact.getPath());
    }

    @Test
    void testGetArtifactFromLocal() throws PackageException, IOException {
        String repoKey = "repoKey";
        String path = "path";
        RepoPath repoPath = RepoPathFactory.create(repoKey, path);
        LocalRepo<LocalRepoDescriptor, LocalRepoConfig> localRepo = Mockito.mock(LocalRepo.class);
        VfsFile<?> file = Mockito.mock(VfsFile.class);
        Mockito.when(file.getStream()).thenReturn(new StringInputStream("Test"));
        Mockito.when(localRepo.getImmutableFile(repoPath)).thenReturn(file);
        Mockito.when(localRepo.isReal()).thenReturn(true);
        Mockito.when(localRepo.getKey()).thenReturn(repoKey);
        Mockito.when(file.getRepoKey()).thenReturn(repoKey);
        Mockito.when(file.getRepoPath()).thenReturn(repoPath);
        Mockito.when(file.getPath()).thenReturn(path);
        Mockito.when(file.getProperties()).thenReturn(new PropertiesImpl());
        Mockito.when(localRepo.getImmutableFile(repoPath)).thenReturn(file);
        LocalRepoDescriptor descriptor = mock(LocalRepoDescriptor.class);
        Mockito.when(descriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(localRepo.getDescriptor()).thenReturn(descriptor);
        RepoResource info = info(repoKey, path);
        Mockito.when(localRepo.getInfo(any())).thenReturn(info);
        Mockito.when(repositoryService.repositoryByKey(repoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.localFederatedOrCachedRepositoryByKey(repoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.localOrFederatedRepositoryByKey(repoKey)).thenReturn(localRepo);
        PackageArtifact artifact = packageDownloadService.getArtifact(repoKey, path);
        Assertions.assertEquals(repoKey, artifact.getRepoKey());
        Assertions.assertEquals(path, artifact.getPath());
    }

    @Test
    void testExecuteRemoteHttpRequestRemoteNotExists() {
        PackageException exception = Assertions.assertThrows(PackageException.class,
                () -> packageDownloadService.executeRemoteHttpRequest("remote", null));
        Assertions.assertEquals("Remote repo remote does not exist", exception.getInternalReason());
        Assertions.assertEquals("Remote repo remote does not exist", exception.getPublicReason());
        Assertions.assertEquals(SC_NOT_FOUND, exception.getStatus());
    }

    @ParameterizedTest
    @MethodSource("testExecuteRemoteHttpRequestRemoteNotAvailableProvider")
    void testExecuteRemoteHttpRequestRemoteNotAvailable(boolean isOffline, boolean isAssumedOffline,
            boolean isBlackedOut) {
        RemoteRepo remoteRepo = mock(RemoteRepo.class);
        when(remoteRepo.isOffline()).thenReturn(isOffline);
        when(remoteRepo.isAssumedOffline()).thenReturn(isAssumedOffline);
        when(remoteRepo.isBlackedOut()).thenReturn(isBlackedOut);
        when(repositoryService.remoteRepositoryByKey("remote")).thenReturn(remoteRepo);
        PackageException exception = Assertions.assertThrows(PackageException.class,
                () -> packageDownloadService.executeRemoteHttpRequest("remote", null));
        Assertions.assertEquals("Remote repo remote is offline", exception.getInternalReason());
        Assertions.assertEquals("Remote repo remote is offline", exception.getPublicReason());
        Assertions.assertEquals(SC_NOT_FOUND, exception.getStatus());
    }

    @Test
    void testExecuteRemoteHttpRequest() throws IOException, PackageException {
        HttpRepo remoteRepo = mock(HttpRepo.class);
        CloseableHttpResponse closeableHttpResponse = mock(CloseableHttpResponse.class);
        HttpRequestBase httpRequestBase = mock(HttpRequestBase.class);
        when(remoteRepo.executeMethod(httpRequestBase)).thenReturn(closeableHttpResponse);
        when(repositoryService.remoteRepositoryByKey("remote")).thenReturn(remoteRepo);
        CloseableHttpResponse response = packageDownloadService.executeRemoteHttpRequest("remote", httpRequestBase);
        Assertions.assertEquals(closeableHttpResponse, response);
        verify(remoteRepo, times(1)).executeMethod(any());
    }

    @Test
    void testGetArtifactFromRemote() throws PackageException, IOException {
        String remoteRepoKey = "remoteRepoKey";
        String localRepoKey = "localRepoKey";
        String path = "path";
        RepoPath remoteRepoPath = RepoPathFactory.create(remoteRepoKey, path);
        RepoPath localRepoPath = RepoPathFactory.create(localRepoKey, path);
        RemoteRepo<RemoteRepoDescriptor, RemoteRepoConfig> remoteRepo = Mockito.mock(RemoteRepo.class);
        LocalRepo<LocalRepoDescriptor, LocalRepoConfig> localRepo = Mockito.mock(LocalRepo.class);
        VfsFile<?> file = Mockito.mock(VfsFile.class);
        Mockito.when(file.getStream()).thenReturn(new StringInputStream("Test"));
        Mockito.when(localRepo.getImmutableFile(localRepoPath)).thenReturn(file);
        LocalRepoDescriptor localRepoDescriptor = mock(LocalRepoDescriptor.class);
        Mockito.when(localRepoDescriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(localRepo.getDescriptor()).thenReturn(localRepoDescriptor);
        Mockito.when(localRepo.getKey()).thenReturn(localRepoKey);
        Mockito.when(remoteRepo.isReal()).thenReturn(true);
        Mockito.when(remoteRepo.isStoreArtifactsLocally()).thenReturn(true);
        Mockito.when(remoteRepo.getKey()).thenReturn(remoteRepoKey);
        Mockito.when(file.getRepoKey()).thenReturn(remoteRepoKey);
        Mockito.when(file.getRepoPath()).thenReturn(remoteRepoPath);
        Mockito.when(file.getPath()).thenReturn(path);
        Mockito.when(file.getProperties()).thenReturn(new PropertiesImpl());
        RemoteRepoDescriptor descriptor = mock(RemoteRepoDescriptor.class);
        Mockito.when(descriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(remoteRepo.getDescriptor()).thenReturn(descriptor);
        RepoResource remote = remoteInfo(remoteRepoKey, path);
        RepoResource local = info(remoteRepoKey, path);
        Mockito.when(remoteRepo.getInfo(any())).thenReturn(remote);
        Mockito.when(localRepo.getInfo(any())).thenReturn(local);
        Mockito.when(repositoryService.repositoryByKey(remoteRepoKey)).thenReturn(remoteRepo);
        Mockito.when(repositoryService.remoteRepositoryByKey(remoteRepoKey)).thenReturn(remoteRepo);
        Mockito.when(repositoryService.repositoryByKey(remoteRepoKey)).thenReturn(remoteRepo);
        Mockito.when(repositoryService.localFederatedOrCachedRepositoryByKey(remoteRepoKey)).thenReturn(localRepo);
        PackageArtifact artifact = packageDownloadService.getArtifact(remoteRepoKey, path);
        Assertions.assertEquals(remoteRepoKey, artifact.getRepoKey());
        Assertions.assertEquals(path, artifact.getPath());
    }

    @Test
    void testGetArtifactFromRemoteNoCache() throws PackageException, IOException {
        String remoteRepoKey = "remoteRepoKey";
        String path = "path";
        RemoteRepo<RemoteRepoDescriptor, ?> remoteRepo = mock(RemoteRepo.class);
        Mockito.when(remoteRepo.isReal()).thenReturn(true);
        Mockito.when(remoteRepo.getKey()).thenReturn(remoteRepoKey);
        RemoteRepoDescriptor descriptor = mock(RemoteRepoDescriptor.class);
        Mockito.when(descriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(remoteRepo.getDescriptor()).thenReturn(descriptor);
        RepoResource remoteResource = remoteInfo(remoteRepoKey, path);
        MutableRepoResourceInfo info = mock(MutableRepoResourceInfo.class);
        Mockito.when(info.getRepoPath()).thenReturn(RepoPathFactory.create(remoteRepoKey, path));
        when(remoteResource.getInfo()).thenReturn(info);
        Mockito.when(remoteRepo.getInfo(any())).thenReturn(remoteResource);
        Mockito.when(repositoryService.repositoryByKey(remoteRepoKey)).thenReturn(remoteRepo);
        Mockito.when(repositoryService.remoteRepositoryByKey(remoteRepoKey)).thenReturn(remoteRepo);
        PackageArtifact artifact = packageDownloadService.getArtifact(remoteRepoKey, path);
        Assertions.assertEquals(remoteRepoKey, artifact.getRepoKey());
        Assertions.assertEquals(path, artifact.getPath());
    }

    @Test
    void testGetStreamFromLocal() throws PackageException, IOException {
        String repoKey = "repoKey";
        String path = "path";
        RepoPath repoPath = RepoPathFactory.create(repoKey, path);
        LocalRepo<?, ?> localRepo = Mockito.mock(LocalRepo.class);
        VfsFile<?> file = Mockito.mock(VfsFile.class);
        Mockito.when(file.getStream()).thenReturn(new StringInputStream("Test"));
        Mockito.when(localRepo.getImmutableFile(repoPath)).thenReturn(file);
        Mockito.when(localRepo.isReal()).thenReturn(true);
        Mockito.when(localRepo.getKey()).thenReturn(repoKey);
        RepoResource info = info(repoKey, path);
        Mockito.when(localRepo.getInfo(any())).thenReturn(info);
        Mockito.when(localRepo.getKey()).thenReturn(repoKey);
        Mockito.when(repositoryService.repositoryByKey(repoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.localFederatedOrCachedRepositoryByKey(repoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.localOrFederatedRepositoryByKey(repoKey)).thenReturn(localRepo);
        PackageDownloadContext context = new PackageDownloadContext(repoKey, path, "alter");
        context.header("key", "val");
        InputStream stream = packageDownloadService.getStream(context);
        Assertions.assertEquals("Test", IOUtils.toString(stream, StandardCharsets.UTF_8));

        ArgumentCaptor<InternalRequestContext> captor = ArgumentCaptor.forClass(InternalRequestContext.class);
        verify(localRepo).getInfo(captor.capture());
        final List<InternalRequestContext> params = captor.getAllValues();
        Assertions.assertEquals("val", params.get(0).getRequest().getHeaderValues("key").nextElement());
        InternalArtifactoryRequest originRequest = (InternalArtifactoryRequest) params.get(0).getRequest();
        Assertions.assertEquals("alter",
                originRequest.getParameter(ArtifactoryHeaders.PARAM_ALTERNATIVE_REMOTE_DOWNLOAD_URL));
    }

    @Test
    void testGetStreamFromRemote() throws PackageException, IOException {
        String remoteRepoKey = "remoteRepoKey";
        String localRepoKey = "localRepoKey";
        String path = "path";
        RepoPath remoteRepoPath = RepoPathFactory.create(remoteRepoKey, path);
        RepoPath localRepoPath = RepoPathFactory.create(localRepoKey, path);
        RemoteRepo<RemoteRepoDescriptor, ?> remoteRepo = mock(RemoteRepo.class);
        LocalRepo<LocalRepoDescriptor, ?> localRepo = mock(LocalRepo.class);
        VfsFile<?> file = mock(VfsFile.class);
        Mockito.when(file.getStream()).thenReturn(new StringInputStream("Test"));
        Mockito.when(localRepo.getImmutableFile(localRepoPath)).thenReturn(file);
        LocalRepoDescriptor localRepoDescriptor = mock(LocalRepoDescriptor.class);
        Mockito.when(localRepoDescriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(localRepo.getDescriptor()).thenReturn(localRepoDescriptor);
        Mockito.when(localRepo.getKey()).thenReturn(localRepoKey);
        Mockito.when(remoteRepo.isReal()).thenReturn(true);
        Mockito.when(remoteRepo.isStoreArtifactsLocally()).thenReturn(true);
        Mockito.when(remoteRepo.getKey()).thenReturn(remoteRepoKey);
        Mockito.when(file.getRepoKey()).thenReturn(remoteRepoKey);
        Mockito.when(file.getRepoPath()).thenReturn(remoteRepoPath);
        Mockito.when(file.getPath()).thenReturn(path);
        Mockito.when(file.getProperties()).thenReturn(new PropertiesImpl());
        RemoteRepoDescriptor descriptor = mock(RemoteRepoDescriptor.class);
        Mockito.when(descriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(remoteRepo.getDescriptor()).thenReturn(descriptor);
        RepoResource remoteResource = remoteInfo(remoteRepoKey, path);
        RepoResource localResource = info(remoteRepoKey, path);
        Mockito.when(remoteRepo.getInfo(any())).thenReturn(remoteResource);
        Mockito.when(localRepo.getInfo(any())).thenReturn(localResource);
        Mockito.when(repositoryService.repositoryByKey(remoteRepoKey)).thenReturn(remoteRepo);
        Mockito.when(repositoryService.remoteRepositoryByKey(remoteRepoKey)).thenReturn(remoteRepo);
        Mockito.when(repositoryService.repositoryByKey(remoteRepoKey)).thenReturn(remoteRepo);
        Mockito.when(repositoryService.localFederatedOrCachedRepositoryByKey(remoteRepoKey)).thenReturn(localRepo);
        PackageDownloadContext context = new PackageDownloadContext(remoteRepoKey, path, "alter");
        context.header("key", "val");
        InputStream stream = packageDownloadService.getStream(context);
        Assertions.assertEquals("Test", IOUtils.toString(stream, StandardCharsets.UTF_8));

        ArgumentCaptor<InternalRequestContext> captor = ArgumentCaptor.forClass(InternalRequestContext.class);
        verify(remoteRepo).getInfo(captor.capture());
        final List<InternalRequestContext> params = captor.getAllValues();
        Assertions.assertEquals("val", params.get(0).getRequest().getHeaderValues("key").nextElement());
        InternalArtifactoryRequest originRequest = (InternalArtifactoryRequest) params.get(0).getRequest();
        Assertions.assertEquals("alter",
                originRequest.getParameter(ArtifactoryHeaders.PARAM_ALTERNATIVE_REMOTE_DOWNLOAD_URL));
    }

    @Test
    void testGetStreamFromRemoteNoCache() throws PackageException, IOException, RepoRejectException {
        String remoteRepoKey = "remoteRepoKey";
        String path = "path";
        RemoteRepo<RemoteRepoDescriptor, ?> remoteRepo = mock(RemoteRepo.class);
        Mockito.when(remoteRepo.isReal()).thenReturn(true);
        Mockito.when(remoteRepo.getKey()).thenReturn(remoteRepoKey);
        RemoteRepoDescriptor descriptor = mock(RemoteRepoDescriptor.class);
        Mockito.when(descriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(remoteRepo.getDescriptor()).thenReturn(descriptor);
        RepoResource remoteResource = remoteInfo(remoteRepoKey, path);
        MutableRepoResourceInfo remoteInfo = mock(MutableRepoResourceInfo.class);
        Mockito.when(remoteInfo.getRepoPath()).thenReturn(RepoPathFactory.create(remoteRepoKey, path));
        when(remoteResource.getInfo()).thenReturn(remoteInfo);
        Mockito.when(remoteRepo.getInfo(any())).thenReturn(remoteResource);
        Mockito.when(repositoryService.repositoryByKey(remoteRepoKey)).thenReturn(remoteRepo);
        Mockito.when(repositoryService.remoteRepositoryByKey(remoteRepoKey)).thenReturn(remoteRepo);
        ResourceStreamHandle resourceStreamHandle = mock(ResourceStreamHandle.class);
        when(resourceStreamHandle.getInputStream()).thenReturn(new StringInputStream("Test"));
        when(repositoryService.getResourceStreamHandle(any(), eq(remoteRepo), eq(remoteResource))).thenReturn(
                resourceStreamHandle);
        PackageDownloadContext context = new PackageDownloadContext(remoteRepoKey, path, "alter");
        context.header("key", "val");
        InputStream stream = packageDownloadService.getStream(context);
        Assertions.assertEquals("Test", IOUtils.toString(stream, StandardCharsets.UTF_8));

        ArgumentCaptor<InternalRequestContext> captor = ArgumentCaptor.forClass(InternalRequestContext.class);
        verify(remoteRepo).getInfo(captor.capture());
        final List<InternalRequestContext> params = captor.getAllValues();
        Assertions.assertEquals("val", params.get(0).getRequest().getHeaderValues("key").nextElement());
        InternalArtifactoryRequest originRequest = (InternalArtifactoryRequest) params.get(0).getRequest();
        Assertions.assertEquals("alter",
                originRequest.getParameter(ArtifactoryHeaders.PARAM_ALTERNATIVE_REMOTE_DOWNLOAD_URL));
    }

    @Test
    void testGetResource() throws IOException, PackageException {
        String repoKey = "repoKey";
        String path = "path";
        RepoPath repoPath = RepoPathFactory.create(repoKey, path);
        LocalRepo<?, ?> localRepo = Mockito.mock(LocalRepo.class);
        VfsFile<?> file = Mockito.mock(VfsFile.class);
        Mockito.when(file.getStream()).thenReturn(new StringInputStream("Test"));
        RepoResource resource = mock(RepoResource.class);
        Mockito.when(resource.getRepoPath()).thenReturn(repoPath);
        Mockito.when(resource.getRepoKey()).thenReturn(repoKey);
        Mockito.when(resource.getPath()).thenReturn(path);
        Mockito.when(localRepo.getInfo(any())).thenReturn(resource);
        Mockito.when(localRepo.isReal()).thenReturn(true);
        Mockito.when(repositoryService.repositoryByKey(repoKey)).thenReturn(localRepo);
        BaseResource result = packageDownloadService.getResource(new PackageDownloadContext(repoKey, path));
        Assertions.assertEquals(repoKey, result.getRepoKey());
        Assertions.assertEquals(path, result.getPath());
    }

    @Test
    void testGetResourceFromVirtualCacheRepoNotFound() {
        try {
            packageDownloadService.getResourceFromVirtualCache("repo1", "a/b/c.json");
            Assertions.fail("Should get bad request");
        } catch (PackageException e) {
            Assertions.assertEquals(e.getStatus(), HttpStatus.SC_NOT_FOUND);
        }
    }

    @Test
    void testGetExpiredResourceFromVirtualCache() throws PackageException {
        when(repositoryService.virtualRepositoryByKey(Mockito.eq("repo1"))).thenReturn(virtualRepo);
        RepoResource resource = mock(RepoResource.class);
        when(resource.isFound()).thenReturn(true);
        when(virtualRepo.getInfoFromVirtualCache(any())).thenReturn(resource);
        when(virtualRepo.isCachedResourceExpired(any())).thenReturn(true);
        BaseResource cachedResource = packageDownloadService.getResourceFromVirtualCache("repo1",
                "a/b/c.json");
        Assertions.assertTrue(cachedResource.isExpired(), "Resource should be expired");

    }

    @Test
    void testGetResourceFromVirtualCache() throws PackageException {
        when(repositoryService.virtualRepositoryByKey(Mockito.eq("repo1"))).thenReturn(virtualRepo);
        RepoResource resource = mock(RepoResource.class);
        when(virtualRepo.getInfoFromVirtualCache(any())).thenReturn(resource);
        BaseResource cachedResource = packageDownloadService.getResourceFromVirtualCache("repo1",
                "a/b/c.json");
        Assertions.assertNotNull(cachedResource);
    }

    @Test
    void testGetStreamFromVirtualCacheRepoNotFound() {
        PackageException exception = Assertions.assertThrows(PackageException.class,
                () -> packageDownloadService.getStreamFromVirtualCache("virtual", ".security/file"));
        Assertions.assertEquals("Virtual repo virtual does not exist", exception.getInternalReason());
        Assertions.assertEquals("Virtual repo virtual does not exist", exception.getPublicReason());
        Assertions.assertEquals(SC_NOT_FOUND, exception.getStatus());
    }

    @Test
    void testGetStreamFromVirtualCacheResourceNotFound() {
        String repoKey = "virtual";
        String path = "path";
        VirtualRepo virtualRepo = mock(VirtualRepo.class);
        when(repositoryService.virtualRepositoryByKey(repoKey)).thenReturn(virtualRepo);
        when(virtualRepo.getInfoFromVirtualCache(any())).thenReturn(
                new UnfoundRepoResource(null, null));
        PackageException exception = Assertions.assertThrows(PackageException.class,
                () -> packageDownloadService.getStreamFromVirtualCache(repoKey, path));
        Assertions.assertEquals("Failed to find artifact in virtual repo:virtual, cache path:path, cause: not found",
                exception.getInternalReason());
        Assertions.assertEquals("Not found", exception.getPublicReason());
        Assertions.assertEquals(SC_NOT_FOUND, exception.getStatus());
    }

    @Test
    void testGetStreamFromVirtualCacheResourceExpired() {
        String repoKey = "virtual";
        String path = "path";
        VirtualRepo virtualRepo = mock(VirtualRepo.class);
        when(repositoryService.virtualRepositoryByKey(repoKey)).thenReturn(virtualRepo);
        RepoResource expiredResource = mock(RepoResource.class);
        when(expiredResource.isFound()).thenReturn(true);
        when(virtualRepo.isCachedResourceExpired(expiredResource)).thenReturn(true);
        when(virtualRepo.getInfoFromVirtualCache(any())).thenReturn(expiredResource);
        PackageException exception = Assertions.assertThrows(PackageException.class,
                () -> packageDownloadService.getStreamFromVirtualCache(repoKey, path));
        Assertions.assertEquals("Failed to find artifact in virtual repo:virtual, cache path:path, cause: not found",
                exception.getInternalReason());
        Assertions.assertEquals("Not found", exception.getPublicReason());
        Assertions.assertEquals(SC_NOT_FOUND, exception.getStatus());
    }

    @Test
    void testGetStreamFromVirtualCacheResource() throws PackageException, RepoRejectException, IOException {
        String repoKey = "virtual";
        String path = "path";
        VirtualRepo virtualRepo = mock(VirtualRepo.class);
        when(repositoryService.virtualRepositoryByKey(repoKey)).thenReturn(virtualRepo);
        RepoResource resource = mock(RepoResource.class);
        when(resource.isFound()).thenReturn(true);
        when(virtualRepo.isCachedResourceExpired(resource)).thenReturn(false);
        when(virtualRepo.getInfoFromVirtualCache(any())).thenReturn(resource);
        ResourceStreamHandle resourceStreamHandle = mock(ResourceStreamHandle.class);
        InputStream inputStream = mock(InputStream.class);
        when(resourceStreamHandle.getInputStream()).thenReturn(inputStream);
        when(repositoryService.getResourceStreamHandle(any(), eq(virtualRepo), eq(resource))).thenReturn(
                resourceStreamHandle);
        InputStream res = packageDownloadService.getStreamFromVirtualCache(repoKey, path);
        Assertions.assertEquals(inputStream, res);
    }

    @NonNull
    private RepoResource info(String repoKey, String path) {
        RepoResource info = mock(RepoResource.class);
        Mockito.when(info.getRepoKey()).thenReturn(repoKey);
        Mockito.when(info.getPath()).thenReturn(path);
        Mockito.when(info.isExpired()).thenReturn(false);
        Mockito.when(info.isFound()).thenReturn(true);
        Mockito.when(info.isBlocked()).thenReturn(false);
        Mockito.when(info.getRepoPath()).thenReturn(RepoPathFactory.create(repoKey, path));
        return info;
    }

    @NonNull
    private RepoResource remoteInfo(String repoKey, String path) {
        RemoteRepoResource info = mock(RemoteRepoResource.class);
        Mockito.when(info.getRepoKey()).thenReturn(repoKey);
        Mockito.when(info.getPath()).thenReturn(path);
        Mockito.when(info.isExpired()).thenReturn(false);
        Mockito.when(info.isFound()).thenReturn(true);
        Mockito.when(info.isBlocked()).thenReturn(false);
        Mockito.when(info.getRepoPath()).thenReturn(RepoPathFactory.create(repoKey, path));
        return info;
    }

    private static Stream<Arguments> testExecuteRemoteHttpRequestRemoteNotAvailableProvider() {
        return Stream.of(
                Arguments.of(true, false, false),
                Arguments.of(false, true, false),
                Arguments.of(false, false, true));
    }
}