package org.artifactory.ph;

import com.google.common.collect.Lists;
import com.jfrog.ph.acl.exception.PackageException;
import org.apache.http.HttpStatus;
import org.artifactory.api.security.access.CreatedTokenInfo;
import org.artifactory.api.security.access.TokenNotFoundException;
import org.artifactory.api.security.access.UserTokenSpec;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.model.xstream.security.UserImpl;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.security.InternalSecurityService;
import org.artifactory.security.SaltedPassword;
import org.artifactory.security.SimpleUser;
import org.artifactory.security.access.AccessService;
import org.artifactory.security.access.CreatedTokenInfoImpl;
import org.artifactory.security.props.auth.OauthManager;
import org.artifactory.storage.security.service.UserGroupStoreService;
import org.jfrog.access.client.AccessClientException;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.context.SecurityContextImpl;

import java.util.concurrent.Callable;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyBoolean;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.ArgumentMatchers.refEq;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.reset;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;
import static org.testng.AssertJUnit.assertFalse;

class PackageSecurityServiceImplTest {

    @Mock
    private InternalSecurityService securityService;
    @Mock
    AccessService accessService;
    @Mock
    AuthenticationManager authenticationManager;
    @Mock
    UserGroupStoreService userGroupStoreService;
    @Mock
    OauthManager tokenManager;

    private PackageSecurityServiceImpl packageSecurityService;

    @BeforeEach
    public void setUp() {
        openMocks(this);
        packageSecurityService = new PackageSecurityServiceImpl(securityService, accessService,
                authenticationManager, userGroupStoreService, tokenManager);
    }

    @AfterEach
    void tearDown() {
        SecurityContextHolder.clearContext();
    }

    @Test
    void testCanRead() {
        when(securityService.canRead(any(RepoPath.class))).thenReturn(false);
        assertFalse(packageSecurityService.canRead("repo", "path"));
        verify(securityService, times(1)).canRead(eq(RepoPathFactory.create("repo", "path")));

        reset(securityService);

        when(securityService.canRead(any(RepoPath.class))).thenReturn(true);
        assertTrue(packageSecurityService.canRead("repo", "path"));
        verify(securityService, times(1)).canRead(eq(RepoPathFactory.create("repo", "path")));
    }

    @Test
    void testCanWrite() {
        when(securityService.canDeploy(any(RepoPath.class))).thenReturn(false);
        assertFalse(packageSecurityService.canWrite("repo", "path"));
        verify(securityService, times(1)).canDeploy(eq(RepoPathFactory.create("repo", "path")));

        reset(securityService);

        when(securityService.canDeploy(any(RepoPath.class))).thenReturn(true);
        assertTrue(packageSecurityService.canWrite("repo", "path"));
        verify(securityService, times(1)).canDeploy(eq(RepoPathFactory.create("repo", "path")));
    }

    @Test
    void testCanDelete() {
        when(securityService.canDelete(any(RepoPath.class))).thenReturn(false);
        assertFalse(packageSecurityService.canDelete("repo", "path"));
        verify(securityService, times(1)).canDelete(eq(RepoPathFactory.create("repo", "path")));

        reset(securityService);

        when(securityService.canDelete(any(RepoPath.class))).thenReturn(true);
        assertTrue(packageSecurityService.canDelete("repo", "path"));
        verify(securityService, times(1)).canDelete(eq(RepoPathFactory.create("repo", "path")));
    }

    @Test
    void testCanAnnotate() {
        when(securityService.canAnnotate(any(RepoPath.class))).thenReturn(false);
        assertFalse(packageSecurityService.canAnnotate("repo", "path"));
        verify(securityService, times(1)).canAnnotate(eq(RepoPathFactory.create("repo", "path")));

        reset(securityService);

        when(securityService.canAnnotate(any(RepoPath.class))).thenReturn(true);
        assertTrue(packageSecurityService.canAnnotate("repo", "path"));
        verify(securityService, times(1)).canAnnotate(eq(RepoPathFactory.create("repo", "path")));
    }

    @Test
    void testCanManage() {
        when(securityService.canManage(any(RepoPath.class))).thenReturn(false);
        assertFalse(packageSecurityService.canManage("repo", "path"));
        verify(securityService, times(1)).canManage(eq(RepoPathFactory.create("repo", "path")));

        reset(securityService);

        when(securityService.canManage(any(RepoPath.class))).thenReturn(true);
        assertTrue(packageSecurityService.canManage("repo", "path"));
        verify(securityService, times(1)).canManage(eq(RepoPathFactory.create("repo", "path")));
    }

    @Test
    void testCallAsSystem() throws Exception {
        Mockito.doAnswer(invocationOnMock -> ((Callable) invocationOnMock.getArgument(0)).call())
                .when(securityService).callAsSystem(any(Callable.class));
        Callable<Boolean> callable = () -> true;
        assertTrue(securityService.callAsSystem(callable));
        verify(securityService, times(1)).callAsSystem(eq(callable));
    }

    @Test
    void testGetCurrentUserName() {
        when(securityService.currentUsername()).thenReturn("user");
        assertEquals("user", packageSecurityService.getCurrentUserName());
        verify(securityService, times(1)).currentUsername();
    }

    @Test
    void testIsAuthenticated() {
        when(securityService.isAuthenticated()).thenReturn(true);
        assertTrue(packageSecurityService.isAuthenticated());
        verify(securityService, times(1)).isAuthenticated();
    }

    @Test
    void testLoginAndGenerateAccessTokenWithBasicAuth() throws PackageException {
        String user = "USER";
        String password = "PASSWORD";
        Long genericTokensExpirationTimeSecs = (long) (86400 * 30);
        Authentication authentication = Mockito.mock(Authentication.class);
        UsernamePasswordAuthenticationToken authenticationToken =
                new UsernamePasswordAuthenticationToken(user, password);
        mockGenericTokensExpirationTimeSecs(genericTokensExpirationTimeSecs);
        when(authenticationManager.authenticate(authenticationToken)).thenReturn(authentication);
        UserTokenSpec tokenSpec = UserTokenSpec.create(user)
                .expiresIn(genericTokensExpirationTimeSecs)
                .refreshable(false)
                .scope(Lists.newArrayList("member-of-groups:*"));
        CreatedTokenInfo createdTokenInfo = new CreatedTokenInfoImpl("a", "b", "c", "d", 1L, "f");
        when(accessService.createToken(any(UserTokenSpec.class), anyBoolean(), anyBoolean())).thenReturn(
                createdTokenInfo);
        packageSecurityService.loginAndGenerateAccessTokenWithBasicAuth(user, password);
        verify(accessService, times(1)).createToken(refEq(tokenSpec), anyBoolean(), anyBoolean());
    }

    private void mockGenericTokensExpirationTimeSecs(Long genericTokensExpirationTimeSecs) {
        ArtifactoryHome artifactoryHome =  Mockito.mock(ArtifactoryHome.class);
        ArtifactorySystemProperties artifactorySystemProperties = Mockito.mock(ArtifactorySystemProperties.class);
        ArtifactoryHome.bind(artifactoryHome);
        when(ArtifactoryHome.get().getArtifactoryProperties()).thenReturn(artifactorySystemProperties);
        when(artifactorySystemProperties.getLongProperty(any())).thenReturn(genericTokensExpirationTimeSecs);
    }
    @Test
    void testRevokeToken() throws PackageException {
        packageSecurityService.revokeToken("token");
        verify(accessService).revokeToken("token");
    }

    @Test
    void testRevokeTokenOnNonRevocableToken() throws PackageException {
        doThrow(new AccessClientException("")).when(accessService).revokeToken("NonRevocableToken");
        packageSecurityService.revokeToken("NonRevocableToken");
        verify(accessService).revokeToken("NonRevocableToken");
    }

    @Test
    void testRevokeTokenOnNonExistingToken()  {
        doThrow(new TokenNotFoundException("")).when(accessService).revokeToken("NonExistingToken");
        try {
            packageSecurityService.revokeToken("NonExistingToken");
            fail("Expected PackageException");
        } catch (PackageException e) {
            verify(accessService).revokeToken("NonExistingToken");
            assertEquals("Logout locally failed", e.getMessage());
            assertEquals(HttpStatus.SC_BAD_REQUEST, e.getStatus());
        }
    }

    @Test
    void testGetCurrentUserDetailsWhenUnAuthenticated() {
        try {
            packageSecurityService.getCurrentUserDetails();
            fail("Expected PackageException");
        } catch (PackageException e) {
            assertEquals("User is unauthenticated", e.getMessage());
        }
    }

    @Test
    void testGetCurrentUserDetails() throws PackageException {
        UserImpl userInfo = new UserImpl("user1");
        userInfo.setEmail("user1@jfrog.com");
        String password = "PASSWORD";
        userInfo.setPassword(new SaltedPassword(password, ""));
        SecurityContext securityContext = new SecurityContextImpl();
        Authentication authentication =
                new UsernamePasswordAuthenticationToken(new SimpleUser(userInfo), password, null);
        securityContext.setAuthentication(authentication);
        SecurityContextHolder.setContext(securityContext);
        when(userGroupStoreService.findUser("user1")).thenReturn(userInfo);
        when(securityService.createEncryptedPasswordIfNeeded(userInfo, password)).thenReturn(password);
        com.jfrog.ph.acl.model.SimpleUser resultUser = packageSecurityService.getCurrentUserDetails();
        assertNotNull(resultUser);
        assertEquals("user1", resultUser.getName());
        assertEquals(password, resultUser.getPassword());
        assertEquals("user1@jfrog.com", resultUser.getEmail());
    }

}