package org.artifactory.ph.virtual;

import com.jfrog.ph.acl.exception.PackageException;
import com.jfrog.ph.acl.model.PackageDownloadContext;
import com.jfrog.ph.acl.resource.PackageUnfoundedResource;
import com.jfrog.ph.acl.virtual.resolution.PackageRepo;
import com.jfrog.ph.acl.virtual.resolution.PackageVirtualController;
import com.jfrog.ph.acl.virtual.resolution.VirtualResolutionStrategyType;
import org.apache.http.HttpStatus;
import org.artifactory.common.resource.BaseResource;
import org.junit.jupiter.api.Assertions;
import org.testng.annotations.Test;

import java.util.List;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.ArgumentMatchers.refEq;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.when;

/**
 * Tests the Virtual repository visitor logic
 */
@Test
public class VirtualRepoTraversalResolverTest extends VirtualRepoTraversalBaseTest {

    /**
     * Test virtual repo resolution
     * 1. upload two artifacts into two different repositories that belongs to virtual repo
     * 2. try to download and expect to get first artifact
     */
    @Test
    public void testResolverExistTwoReturnFirst() throws PackageException {
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any(), eq(false));
        BaseResource result1 = resource(localRepoKey1, path, null);
        BaseResource result2 = resource(localRepoKey2, path, null);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey1, path))))
                .thenReturn(result1);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey2, path))))
                .thenReturn(result2);

        PackageVirtualController controller = controller(VirtualResolutionStrategyType.RESOLVER);
        BaseResource resource =
                virtualRepoTraversal.resolve(new PackageDownloadContext(virtualRepoKey, path), controller,
                        getResourceFunc(controller));
        Assertions.assertEquals(localRepoKey1, resource.getRepoKey());
        Assertions.assertEquals(path, resource.getPath());
    }

    /**
     * Test virtual repo resolution
     * 1. upload two different artifacts into two different repositories that belongs to virtual repo
     * 2. try to download one of them and expect to find one artifact
     */
    @Test
    public void testExist1ReturnReturnFirst() throws PackageException {
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any(), eq(false));
        BaseResource result1 = resource(localRepoKey1, path1, null);
        BaseResource result2 = resource(localRepoKey2, path2, null);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey1, path1))))
                .thenReturn(result1);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey2, path2))))
                .thenReturn(result2);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey1, path2))))
                .thenReturn(new PackageUnfoundedResource(localRepoKey1, path2, 404, ""));
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey2, path1))))
                .thenReturn(new PackageUnfoundedResource(localRepoKey2, path1, 404, ""));
        PackageVirtualController controller = controller(VirtualResolutionStrategyType.RESOLVER);
        BaseResource resource1 =
                virtualRepoTraversal.resolve(new PackageDownloadContext(virtualRepoKey, path1), controller,
                        getResourceFunc(controller));
        Assertions.assertEquals(localRepoKey1, resource1.getRepoKey());
        Assertions.assertEquals(path1, resource1.getPath());
        BaseResource resource2 =
                virtualRepoTraversal.resolve(new PackageDownloadContext(virtualRepoKey, path2), controller,
                        getResourceFunc(controller));
        Assertions.assertEquals(localRepoKey2, resource2.getRepoKey());
        Assertions.assertEquals(path2, resource2.getPath());
    }

    /**
     * Checks the virtual response code when both sub repositories return not found - the virtual repository should
     * return not found.
     */
    @Test
    public void testNotFound() throws PackageException {
        testExpectedError(HttpStatus.SC_NOT_FOUND, HttpStatus.SC_NOT_FOUND, HttpStatus.SC_NOT_FOUND);
    }

    /**
     * Checks the virtual response code when one of the repositories send not found and the other forbidden  - the
     * virtual repository should return forbidden.
     */
    @Test
    public void testForbiddenAndNotFound() throws PackageException {
        testExpectedError(HttpStatus.SC_NOT_FOUND, HttpStatus.SC_FORBIDDEN, HttpStatus.SC_FORBIDDEN);

        testExpectedError(HttpStatus.SC_FORBIDDEN, HttpStatus.SC_NOT_FOUND, HttpStatus.SC_FORBIDDEN);
    }

    /**
     * Checks the virtual response code when one of the repositories send unauthorized and the other not found  - the
     * virtual repository should return unauthorized.
     */
    @Test
    public void testUnAuthorizedAndNotFound() throws PackageException {
        testExpectedError(HttpStatus.SC_NOT_FOUND, HttpStatus.SC_UNAUTHORIZED,
                HttpStatus.SC_UNAUTHORIZED);

        testExpectedError(HttpStatus.SC_UNAUTHORIZED, HttpStatus.SC_NOT_FOUND,
                HttpStatus.SC_UNAUTHORIZED);
    }

    /**
     * Checks the virtual response code when one of the repositories send unauthorized and the other forbidden  - the
     * virtual repository should return unauthorized.
     */
    @Test
    public void testUnAuthorizedAndForbidden() throws PackageException {
        testExpectedError(HttpStatus.SC_UNAUTHORIZED, HttpStatus.SC_FORBIDDEN,
                HttpStatus.SC_UNAUTHORIZED);

        testExpectedError(HttpStatus.SC_FORBIDDEN, HttpStatus.SC_UNAUTHORIZED,
                HttpStatus.SC_UNAUTHORIZED);
    }

    private void testExpectedError(int firstResponseCode, int secondResponseCode,
            int virtualExpectedResponseCode)
            throws PackageException {
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any(), eq(false));
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey1, path1))))
                .thenReturn(new PackageUnfoundedResource(localRepoKey1, path1, firstResponseCode, ""));
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey2, path1))))
                .thenReturn(new PackageUnfoundedResource(localRepoKey2, path1, secondResponseCode, ""));
        PackageVirtualController controller = controller(VirtualResolutionStrategyType.RESOLVER);
        BaseResource resource =
                virtualRepoTraversal.resolve(new PackageDownloadContext(virtualRepoKey, path1), controller,
                        getResourceFunc(controller));
        Assertions.assertFalse(resource.isFound());
        Assertions.assertEquals(virtualExpectedResponseCode, ((PackageUnfoundedResource) resource).getStatusCode());
        Assertions.assertEquals(path1, resource.getPath());
    }
}
