package org.artifactory.projects.interceptor;

import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.FederatedRepoConfig;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.VirtualRepoConfig;
import org.artifactory.projects.ProjectsService;
import org.artifactory.repo.RepoConfigChange;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.access.client.AccessClientException;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.List;
import java.util.Map;

import static org.artifactory.repo.RepoDetailsType.LOCAL;
import static org.mockito.Mockito.anyString;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

public class ProjectRepoConfigurationInterceptorTest extends ArtifactoryHomeBoundTest {

    @Mock
    ProjectsService projectsService;

    @BeforeMethod
    public void setup() {
        MockitoAnnotations.openMocks(this);
    }

    @Test
    public void testOnBeforeSaveWhenNeedToRemoveFromLocalOnly() {
        ProjectRepoConfigurationInterceptor interceptor = new ProjectRepoConfigurationInterceptor(projectsService);
        when(projectsService.getProjectKeyOfRepository("repo")).thenReturn("prjx");

        interceptor.onBeforeSave(Map.of(LOCAL, List.of(new RepoConfigChange("repo",
                RepoConfigChange.ConfigAction.REPO_DELETE, null,
                LocalRepoConfig.create("repoKey", RepoType.Generic)))));

        verify(projectsService).removeRepositoryFromCurrentProject("repo");
    }

    @Test
    public void testOnBeforeSaveWhenNeedToRemoveFromFederatedOnly() {
        ProjectRepoConfigurationInterceptor interceptor = new ProjectRepoConfigurationInterceptor(projectsService);
        when(projectsService.getProjectKeyOfRepository("repo")).thenReturn("prjx");

        interceptor.onBeforeSave(Map.of(RepoDetailsType.FEDERATED, List.of(new RepoConfigChange("repo",
                RepoConfigChange.ConfigAction.REPO_DELETE, null,
                FederatedRepoConfig.create("repoKey", RepoType.Generic)))));

        verify(projectsService).removeRepositoryFromCurrentProject("repo");
    }

    @Test
    public void testOnBeforeSaveWhenNeedToRemoveFromLocalAndRemote() {
        ProjectRepoConfigurationInterceptor interceptor = new ProjectRepoConfigurationInterceptor(projectsService);
        when(projectsService.getProjectKeyOfRepository("deletelocal")).thenReturn("prjx");
        when(projectsService.getProjectKeyOfRepository("deleteremote")).thenReturn("prjx");

        interceptor.onBeforeSave(Map.of(LOCAL,
                List.of(new RepoConfigChange("deletelocal", RepoConfigChange.ConfigAction.REPO_DELETE, null,
                        LocalRepoConfig.create("repoKey", RepoType.Generic))),
                RepoDetailsType.REMOTE,
                List.of(new RepoConfigChange("deleteremote", RepoConfigChange.ConfigAction.REPO_DELETE, null,
                        RemoteRepoConfig.create("repoKey", RepoType.Generic)))));

        verify(projectsService).removeRepositoryFromCurrentProject("deletelocal");
        verify(projectsService).removeRepositoryFromCurrentProject("deleteremote");
    }

    @Test
    public void testOnBeforeSaveWhenNeedToRemoveFromLocalAndVirtual() {
        ProjectRepoConfigurationInterceptor interceptor = new ProjectRepoConfigurationInterceptor(projectsService);
        when(projectsService.getProjectKeyOfRepository("deletelocal")).thenReturn("prjx");
        when(projectsService.getProjectKeyOfRepository("deletevirtual")).thenReturn("prjx");

        interceptor.onBeforeSave(Map.of(LOCAL,
                List.of(new RepoConfigChange("deletelocal", RepoConfigChange.ConfigAction.REPO_DELETE, null,
                        LocalRepoConfig.create("repoKey", RepoType.Generic))),
                RepoDetailsType.VIRTUAL,
                List.of(new RepoConfigChange("deletevirtual", RepoConfigChange.ConfigAction.REPO_DELETE, null,
                        VirtualRepoConfig.create("repoKey", RepoType.Generic)))));

        verify(projectsService).removeRepositoryFromCurrentProject("deletelocal");
        verify(projectsService).removeRepositoryFromCurrentProject("deletevirtual");
    }


    @Test
    public void testOnBeforeSaveWhenNeedToRemoveFromRemoteAndVirtual() {
        ProjectRepoConfigurationInterceptor interceptor = new ProjectRepoConfigurationInterceptor(projectsService);
        when(projectsService.getProjectKeyOfRepository("deleteremote")).thenReturn("prjx");
        when(projectsService.getProjectKeyOfRepository("deletevirtual")).thenReturn("prjx");

        interceptor.onBeforeSave(Map.of(RepoDetailsType.REMOTE,
                List.of(new RepoConfigChange("deleteremote", RepoConfigChange.ConfigAction.REPO_DELETE, null,
                        RemoteRepoConfig.create("repoKey", RepoType.Generic))),
                RepoDetailsType.VIRTUAL,
                List.of(new RepoConfigChange("deletevirtual", RepoConfigChange.ConfigAction.REPO_DELETE, null,
                        VirtualRepoConfig.create("repoKey", RepoType.Generic)))));

        verify(projectsService).removeRepositoryFromCurrentProject("deleteremote");
        verify(projectsService).removeRepositoryFromCurrentProject("deletevirtual");
    }

    @Test
    public void testOnBeforeSaveWhenThereNoNeedToRemove() {
        ProjectRepoConfigurationInterceptor interceptor = new ProjectRepoConfigurationInterceptor(projectsService);

        interceptor.onBeforeSave(Map.of(LOCAL,
                List.of(new RepoConfigChange("repo", RepoConfigChange.ConfigAction.REPO_CREATE,
                        LocalRepoConfig.create("repoKey", RepoType.Generic), null))
        ));

        verify(projectsService, never()).getProjectKeyOfRepository(anyString());
        verify(projectsService, never()).removeRepositoryFromCurrentProject(anyString());
    }

    @Test
    public void testOnBeforeSaveWithNoChanges() {
        ProjectRepoConfigurationInterceptor interceptor = new ProjectRepoConfigurationInterceptor(projectsService);

        interceptor.onBeforeSave(Map.of());

        verify(projectsService, never()).getProjectKeyOfRepository(anyString());
        verify(projectsService, never()).removeRepositoryFromCurrentProject(anyString());
    }

    @Test
    public void testAccessClientExceptionFromAccess() {
        ProjectRepoConfigurationInterceptor interceptor = new ProjectRepoConfigurationInterceptor(projectsService);
        doThrow(new AccessClientException("Unable to delete")).when(projectsService)
                .removeRepositoryFromCurrentProject(anyString());

        interceptor.onBeforeSave(Map.of(LOCAL, List.of(new RepoConfigChange("repo",
                RepoConfigChange.ConfigAction.REPO_DELETE, null,
                LocalRepoConfig.create("repoKey", RepoType.Generic)))));

        verify(projectsService).removeRepositoryFromCurrentProject("repo");
    }

    @Test
    public void testNPEFromAccess() {
        ProjectRepoConfigurationInterceptor interceptor = new ProjectRepoConfigurationInterceptor(projectsService);

        doThrow(new NullPointerException("null")).when(projectsService)
                .removeRepositoryFromCurrentProject(anyString());

        interceptor.onBeforeSave(Map.of(LOCAL, List.of(new RepoConfigChange("repo",
                RepoConfigChange.ConfigAction.REPO_DELETE, null,
                LocalRepoConfig.create("repoKey", RepoType.Generic)))));

        verify(projectsService).removeRepositoryFromCurrentProject("repo");
    }

    @Test
    public void testConvertedRepo() {
        ProjectRepoConfigurationInterceptor interceptor = new ProjectRepoConfigurationInterceptor(projectsService);
        interceptor.onBeforeSave(Map.of(LOCAL, List.of(new RepoConfigChange("repo",
                RepoConfigChange.ConfigAction.REPO_UPDATE, FederatedRepoConfig.create("repoKey", RepoType.Generic),
                LocalRepoConfig.create("repoKey", RepoType.Generic)))));

        verify(projectsService, times(0)).removeRepositoryFromCurrentProject("repo");
    }
}