/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.repo.cache.expirable;

import com.google.common.collect.Maps;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.spring.ArtifactoryApplicationContext;
import org.mockito.Mockito;
import org.testng.annotations.Test;

import java.util.Map;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertTrue;

/**
 * @author Noam Y. Tenne
 */
public class CacheExpiryImplTest {

    @Test(expectedExceptions = NullPointerException.class)
    public void testInitWithNullContext() {
        new CacheExpiryImpl().init();
    }

    @Test
    public void testInitWithNullBeanName() {
        ArtifactoryApplicationContext context = Mockito.mock(ArtifactoryApplicationContext.class);
        when(context.beansForType(CacheExpiryChecker.class))
                .thenReturn(Maps.newHashMap());

        CacheExpiryImpl cacheExpiry = new CacheExpiryImpl();
        cacheExpiry.setApplicationContext(context);
        cacheExpiry.init();
        Mockito.verify(context, Mockito.times(1)).beansForType(CacheExpiryChecker.class);
    }

    @SuppressWarnings("DataFlowIssue")
    @Test
    public void testIsExpirableWithNullPath() {
        assertFalse(new CacheExpiryImpl().isExpirable(null, null, null), "Null path should never be expirable.");
        assertFalse(new CacheExpiryImpl().isLocalGenerated(null, null, null), "Null path should never be allowed.");
        assertFalse(new CacheExpiryImpl().isMetadata(null, null, null), "Null path should never be allowed.");
        assertFalse(new CacheExpiryImpl().shouldExcludePropertyFromMirroring(null, null),
                "Null path should never be allowed.");
    }

    @Test
    public void testIsExpirableWithValidPathAndNoExpirables() {
        assertFalse(new CacheExpiryImpl().isExpirable(null, null, "afdafasdf"),
                "Nothing should be expirable when no expiry criteria is set.");
        assertFalse(new CacheExpiryImpl().isLocalGenerated(RepoType.Alpine, "alpine", "afdafasdf"),
                "Nothing should be expirable when no expiry criteria is set.");
        assertFalse(new CacheExpiryImpl().isMetadata(RepoType.Alpine, "alpine", "afdafasdf"),
                "Nothing should be expirable when no expiry criteria is set.");
    }

    @Test
    public void testIsExpirable() {
        CacheExpiryChecker rootExpirable = mockChecker(new RepoType[] {RepoType.Maven}, false);
        CacheExpiryChecker acceptingExpirable = mockChecker(new RepoType[] {RepoType.Generic, RepoType.Maven}, true);

        Map<String, CacheExpiryChecker> expirableHashMap = Map.of(
                "root", rootExpirable,
                "accepting", acceptingExpirable
        );
        String path = "asdfasdf";
        String repoKey = "root";
        boolean expirable = testExpirable(expirableHashMap, repoKey, path);
        assertTrue(expirable, "Path should be expirable.");
        Mockito.verify(acceptingExpirable, times(1)).isExpirable(RepoType.Maven, repoKey, path);
    }

    @Test
    public void testIsNotExpirable() {
        CacheExpiryChecker rootExpirable = mockChecker(new RepoType[] {RepoType.Maven}, false);
        CacheExpiryChecker differentType = mockChecker(new RepoType[] {RepoType.Generic, RepoType.Gems}, true);

        Map<String, CacheExpiryChecker> expirableHashMap = Map.of(
                "root", rootExpirable,
                "accepting", differentType
        );
        String path = "asdfasdf";
        String repoKey = "root";
        boolean expirable = testExpirable(expirableHashMap, repoKey, path);
        assertFalse(expirable, "Path should not be expirable.");
        Mockito.verify(differentType, times(0)).isExpirable(RepoType.Maven, repoKey, path);
        Mockito.verify(rootExpirable, times(1)).isExpirable(RepoType.Maven, repoKey, path);
    }

    private static boolean testExpirable(Map<String, CacheExpiryChecker> expirableHashMap, String repoKey,
            String path) {
        ArtifactoryApplicationContext context = Mockito.mock(ArtifactoryApplicationContext.class);
        when(context.beansForType(CacheExpiryChecker.class)).thenReturn(expirableHashMap);

        CacheExpiryImpl cacheExpiry = new CacheExpiryImpl();
        cacheExpiry.setApplicationContext(context);

        cacheExpiry.init();
        Mockito.verify(context, Mockito.times(1)).beansForType(CacheExpiryChecker.class);
        return cacheExpiry.isExpirable(RepoType.Maven, repoKey, path);
    }

    private static CacheExpiryChecker mockChecker(RepoType[] types, boolean expirable) {
        CacheExpiryChecker rootExpirable = Mockito.mock(CacheExpiryChecker.class);
        when(rootExpirable.isExpirable(any(), any(), any())).thenReturn(expirable);
        when(rootExpirable.getTypes()).thenReturn(types);
        return rootExpirable;
    }

    private static CacheExpiryChecker mockGeneratedChecker(RepoType[] types, boolean localGenerated) {
        CacheExpiryChecker rootExpirable = Mockito.mock(CacheExpiryChecker.class);
        when(rootExpirable.isLocalGenerated(any(), any(), any())).thenReturn(localGenerated);
        when(rootExpirable.getTypes()).thenReturn(types);
        return rootExpirable;
    }

    private static CacheExpiryChecker mockMetadataChecker(RepoType[] types) {
        CacheExpiryChecker rootExpirable = Mockito.mock(CacheExpiryChecker.class);
        when(rootExpirable.isMetadata(any(), any(), any())).thenReturn(true);
        when(rootExpirable.getTypes()).thenReturn(types);
        return rootExpirable;
    }

    @Test
    public void testIsLocalGenerated() {
        CacheExpiryChecker rootExpirable = mockGeneratedChecker(new RepoType[] {RepoType.Maven}, false);
        CacheExpiryChecker acceptingExpirable =
                mockGeneratedChecker(new RepoType[] {RepoType.Generic, RepoType.Maven}, true);

        Map<String, CacheExpiryChecker> expirableHashMap = Map.of(
                "root", rootExpirable,
                "accepting", acceptingExpirable
        );
        String path = "asdfasdf";
        String repoKey = "root";
        boolean expirable = testIsLocalGenerated(expirableHashMap, repoKey, path);
        assertTrue(expirable, "Path should be local generated.");
        Mockito.verify(acceptingExpirable, times(1)).isLocalGenerated(RepoType.Maven, repoKey, path);
    }

    @Test
    public void testIsMetadata() {
        CacheExpiryChecker rootExpirable = mockGeneratedChecker(new RepoType[] {RepoType.Maven}, false);
        CacheExpiryChecker acceptingExpirable =
                mockMetadataChecker(new RepoType[] {RepoType.Generic, RepoType.Maven});

        Map<String, CacheExpiryChecker> expirableHashMap = Map.of(
                "root", rootExpirable,
                "accepting", acceptingExpirable
        );
        String path = "asdfasdf";
        String repoKey = "root";
        boolean expirable = testIsMetadata(expirableHashMap, repoKey, path);
        assertTrue(expirable, "Path should be local generated.");
        Mockito.verify(acceptingExpirable, times(1)).isMetadata(RepoType.Maven, repoKey, path);
    }

    @Test
    public void testIsNotLocalGenerated() {
        CacheExpiryChecker rootExpirable = mockGeneratedChecker(new RepoType[] {RepoType.Maven}, false);
        CacheExpiryChecker differentChecker =
                mockGeneratedChecker(new RepoType[] {RepoType.Generic, RepoType.Gems}, true);

        Map<String, CacheExpiryChecker> expirableHashMap = Map.of(
                "root", rootExpirable,
                "accepting", differentChecker
        );
        String path = "asdfasdf";
        String repoKey = "root";
        boolean expirable = testIsLocalGenerated(expirableHashMap, repoKey, path);
        assertFalse(expirable, "Path should not be local generated.");
        Mockito.verify(differentChecker, times(0)).isLocalGenerated(RepoType.Maven, repoKey, path);
        Mockito.verify(rootExpirable, times(1)).isLocalGenerated(RepoType.Maven, repoKey, path);
    }

    private static boolean testIsLocalGenerated(Map<String, CacheExpiryChecker> expirableHashMap, String repoKey,
            String path) {
        ArtifactoryApplicationContext context = Mockito.mock(ArtifactoryApplicationContext.class);
        when(context.beansForType(CacheExpiryChecker.class)).thenReturn(expirableHashMap);

        CacheExpiryImpl cacheExpiry = new CacheExpiryImpl();
        cacheExpiry.setApplicationContext(context);

        cacheExpiry.init();
        Mockito.verify(context, Mockito.times(1)).beansForType(CacheExpiryChecker.class);
        return cacheExpiry.isLocalGenerated(RepoType.Maven, repoKey, path);
    }

    private static boolean testIsMetadata(Map<String, CacheExpiryChecker> expirableHashMap, String repoKey,
            String path) {
        ArtifactoryApplicationContext context = Mockito.mock(ArtifactoryApplicationContext.class);
        when(context.beansForType(CacheExpiryChecker.class)).thenReturn(expirableHashMap);

        CacheExpiryImpl cacheExpiry = new CacheExpiryImpl();
        cacheExpiry.setApplicationContext(context);

        cacheExpiry.init();
        Mockito.verify(context, Mockito.times(1)).beansForType(CacheExpiryChecker.class);
        return cacheExpiry.isMetadata(RepoType.Maven, repoKey, path);
    }
}