package org.artifactory.repo.interceptor;

import lombok.NonNull;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.maven.MavenMetadataService;
import org.artifactory.api.maven.MavenMetadataWorkItem;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.MutableStatusHolder;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.descriptor.repo.SnapshotVersionBehavior;
import org.artifactory.fs.FileInfo;
import org.artifactory.md.Properties;
import org.artifactory.metadata.service.indexer.MetadataServiceIndexHandler;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.model.xstream.fs.FileInfoImpl;
import org.artifactory.model.xstream.fs.PropertiesImpl;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.db.DbLocalRepo;
import org.artifactory.repo.db.DbStoringRepoMixin;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.sapi.fs.VfsItem;
import org.artifactory.sapi.interceptor.context.DeleteContext;
import org.artifactory.sapi.interceptor.context.InterceptorCreateContext;
import org.artifactory.storage.db.fs.model.DbMutableFile;
import org.artifactory.storage.fs.MutableVfsFile;
import org.artifactory.storage.fs.repo.StoringRepo;
import org.artifactory.storage.fs.service.PropertiesService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.argThat;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;

/**
 * @author Yevdo Abramov
 * Created on 01/10/2020
 */
public class MavenMetadataCalculationInterceptorTest extends ArtifactoryHomeBoundTest {

    @Mock
    private MavenMetadataService mavenMetadataService;
    @Mock
    private InternalRepositoryService repoService;
    @Mock
    private CentralConfigService centralConfigService;
    @Mock
    private ArtifactoryContext artifactoryContext;
    @Mock
    private PropertiesService propertiesService;
    @Mock
    private MetadataServiceIndexHandler metadataServiceIndexHandler;

    private MavenMetadataCalculationInterceptor interceptor;
    private AutoCloseable autoCloseable;

    @BeforeMethod
    public void setUp() {
        autoCloseable = MockitoAnnotations.openMocks(this);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        interceptor = new MavenMetadataCalculationInterceptor();
        interceptor.setMetadataServiceIndexHandler(metadataServiceIndexHandler);
        interceptor.setMavenMetadataService(mavenMetadataService);
        interceptor.setRepoService(repoService);
        doReturn(propertiesService).when(artifactoryContext).beanForType(PropertiesService.class);
    }

    @AfterMethod
    public void tearDown() throws Exception {
        ArtifactoryContextThreadBinder.unbind();
        verifyNoMoreInteractions(mavenMetadataService, repoService, propertiesService, metadataServiceIndexHandler);
        autoCloseable.close();
    }

    @Test(testName = "test afterCreate with mavenPlugin indicator true or false")
    public void testAfterCreateForMavenPluginIndicator() {
        // given
        MutableStatusHolder statusHolder = mock(MutableStatusHolder.class);
        InterceptorCreateContext context = mock(InterceptorCreateContext.class);
        System.setProperty(ConstantValues.mvnLockMetadata.getPropertyName(), "false");
        ArtifactoryHome.get().initArtifactorySystemProperties();
        VfsItem fsFile =
                mockFsFile("org/codehaus/mojo/maven-buildnumber-plugin/0.9.4/maven-buildnumber-plugin-0.9.4.pom");

        Properties properties = new PropertiesImpl();
        when(propertiesService.getProperties(fsFile.getInfo())).thenReturn(properties);

        // when
        interceptor.afterCreate(fsFile, statusHolder, context);

        // then
        verify(repoService).localOrFederatedRepositoryByKey(fsFile.getRepoKey());
        verify(propertiesService, times(2)).getProperties(fsFile.getInfo());
        verify(mavenMetadataService).calculateMavenMetadataAsync(argThat(workItem -> !workItem.isMavenPlugin()));
        verify(metadataServiceIndexHandler).indexPackageAsync(fsFile.getInfo(), properties);

        // given
        properties.put(org.artifactory.api.properties.PropertiesService.MAVEN_PLUGIN_PROPERTY_NAME, "true");
        Mockito.clearInvocations(repoService, mavenMetadataService, propertiesService, metadataServiceIndexHandler);

        // when
        interceptor.afterCreate(fsFile, statusHolder, context);

        // then
        verify(repoService).localOrFederatedRepositoryByKey(fsFile.getRepoKey());
        verify(propertiesService, times(2)).getProperties(fsFile.getInfo());
        verify(mavenMetadataService).calculateMavenMetadataAsync(argThat(MavenMetadataWorkItem::isMavenPlugin));
        verify(metadataServiceIndexHandler).indexPackageAsync(fsFile.getInfo(), properties);

    }

    @Test(testName = "test afterDelete with mavenPlugin indicator true or false")
    public void testAfterDeleteForMavenPluginIndicator() {
        // given
        MutableStatusHolder statusHolder = mock(MutableStatusHolder.class);
        VfsItem fsFile =
                mockFsFile("org/codehaus/mojo/maven-buildnumber-plugin/0.9.4/maven-buildnumber-plugin-0.9.4.pom");

        DeleteContext context = new DeleteContext(fsFile.getRepoPath()).triggerMetadataCalculation(true);
        Properties properties = new PropertiesImpl();
        when(propertiesService.getProperties(fsFile.getInfo())).thenReturn(properties);

        // when
        interceptor.afterDelete(fsFile, statusHolder, context);

        // then
        verify(repoService, times(2)).localOrFederatedRepositoryByKey(fsFile.getRepoKey());
        verify(propertiesService, times(2)).getProperties(fsFile.getInfo());
        verify(mavenMetadataService, times(2))
                .calculateMavenMetadataAsync(argThat(workItem -> !workItem.isMavenPlugin()));

        // given
        properties.put(org.artifactory.api.properties.PropertiesService.MAVEN_PLUGIN_PROPERTY_NAME, "true");
        Mockito.clearInvocations(repoService, mavenMetadataService, propertiesService);

        // when
        interceptor.afterDelete(fsFile, statusHolder, context);

        // then
        verify(repoService, times(2)).localOrFederatedRepositoryByKey(fsFile.getRepoKey());
        verify(propertiesService, times(2)).getProperties(fsFile.getInfo());
        verify(mavenMetadataService, times(2))
                .calculateMavenMetadataAsync(argThat(MavenMetadataWorkItem::isMavenPlugin));
    }

    @Test
    public void testAfterDeleteMavenDeployer() {
        MutableStatusHolder statusHolder = mock(MutableStatusHolder.class);
        LocalRepo localRepo = mock(DbLocalRepo.class);
        LocalRepoDescriptor descriptor = mock(LocalRepoDescriptor.class);
        Properties properties = new PropertiesImpl();

        VfsItem fsFile = mockFsFile("org/codehaus/mojo/maven-buildnumber-plugin/" +
                "7.x-SNAPSHOT/jfrog-control-plane-bin-7.x-20210705.101611-3.jar");

        DeleteContext context = new DeleteContext(fsFile.getRepoPath()).triggerMetadataCalculation(true);
        when(repoService.localOrFederatedRepositoryByKey(fsFile.getRepoKey())).thenReturn(localRepo);
        when(localRepo.getMavenSnapshotVersionBehavior()).thenReturn(SnapshotVersionBehavior.DEPLOYER);
        when(localRepo.getDescriptor()).thenReturn(descriptor);
        when(fsFile.getProperties()).thenReturn(properties);

        interceptor.afterDelete(fsFile, statusHolder, context);
        verify(repoService, times(1)).localOrFederatedRepositoryByKey(fsFile.getRepoKey());
        verify(mavenMetadataService, times(0))
                .calculateMavenMetadataAsync(argThat(MavenMetadataWorkItem::isMavenPlugin));
    }

    @Test
    public void testAfterDeleteMavenAsUsual() {
        MutableStatusHolder statusHolder = mock(MutableStatusHolder.class);
        LocalRepo localRepo = mock(DbLocalRepo.class);
        LocalRepoDescriptor descriptor = mock(LocalRepoDescriptor.class);
        Properties properties = new PropertiesImpl();

        VfsItem fsFile = mockFsFile("org/codehaus/mojo/maven-buildnumber-plugin/" +
                "7.x-SNAPSHOT/jfrog-control-plane-bin-7.x-20210705.101611-3.jar");

        DeleteContext context = new DeleteContext(fsFile.getRepoPath()).triggerMetadataCalculation(true);
        when(repoService.localOrFederatedRepositoryByKey(fsFile.getRepoKey())).thenReturn(localRepo);
        when(localRepo.getMavenSnapshotVersionBehavior()).thenReturn(SnapshotVersionBehavior.UNIQUE);
        when(localRepo.getDescriptor()).thenReturn(descriptor);
        when(fsFile.getProperties()).thenReturn(properties);

        interceptor.afterDelete(fsFile, statusHolder, context);
        verify(repoService, times(2)).localOrFederatedRepositoryByKey(fsFile.getRepoKey());
        verify(propertiesService, times(1)).getProperties(fsFile.getInfo());
        verify(mavenMetadataService, times(1)).calculateMavenMetadataAsync(any());
    }

    @NonNull
    private MutableVfsFile mockFsFile(@NonNull String path) {
        LocalRepoConfig localRepoConfig = LocalRepoConfig.create("local_repo", RepoType.Generic);
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();// mock(LocalRepoDescriptor.class);
        LocalRepo localRepo = new DbLocalRepo(localRepoDescriptor, localRepoConfig, repoService,
                null, centralConfigService);
        localRepoDescriptor.setType(RepoType.Maven);
        StoringRepo storingRepo = new DbStoringRepoMixin(localRepoDescriptor, null, null);
        FileInfo fileInfo = new FileInfoImpl(new RepoPathImpl("local_repo", path, false));
        MutableVfsFile file = new DbMutableFile(storingRepo, 1, fileInfo);
        when(repoService.localOrFederatedRepositoryByKey("local_repo")).thenReturn(localRepo);
        return file;
    }
}
