package org.artifactory.repo.interceptor.config;

import com.google.common.collect.Lists;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.config.RepositoryConfigService;
import org.artifactory.config.ConfigurationException;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.RepoLayout;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.VirtualRepoConfig;
import org.artifactory.repo.RepoConfigChange;
import org.artifactory.repo.RepoDetailsType;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_CREATE;
import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_DELETE;
import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_UPDATE;
import static org.artifactory.repo.RepoDetailsType.LOCAL;
import static org.artifactory.repo.RepoDetailsType.REMOTE;
import static org.artifactory.repo.RepoDetailsType.VIRTUAL;
import static org.artifactory.repo.config.RepoConfigDefaultValues.DEFAULT_REPO_LAYOUT;
import static org.artifactory.repo.system.SystemRepos.BUILDINFO_REPO_KEY;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertNull;

/**
 * @author Tamar Rosen
 */
public class BuildInfoRepoConfigInterceptorTest {
    private BuildInfoRepoConfigInterceptor interceptor;
    private RepositoryConfigService repositoryConfigService;
    private CentralConfigService centralConfigService;

    @BeforeMethod
    public void setup() {
        repositoryConfigService = mock(RepositoryConfigService.class);
        centralConfigService = mock(CentralConfigService.class);
        interceptor = new BuildInfoRepoConfigInterceptor();
        interceptor.setRepositoryConfigService(repositoryConfigService);
        interceptor.setCentralConfigService(centralConfigService);
        when(centralConfigService.getDescriptor()).thenReturn(mock(CentralConfigDescriptor.class));
    }

    @Test(expectedExceptions = ConfigurationException.class)
    public void testNoMoreThanOneBuildInfoRepo() {
        LocalRepoConfig localRepoConfig = LocalRepoConfig.create(BUILDINFO_REPO_KEY, RepoType.BuildInfo);
        interceptor.onBeforeSave(
                Map.of(LOCAL, List.of(new RepoConfigChange(BUILDINFO_REPO_KEY, REPO_CREATE, localRepoConfig, null),
                        new RepoConfigChange(BUILDINFO_REPO_KEY, REPO_CREATE, localRepoConfig, null))));
    }

    @Test(expectedExceptions = ConfigurationException.class)
    public void testNoBuildInfoInRemote() {
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create(BUILDINFO_REPO_KEY, RepoType.BuildInfo);
        interceptor.onBeforeSave(
                Map.of(REMOTE, List.of(new RepoConfigChange(BUILDINFO_REPO_KEY, REPO_CREATE, remoteRepoConfig, null))));
    }

    @Test(expectedExceptions = ConfigurationException.class)
    public void testNoBuildInfoInVirtual() {
        VirtualRepoConfig virtualRepoConfig = VirtualRepoConfig.create(BUILDINFO_REPO_KEY, RepoType.BuildInfo);
        interceptor.onBeforeSave(
                Map.of(VIRTUAL,
                        List.of(new RepoConfigChange(BUILDINFO_REPO_KEY, REPO_CREATE, virtualRepoConfig, null))));
    }

    @Test
    public void testCreateDefaultBuildInfoRepoUponDeletion() {
        LocalRepoConfig oldBuildInfoRepoConfig = LocalRepoConfig.create("repoKey", RepoType.BuildInfo);
        when(repositoryConfigService.findRepositoriesByPackageType(RepoType.BuildInfo))
                .thenReturn(List.of(oldBuildInfoRepoConfig));
        Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChanges =
                new HashMap<>(Map.of(LOCAL, Lists.newArrayList(
                        new RepoConfigChange(BUILDINFO_REPO_KEY, REPO_DELETE, null, oldBuildInfoRepoConfig))));
        interceptor.onBeforeSave(repoConfigChanges);
        List<RepoConfigChange> localRepoConfigChanges = repoConfigChanges.get(LOCAL);
        assertEquals(localRepoConfigChanges.size(), 1);
        assertEquals(localRepoConfigChanges.get(0).getConfigAction(), REPO_UPDATE);
        assertEquals(localRepoConfigChanges.get(0).getRepoKey(), BUILDINFO_REPO_KEY);
    }

    @Test(expectedExceptions = ConfigurationException.class)
    public void testFailToCreateBuildInfoRepoIfFallbackNamesAreTaken() {
        // given - no build info repo exists in storage
        when(repositoryConfigService.findRepositoriesByPackageType(RepoType.BuildInfo)).thenReturn(List.of());
        // given - all fallback build info repo names are already taken
        when(repositoryConfigService.exists(BUILDINFO_REPO_KEY)).thenReturn(true);
        when(repositoryConfigService.exists(BUILDINFO_REPO_KEY + "-1")).thenReturn(true);
        when(repositoryConfigService.exists(BUILDINFO_REPO_KEY + "-2")).thenReturn(true);
        when(repositoryConfigService.exists(BUILDINFO_REPO_KEY + "-3")).thenReturn(true);
        when(repositoryConfigService.exists(BUILDINFO_REPO_KEY + "-4")).thenReturn(true);
        when(repositoryConfigService.exists(BUILDINFO_REPO_KEY + "-5")).thenReturn(true);
        Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChanges = new HashMap<>();
        interceptor.onBeforeSave(repoConfigChanges);
    }

    @Test
    public void testCreateBuildInfoRepoIfDoesNotExist() {
        // given - no build info repo exists in storage
        when(repositoryConfigService.findRepositoriesByPackageType(RepoType.BuildInfo)).thenReturn(List.of());
        // given - build info repo name is already taken
        when(repositoryConfigService.exists(BUILDINFO_REPO_KEY)).thenReturn(true);
        when(repositoryConfigService.exists(BUILDINFO_REPO_KEY + "-1")).thenReturn(false);
        Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChanges = new HashMap<>();
        interceptor.onBeforeSave(repoConfigChanges);
        assertEquals(1, repoConfigChanges.getOrDefault(LOCAL, List.of()).size());
        RepoConfigChange buildInfoCreation = repoConfigChanges.get(LOCAL).get(0);
        assertEquals(buildInfoCreation.getConfigAction(), REPO_CREATE);
        assertEquals(buildInfoCreation.getNewRepoConfig().getKey(), BUILDINFO_REPO_KEY + "-" + 1);
        assertEquals(buildInfoCreation.getNewRepoConfig().getPackageType(), RepoType.BuildInfo);
        assertNull(buildInfoCreation.getNewRepoConfig().getBaseConfig().getRepoLayoutRef());
    }

    @Test
    public void testCreateBuildInfoRepoIfDoesNotExistWithRepoLayoutRef() {
        CentralConfigDescriptor centralConfigDescriptor = mock(CentralConfigDescriptor.class);
        when(centralConfigDescriptor.getRepoLayout(DEFAULT_REPO_LAYOUT)).thenReturn(mock(RepoLayout.class));
        when(centralConfigService.getDescriptor()).thenReturn(centralConfigDescriptor);

        // given - no build info repo exists in storage
        when(repositoryConfigService.findRepositoriesByPackageType(RepoType.BuildInfo)).thenReturn(List.of());
        // given - build info repo name is already taken
        when(repositoryConfigService.exists(BUILDINFO_REPO_KEY)).thenReturn(true);
        when(repositoryConfigService.exists(BUILDINFO_REPO_KEY + "-1")).thenReturn(false);
        Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChanges = new HashMap<>();
        interceptor.onBeforeSave(repoConfigChanges);
        assertEquals(1, repoConfigChanges.getOrDefault(LOCAL, List.of()).size());
        RepoConfigChange buildInfoCreation = repoConfigChanges.get(LOCAL).get(0);
        assertEquals(buildInfoCreation.getConfigAction(), REPO_CREATE);
        assertEquals(buildInfoCreation.getNewRepoConfig().getKey(), BUILDINFO_REPO_KEY + "-" + 1);
        assertEquals(buildInfoCreation.getNewRepoConfig().getPackageType(), RepoType.BuildInfo);
        assertEquals(buildInfoCreation.getNewRepoConfig().getBaseConfig().getRepoLayoutRef(), DEFAULT_REPO_LAYOUT);
    }
}