package org.artifactory.repo.migration.mappers;

import com.google.gson.JsonParser;
import org.apache.commons.io.IOUtils;
import org.artifactory.descriptor.repo.ChecksumPolicyType;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoChecksumPolicyType;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.descriptor.repo.SnapshotVersionBehavior;
import org.artifactory.descriptor.repo.TrashRepoDescriptor;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.descriptor.repo.releasebundle.ReleaseBundlesRepoDescriptor;
import org.artifactory.repo.converter.RepositoriesConfigSubConverter;
import org.artifactory.storage.db.repo.entity.DbRepositoryConfig;
import org.jfrog.common.ResourceUtils;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static org.artifactory.repo.converter.RepositoryConfigBaseConverter.REPO_TYPE_CONFIG_KEY;
import static org.mockito.Mockito.mock;
import static org.testng.Assert.assertEquals;

public class RepoConfigMapperHolderTest {

    @Test(dataProvider = "mappersRegistrationDataProvider")
    public void testMapWithRegisteredClass(Class<RepoDescriptor> repoDescriptorClass, RepoDescriptor repoDescriptor) {
        RepoConfigMapperHolder repoConfigMapperHolder = new RepoConfigMapperHolder(Set.of());
        List<DbRepositoryConfig> dbRepos =
                repoConfigMapperHolder.map(repoDescriptorClass, Map.of("repoKey", repoDescriptor));
        assertEquals(dbRepos.size(), 1);
    }

    @Test(expectedExceptions = IllegalArgumentException.class,
            expectedExceptionsMessageRegExp = "Repository class .* does not have mapper")
    public void testMapWithoutRegisteredClass() {
        RepoConfigMapperHolder repoConfigMapperHolder = new RepoConfigMapperHolder(Set.of());
        repoConfigMapperHolder.map(TrashRepoDescriptor.class, Map.of("trash", mock(TrashRepoDescriptor.class)));
    }

    @Test
    public void testConvertMapOfFieldsToDbEntityWithoutSubConverters() throws IOException {
        LocalRepoDescriptor localRepoDescriptor = createLocalRepoDescriptor();
        RepoConfigMapperHolder repoConfigMapperHolder = new RepoConfigMapperHolder(Set.of());
        List<DbRepositoryConfig> dbRepositoryConfigs = repoConfigMapperHolder.map(LocalRepoDescriptor.class,
                Map.of(localRepoDescriptor.getKey(), localRepoDescriptor));
        assertEquals(dbRepositoryConfigs.size(), 1);

        DbRepositoryConfig dbRepositoryConfig = dbRepositoryConfigs.get(0);
        String infoBlobAsString = new String(dbRepositoryConfig.getRepositoryInfoBlob());
        String expectedInfoBlobAsString = IOUtils.toString(
                ResourceUtils.getResource("/org/artifactory/repo/migration/local-repo-after-migration.json"),
                StandardCharsets.UTF_8);

        assertEquals(JsonParser.parseString(infoBlobAsString), JsonParser.parseString(expectedInfoBlobAsString));
    }

    @Test
    public void testConvertMapOfFieldsToDbEntityWithSubConverters() throws IOException {
        LocalRepoDescriptor localRepoDescriptor = createLocalRepoDescriptor();
        RepoConfigMapperHolder repoConfigMapperHolder =
                new RepoConfigMapperHolder(Set.of(new TestRepoTypeConfigRepositoriesConfigSubConverter()));
        List<DbRepositoryConfig> dbRepositoryConfigs = repoConfigMapperHolder.map(LocalRepoDescriptor.class,
                Map.of(localRepoDescriptor.getKey(), localRepoDescriptor));
        assertEquals(dbRepositoryConfigs.size(), 1);

        DbRepositoryConfig dbRepositoryConfig = dbRepositoryConfigs.get(0);
        String infoBlobAsString = new String(dbRepositoryConfig.getRepositoryInfoBlob());
        String expectedInfoBlobAsString = IOUtils.toString(
                ResourceUtils.getResource("/org/artifactory/repo/migration/local-repo-after-converter.json"),
                StandardCharsets.UTF_8);

        assertEquals(JsonParser.parseString(infoBlobAsString), JsonParser.parseString(expectedInfoBlobAsString));
    }

    @DataProvider
    private static Object[][] mappersRegistrationDataProvider() {
        LocalRepoDescriptor localRepoDescriptor = createLocalRepoDescriptor();

        VirtualRepoDescriptor virtualRepoDescriptor = new VirtualRepoDescriptor();
        virtualRepoDescriptor.setType(RepoType.NuGet);
        virtualRepoDescriptor.setKey("virtual11");

        HttpRepoDescriptor remoteRepoDescriptor = new HttpRepoDescriptor();
        remoteRepoDescriptor.setType(RepoType.Debian);
        remoteRepoDescriptor.setChecksumPolicyType(ChecksumPolicyType.PASS_THRU);
        remoteRepoDescriptor.setKey("remote1");

        FederatedRepoDescriptor fedRepoDescriptor = new FederatedRepoDescriptor();
        fedRepoDescriptor.setType(RepoType.Vagrant);
        fedRepoDescriptor.setSnapshotVersionBehavior(SnapshotVersionBehavior.UNIQUE);
        fedRepoDescriptor.setChecksumPolicyType(LocalRepoChecksumPolicyType.CLIENT);
        fedRepoDescriptor.setKey("fed1");

        ReleaseBundlesRepoDescriptor releaseBundlesRepoDescriptor = new ReleaseBundlesRepoDescriptor();
        releaseBundlesRepoDescriptor.setType(RepoType.SBT);
        releaseBundlesRepoDescriptor.setSnapshotVersionBehavior(SnapshotVersionBehavior.UNIQUE);
        releaseBundlesRepoDescriptor.setChecksumPolicyType(LocalRepoChecksumPolicyType.CLIENT);
        releaseBundlesRepoDescriptor.setKey("bundle1");

        return new Object[][] {
                {LocalRepoDescriptor.class, localRepoDescriptor},
                {RemoteRepoDescriptor.class, remoteRepoDescriptor},
                {VirtualRepoDescriptor.class, virtualRepoDescriptor},
                {FederatedRepoDescriptor.class, fedRepoDescriptor},
                {ReleaseBundlesRepoDescriptor.class, releaseBundlesRepoDescriptor}
        };
    }

    private static LocalRepoDescriptor createLocalRepoDescriptor() {
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setSnapshotVersionBehavior(SnapshotVersionBehavior.UNIQUE);
        localRepoDescriptor.setChecksumPolicyType(LocalRepoChecksumPolicyType.CLIENT);
        localRepoDescriptor.setType(RepoType.Docker);
        localRepoDescriptor.setKey("local1");
        return localRepoDescriptor;
    }

    private class TestRepoTypeConfigRepositoriesConfigSubConverter implements RepositoriesConfigSubConverter {

        @Override
        public boolean isInterested(String key) {
            return key.equals("repoTypeConfig");
        }

        @Override
        public void convert(Map<String, Object> toConvert) {
            Map<String, Object> repoTypeConfig = (Map<String, Object>) toConvert.get(REPO_TYPE_CONFIG_KEY);
            repoTypeConfig.put("snapshotVersionBehavior", "TEST_VALUE_THAT_CONVERTER_UPDATED");
        }
    }
}