package org.artifactory.search.archive;

import org.apache.commons.lang.StringUtils;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.config.CentralConfigKey;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.schedule.Task;
import org.artifactory.schedule.TaskService;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.common.concurrency.ConflictGuard;
import org.jfrog.common.concurrency.JvmConflictGuard;
import org.jfrog.common.config.diff.DataDiff;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.List;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyBoolean;

/**
 * @author Shay Bagants
 */
@Test
public class ArchiveIndexerImplTest extends ArtifactoryHomeBoundTest {

    private ArchiveIndexerImpl archiveIndexer = Mockito.spy(new ArchiveIndexerImpl());

    @Mock
    private TaskService taskService;

    @Mock
    private InternalArtifactoryContext context;

    @Mock
    private CentralConfigDescriptor descriptor;

    @BeforeMethod
    private void setup() {
        MockitoAnnotations.initMocks(this);
        CentralConfigService configService = Mockito.mock(CentralConfigService.class);
        Mockito.when(configService.getDescriptor()).thenReturn(descriptor);
        Mockito.when(context.getCentralConfig()).thenReturn(configService);
        archiveIndexer.taskService = taskService;
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterMethod
    private void cleanup() {
        ArtifactoryContextThreadBinder.unbind();
    }


    @Test(dataProvider = "shouldStopProvider")
    public void shouldStop(ConflictGuard conflictGuard, boolean isReady, boolean pauseOrBreak, boolean isEnabled,
            boolean expectedToStop) {
        Mockito.when(descriptor.isArchiveIndexEnabled()).thenReturn(isEnabled);
        archiveIndexer.conflictGuard = conflictGuard;
        Mockito.when(context.isReady()).thenReturn(isReady);
        Mockito.when(taskService.pauseOrBreak()).thenReturn(pauseOrBreak);
        boolean shouldStop = archiveIndexer.shouldStop();
        Assert.assertEquals(shouldStop, expectedToStop);
    }

    @Test(dataProvider = "initProvider")
    public void testInit(boolean isEnabled, int numOfInvocations) {
        Mockito.when(descriptor.isArchiveIndexEnabled()).thenReturn(isEnabled);
        Mockito.doNothing().when(archiveIndexer).startTask();
        archiveIndexer.init();
        Mockito.verify(archiveIndexer, Mockito.times(numOfInvocations)).startTask();
    }

    @Test
    public void testStartTaskNotEmpty() {
        List<Task> taskList = new ArrayList<>();
        Task task = Mockito.mock(Task.class);
        taskList.add(task);
        Mockito.when(taskService.getActiveTasks(any())).thenReturn(taskList);
        Mockito.doNothing().when(archiveIndexer).startTask();
        archiveIndexer.startTask();
        Mockito.verify(taskService, Mockito.times(0)).startTask(any(), anyBoolean());
    }

    @Test
    public void testStartTaskEmpty() {
        List<Task> taskList = new ArrayList<>();
        Mockito.when(taskService.getActiveTasks(any())).thenReturn(taskList);
        Mockito.when(taskService.startTask(any(), anyBoolean())).thenReturn("");
        archiveIndexer.startTask();
        Mockito.verify(taskService, Mockito.times(1)).startTask(any(), anyBoolean());
    }

    @Test(dataProvider = "reloadProvider")
    public void testReload(String taskToken, boolean newValue, int startTaskInvocations, int stopTaskInvocations) {
        CentralConfigDescriptor oldDescriptor = Mockito.mock(CentralConfigDescriptor.class);
        archiveIndexer.taskToken = taskToken;
        List<DataDiff<?>> configDiff = new ArrayList<>();
        DataDiff dataDiff = new DataDiff(CentralConfigKey.archiveIndexEnabled.toString(), !newValue, newValue);
        configDiff.add(dataDiff);
        Mockito.when(taskService.startTask(any(), anyBoolean())).thenReturn("token");
        Mockito.doNothing().when(taskService).stopTask(any(), anyBoolean());
        archiveIndexer.reload(oldDescriptor, configDiff);
        Mockito.verify(archiveIndexer, Mockito.times(startTaskInvocations)).startTask();
        Mockito.verify(archiveIndexer.taskService, Mockito.times(stopTaskInvocations)).stopTask(any(), anyBoolean());
        if (stopTaskInvocations > 0) {
            Assert.assertNull(archiveIndexer.taskToken);
        }
        if (startTaskInvocations > 0) {
            Assert.assertTrue(StringUtils.isNotBlank(archiveIndexer.taskToken));
            Assert.assertEquals(archiveIndexer.taskToken, "token");
        }
    }

    @AfterMethod
    public void reset() {
        Mockito.reset(archiveIndexer);
    }

    @DataProvider
    private static Object[][] shouldStopProvider() {
        return new Object[][] {
                // conflictGuard, isReady, pauseOrBreak, isEnabled, expectedToStop
                {null, true, false, true, true},
                {null, true, true, true, true},
                {null, false, true, true, true},
                {null, false, false, true, true},
                {new JvmConflictGuard(null), true, false, true, false},
                {new JvmConflictGuard(null), true, true, true, true},
                {new JvmConflictGuard(null), false, true, true, true},
                {new JvmConflictGuard(null), false, false, true, true},
                {new JvmConflictGuard(null), false, false, false, true}
        };
    }

    @DataProvider
    public static Object[][] initProvider() {
        return new Object[][] {
                // isEnabled, timeOfStartInvocation
                {false, 0},
                {true, 1}
        };
    }

    @DataProvider
    public static Object[][] reloadProvider() {
        return new Object[][] {
                //taskToken, newValue, startTaskInvocations, stopTaskInvocations
                {"AAAA", true, 0, 0},
                {"", true, 1, 0},
                {"AAA", false, 0, 1},
                {"", false, 0, 0},
        };
    }
}