package org.artifactory.traffic;

import org.artifactory.api.repo.RepositoryService;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.storage.db.servers.service.ArtifactoryServersCommonService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.traffic.entry.DownloadEntry;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import java.io.File;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;

/**
 * @author dudim
 */
@Test
public class TrafficServiceImplTest extends ArtifactoryHomeBoundTest {

    private AutoCloseable autoCloseable;

    @Mock
    private ArtifactoryServersCommonService serversService;
    @Mock
    private RepositoryService repositoryService;

    @BeforeClass
    public void beforeClass() {
        autoCloseable = MockitoAnnotations.openMocks(this);
        RepoDescriptor repoDescriptor = mock(RepoDescriptor.class);
        when(repositoryService.repoDescriptorByKey(any())).thenReturn(repoDescriptor);
        when(repoDescriptor.getType()).thenReturn(RepoType.Docker);
        when(repositoryService.remoteRepoDescriptorByKey(any())).thenReturn(null);
        when(repositoryService.virtualRepoDescriptorByKey(any())).thenReturn(null);
    }

    @AfterClass
    public void afterClass() throws Exception {
        autoCloseable.close();
    }

    public void handleTransferEntryTest() {
        ArtifactorySystemProperties artifactoryProperties = ArtifactoryHome.get().getArtifactoryProperties();
        TrafficServiceImpl trafficService = new TrafficServiceImpl();
        trafficService.setServersService(serversService);
        trafficService.setRepositoryService(repositoryService);
        DownloadEntry entry = new DownloadEntry("repo:Path4", 51, 8,
                "userAddress4", "userxray");

        TrafficServiceImpl spy;

        artifactoryProperties.setProperty(ConstantValues.trafficCollectionActive.getPropertyName(), "true");
        artifactoryProperties.setProperty(ConstantValues.trafficVersion2Enabled.getPropertyName(), "true");
        spy = spy(trafficService);
        spy.handleTrafficEntry(entry);
        verify(spy, times(1)).logTransferEntry(entry);
        verify(spy, times(1)).logTransferEntryV2(entry, null);

        artifactoryProperties.setProperty(ConstantValues.trafficCollectionActive.getPropertyName(), "true");
        artifactoryProperties.setProperty(ConstantValues.trafficVersion2Enabled.getPropertyName(), "false");
        spy = spy(trafficService);
        spy.handleTrafficEntry(entry);
        verify(spy, times(1)).logTransferEntry(entry);
        verify(spy, times(0)).logTransferEntryV2(entry, null);

        artifactoryProperties.setProperty(ConstantValues.trafficCollectionActive.getPropertyName(), "false");
        artifactoryProperties.setProperty(ConstantValues.trafficVersion2Enabled.getPropertyName(), "false");
        spy = spy(trafficService);
        spy.handleTrafficEntry(entry);
        verify(spy, times(0)).logTransferEntry(entry);
        verify(spy, times(0)).logTransferEntryV2(entry, null);

        artifactoryProperties.setProperty(ConstantValues.trafficCollectionActive.getPropertyName(), "false");
        artifactoryProperties.setProperty(ConstantValues.trafficVersion2Enabled.getPropertyName(), "true");
        spy = spy(trafficService);
        spy.handleTrafficEntry(entry);
        verify(spy, times(0)).logTransferEntry(entry);
        verify(spy, times(0)).logTransferEntryV2(entry, null);
    }

    public void testGetDefaultTrafficLogDir() {
        TrafficServiceImpl trafficServiceSpy = spy(new TrafficServiceImpl());
        trafficServiceSpy.setServersService(serversService);
        trafficServiceSpy.setRepositoryService(repositoryService);
        doReturn(new File("originalPath")).when(trafficServiceSpy).getDefaultTrafficLogDir();
        String logDirPath = trafficServiceSpy.getTrafficLogDir().getPath();
        assertEquals(logDirPath, "originalPath");
    }

    public void testGetNotExistingPathTrafficLogDir() {
        setTrafficLogsDirSystemPropertyProperty("NotExistPath");
        TrafficServiceImpl trafficServiceSpy = spy(new TrafficServiceImpl());
        trafficServiceSpy.setServersService(serversService);
        trafficServiceSpy.setRepositoryService(repositoryService);
        doReturn(new File("originalPath")).when(trafficServiceSpy).getDefaultTrafficLogDir();
        String logDirPath = trafficServiceSpy.getTrafficLogDir().getPath();
        assertEquals(logDirPath, "originalPath");
    }

    public void testGetNewTrafficLogDir() {
        setTrafficLogsDirSystemPropertyProperty("/");
        TrafficServiceImpl trafficServiceSpy = new TrafficServiceImpl();
        trafficServiceSpy.setServersService(serversService);
        trafficServiceSpy.setRepositoryService(repositoryService);
        String logDirPath = trafficServiceSpy.getTrafficLogDir().getPath();
        assertEquals(logDirPath, "/");
    }

    public void testGetEmptyDirectoryTrafficLogDir() {
        setTrafficLogsDirSystemPropertyProperty("");
        TrafficServiceImpl trafficServiceSpy = spy(new TrafficServiceImpl());
        trafficServiceSpy.setServersService(serversService);
        trafficServiceSpy.setRepositoryService(repositoryService);
        doReturn(new File("originalPath")).when(trafficServiceSpy).getDefaultTrafficLogDir();
        String logDirPath = trafficServiceSpy.getTrafficLogDir().getPath();
        assertEquals("originalPath", logDirPath);
    }

    public void testGetSpaceDirectoryTrafficLogDir() {
        setTrafficLogsDirSystemPropertyProperty(" ");
        TrafficServiceImpl trafficServiceSpy = spy(new TrafficServiceImpl());
        trafficServiceSpy.setServersService(serversService);
        trafficServiceSpy.setRepositoryService(repositoryService);
        doReturn(new File("originalPath")).when(trafficServiceSpy).getDefaultTrafficLogDir();
        String logDirPath = trafficServiceSpy.getTrafficLogDir().getPath();
        assertEquals(logDirPath, "originalPath");
    }

    public void testGetNullDirectoryTrafficLogDir() {
        setTrafficLogsDirSystemPropertyProperty(null);
        TrafficServiceImpl trafficServiceSpy = spy(new TrafficServiceImpl());
        trafficServiceSpy.setServersService(serversService);
        trafficServiceSpy.setRepositoryService(repositoryService);
        doReturn(new File("originalPath")).when(trafficServiceSpy).getDefaultTrafficLogDir();
        String logDirPath = trafficServiceSpy.getTrafficLogDir().getPath();
        assertEquals(logDirPath, "originalPath");
    }

    private void setTrafficLogsDirSystemPropertyProperty(String dirPath) {
        ArtifactorySystemProperties artifactoryProperties = ArtifactoryHome.get().getArtifactoryProperties();
        artifactoryProperties.setProperty(ConstantValues.trafficLogsDirectory.getPropertyName(), dirPath);
    }
}