package org.artifactory.util;

import com.github.tomakehurst.wiremock.WireMockServer;
import org.apache.http.HttpStatus;
import org.apache.http.StatusLine;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.artifactory.api.rest.constant.RestConstants;
import org.artifactory.api.rest.constant.SystemRestConstants;
import org.artifactory.repo.http.CloseableHttpExecute;
import org.artifactory.test.WireMockUtils;
import org.mockito.Mockito;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.IOException;
import java.util.function.Function;

import static com.github.tomakehurst.wiremock.client.WireMock.aResponse;
import static com.github.tomakehurst.wiremock.client.WireMock.get;
import static com.github.tomakehurst.wiremock.client.WireMock.urlEqualTo;
import static org.mockito.Mockito.mock;
import static org.testng.Assert.assertEquals;

/**
 * @author Andrey Tashlitsky
 */
public class ArtifactoryVersionUtilTest {

    private static final String SERVICE_VERSION_URL = "/" + RestConstants.PATH_API + "/" + SystemRestConstants.PATH_ROOT
            + "/" + SystemRestConstants.PATH_VERSION_SERVICE;
    private static final String OLD_VERSION_URL = "/" + RestConstants.PATH_API + "/" + SystemRestConstants.PATH_ROOT
            + "/" + SystemRestConstants.PATH_VERSION;

    private WireMockServer wireMockServer;
    private String wireMockUrl;

    @BeforeClass
    public void beforeClass() {
        wireMockServer = WireMockUtils.startWireMockServer();
        wireMockUrl = "http://localhost:" + wireMockServer.port();
    }

    @AfterClass
    public void afterClass() {
        WireMockUtils.stopWireMockServer(wireMockServer);
    }

    @Test(dataProvider = "testGetArtifactoryServiceVersionDataProvider")
    public void testGetArtifactoryServiceVersionWithCloseableHttpClient(int serviceVersionUrlWillReturn,
            int oldVersionUrlWillReturn, int expectedHttpStatus) throws IOException {
        configServiceVersionUrlStub(serviceVersionUrlWillReturn, oldVersionUrlWillReturn);
        try (CloseableHttpClient closeableHttpClient = HttpClients.createDefault()) {
            try (CloseableHttpResponse artifactoryServiceVersion =
                         ArtifactoryVersionUtil.getArtifactoryServiceVersion(closeableHttpClient, wireMockUrl, null)) {
                assertEquals(artifactoryServiceVersion.getStatusLine().getStatusCode(), expectedHttpStatus);
            }
        }
    }

    @Test(dataProvider = "testGetArtifactoryServiceVersionBackwardCompatibilityDataProvider")
    public void testGetArtiServiceVersionBackwardCompatibilityCloseWhenVersionNotFound(int serviceVersionUrlWillReturn,
            int times) throws IOException {
        CloseableHttpResponse response = Mockito.mock(CloseableHttpResponse.class);
        response.setStatusCode(serviceVersionUrlWillReturn);
        Function<HttpGet, CloseableHttpResponse> function = Mockito.mock(Function.class);
        Mockito.when(function.apply(Mockito.any())).thenReturn(response);
        StatusLine statusLine = mock(StatusLine.class);
        Mockito.when(response.getStatusLine()).thenReturn(statusLine);
        Mockito.when(response.getStatusLine().getStatusCode()).thenReturn(serviceVersionUrlWillReturn);
        try (CloseableHttpResponse artifactoryServiceVersion =
                     ArtifactoryVersionUtil.getServiceVersionWithBackwardCompatibility(function, null, null, null)) {
            Mockito.verify(artifactoryServiceVersion, Mockito.times(times)).close();
        }
    }

    @Test(dataProvider = "testGetArtifactoryServiceVersionDataProvider")
    public void testGetArtifactoryServiceVersionWithCloseableHttpExecute(int serviceVersionUrlWillReturn,
            int oldVersionUrlWillReturn, int expectedHttpStatus) throws IOException {
        configServiceVersionUrlStub(serviceVersionUrlWillReturn, oldVersionUrlWillReturn);
        try (CloseableHttpClient closeableHttpClient = HttpClients.createDefault()) {
            CloseableHttpExecuteStub closeableHttpExecute = new CloseableHttpExecuteStub(closeableHttpClient);
            try (CloseableHttpResponse artifactoryServiceVersion =
                         ArtifactoryVersionUtil.getArtifactoryServiceVersion(closeableHttpExecute, wireMockUrl, null)) {
                assertEquals(artifactoryServiceVersion.getStatusLine().getStatusCode(), expectedHttpStatus);
            }
        }
    }

    @DataProvider
    public static Object[][] testGetArtifactoryServiceVersionDataProvider() {
        return new Object[][] {
                {HttpStatus.SC_NOT_FOUND, HttpStatus.SC_NOT_FOUND, HttpStatus.SC_NOT_FOUND},
                {HttpStatus.SC_NOT_FOUND, HttpStatus.SC_OK, HttpStatus.SC_OK},
                {HttpStatus.SC_OK, HttpStatus.SC_NOT_FOUND, HttpStatus.SC_OK},
                {HttpStatus.SC_OK, HttpStatus.SC_OK, HttpStatus.SC_OK},
                {HttpStatus.SC_INTERNAL_SERVER_ERROR, HttpStatus.SC_OK, HttpStatus.SC_INTERNAL_SERVER_ERROR},
        };
    }

    @DataProvider
    public static Object[][] testGetArtifactoryServiceVersionBackwardCompatibilityDataProvider() {
        return new Object[][] {
                {HttpStatus.SC_NOT_FOUND, 1},
                {HttpStatus.SC_OK, 0}
        };
    }

    private void configServiceVersionUrlStub(int serviceVersionHttpStatus, int oldVersionHttpStatus) {
        wireMockServer.stubFor(
                get(urlEqualTo(SERVICE_VERSION_URL)).willReturn(aResponse().withStatus(serviceVersionHttpStatus)));
        wireMockServer.stubFor(
                get(urlEqualTo(OLD_VERSION_URL)).willReturn(aResponse().withStatus(oldVersionHttpStatus)));
    }

    private static class CloseableHttpExecuteStub implements CloseableHttpExecute {

        private final CloseableHttpClient closeableHttpClient;

        public CloseableHttpExecuteStub(CloseableHttpClient closeableHttpClient) {
            this.closeableHttpClient = closeableHttpClient;
        }

        @Override
        public CloseableHttpResponse executeMethod(HttpRequestBase method) throws IOException {
            return closeableHttpClient.execute(method);
        }
    }
}
