/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.aql.result;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonPropertyOrder;
import com.fasterxml.jackson.annotation.PropertyAccessor;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import org.artifactory.aql.model.AqlDomainEnum;
import org.artifactory.aql.model.AqlPermissionProvider;
import org.artifactory.aql.model.AqlPhysicalFieldEnum;
import org.artifactory.aql.util.AqlUtils;
import org.artifactory.repo.RepoPath;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.Closeable;
import java.sql.ResultSet;

/**
 * @author Gidi Shabat
 */
public abstract class AqlRestResult implements Closeable {
    public static final String QUERY_PREFIX = "\n{\n\"results\" : [ ";
    protected static final String NUMBER_OF_ROWS = "<NUMBER_OF_ROWS>";
    protected static final String QUERY_POSTFIX = " ],\n\"range\" : " + NUMBER_OF_ROWS + "\n}\n";
    protected static final ObjectMapper mapper = createObjectMapper();
    protected static final JsonFactory factory = mapper.getFactory();

    private static final Logger log = LoggerFactory.getLogger(AqlRestResult.class);

    protected AqlPermissionProvider permissionProvider;

    AqlRestResult(AqlPermissionProvider permissionProvider) {
        this.permissionProvider = permissionProvider;
    }

    protected boolean canRead(AqlDomainEnum domain, ResultSet resultSet) {
        if (permissionProvider.isAdmin()) {
            return true;
        } else {
            if (AqlDomainEnum.items == domain) {
                try {
                    String itemRepo = resultSet.getString(AqlPhysicalFieldEnum.itemRepo.name());
                    String itemPath = resultSet.getString(AqlPhysicalFieldEnum.itemPath.name());
                    String itemName = resultSet.getString(AqlPhysicalFieldEnum.itemName.name());
                    RepoPath repoPath = AqlUtils.fromAql(itemRepo, itemPath, itemName);
                    return permissionProvider.canRead(repoPath);
                } catch (Exception e) {
                    logMinimalFieldError(e);
                }
            }
            return false;
        }
    }

    protected boolean canRead(AqlDomainEnum domain, String repo, String path, String name) {
        if (permissionProvider.isAdmin()) {
            return true;
        } else {
            if (AqlDomainEnum.items == domain) {
                try {
                    RepoPath repoPath = AqlUtils.fromAql(repo, path, name);
                    return permissionProvider.canRead(repoPath);
                } catch (Exception e) {
                    logMinimalFieldError(e);
                }
            }
            return false;
        }
    }

    public abstract byte[] read();

    public abstract byte[] readCompact();

    private static ObjectMapper createObjectMapper() {
        ObjectMapper newMapper = new ObjectMapper();
        newMapper.setSerializationInclusion(JsonInclude.Include.NON_NULL);
        newMapper.setVisibility(PropertyAccessor.ALL, JsonAutoDetect.Visibility.NONE);
        newMapper.setVisibility(PropertyAccessor.FIELD, JsonAutoDetect.Visibility.ANY);
        return newMapper;
    }

    private void logMinimalFieldError(Exception e) {
        log.error("AQL minimal field expectation error: repo, path and name", e);
    }

    @JsonSerialize(include = JsonSerialize.Inclusion.NON_NULL)
    @JsonPropertyOrder(value = {"start", "end", "total"}, alphabetic = true)
    protected class Range {

        @JsonProperty("start_pos")
        protected Long start;
        @JsonProperty("end_pos")
        protected Long end;
        @JsonProperty("total")
        protected Long total;
        @JsonProperty("limit")
        protected Long limited;

        public Range(long start, long end, long limited) {
            this.start = start;
            this.end = end;
            this.limited = Long.MAX_VALUE == limited ? null : limited;
        }

        public Range(long start, long end, long total, long limited) {
            this.start = start;
            this.end = end;
            this.total = total;
            this.limited = Long.MAX_VALUE == limited ? null : limited;
        }
    }
}
