package org.artifactory.federated.metadata;

import org.artifactory.async.Async;
import org.artifactory.model.FederatedRepoConfig;

import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.Map;

/**
 * The service for retrieving the metadata object json and invalidating the cache
 *
 * @author Alexei Vainshtein
 */
public interface FederatedRepoMetadataService {

    /**
     * Returns package metadata of a remote member.
     * First goes to cache, if not cached (or expired), will try secondly in the db.
     * If missed in the db (cached expired or not cached), will go and try from the remote mirror.
     *
     * @param packageType - The package type to get the cached metadata
     * @param url         - The member url
     * @param proxyRef    - The proxy name if configured. Can be null
     * @return the metadata json of the requested package type from the remote member.
     * @throws MetadataRetrievalException upon error fetching the metadata
     */
    String getPackageMetadata(String packageType, String url, String proxyRef) throws MetadataRetrievalException;

    /**
     * Returns package metadata of the running service.
     * This is calculated internally upon demand.
     *
     * @param packageType - The package type to get the metadata for
     * @param baseUrl     - The server base url
     * @return the metadata json of the requested package type from the remote member.
     */
    String getPackageMetadataForLocal(String packageType, String baseUrl);

    /**
     * @return a json of the basic configuration.
     * Represents object {@link org.artifactory.model.FederatedRepoConfig}
     */
    String getFederatedRepoConfigMetadata();


    /**
     * Invalidate the cache for the specific member
     *
     * @param packageType - The package type to be invalidated
     * @param baseUrl     - The member url
     * @param proxyRef    - The proxy name if configured. Can be null
     */
    default void invalidateCacheForMember(String packageType, String baseUrl, String proxyRef) {

    }

    /**
     * @return the highest metadata version combined from the basic and the package class's
     */
    default int fetchHighestMetadataVersion() {
        return Integer.MIN_VALUE;
    }

    /**
     * Extracts from the database the relevant metadata per package.
     * Grabs the information from the collection of repositories provided, collects all the required urls that needs
     * to be extracted from those repositories and fetch those from the db
     *
     * @param federatedRepositories - A collection of all the federated repositories to collect the metadata for
     * @return a Map of metadata per member for each repo
     */
    default Map<String /*repo*/, Map<String /*url*/, String /*metadata*/>> getPackagesMetadata(
            Collection<FederatedRepoConfig> federatedRepositories) {
        return Collections.emptyMap();
    }

    /**
     * Extracts from the database the relevant basic metadata.
     * Grabs the information from the collection of repositories provided, collects all the required urls that needs
     * to be extracted from those repositories and fetch those from the db
     *
     * @param federatedRepositories - A collection of all the federated repositories to collect the metadata for
     * @return a Map of metadata per member for each repo
     */
    default Map<String, Map<String, String>> getRemoteBasicMetadata(
            Collection<FederatedRepoConfig> federatedRepositories) {
        return Collections.emptyMap();
    }

    /**
     * This method send this server metadata to the members url
     *
     * @param originUrl       the origin that its metadata was changes
     * @param memberUrl       the member to send the new metadata
     * @param updatedMetaData the newer metadata
     * @param proxyRef        the proxy that enables the origin get to the remote
     * @param updateMembers   a boolean that represents if the other member need to other his members as well.
     *                        During creation of a repo this will be true. This is required to avoid loop and
     *                        re-sending the metadata to the same member that created the repo in the beginning.
     * @throws IOException when failed to put metadata in mirror
     */
    default void sendRemoteMetadataChange(String originUrl, String memberUrl, String updatedMetaData, String proxyRef,
            boolean updateMembers) throws IOException {}

    /**
     * Put the given metadata of the origin url
     * @param typeToMetadata the newer metadata for each package type
     * @param originUrl the member that triggered the update
     */
    default void putMirrorMetadataLocally(String originUrl,
            Map<String /*package type*/, String /*metadata of the package*/> typeToMetadata) {
    }

    /**
     * Updates the fedRepos members with the new metadata
     * @param fedRepos the repositories that were changed
     * @param filterMemberUrl the url that metadata update is not required for.
     */
    @Async
    default void updateMembersMetadata(Collection<FederatedRepoConfig> fedRepos, String filterMemberUrl){}

    /**
     * Retrieves metadata from the other members in async way. This will only put the pulled metadata in the database.
     *
     * @param federatedRepoConfigs - All federated repositories that the metadata from their members need to be fetched.
     */
    @Async
    default void fetchMetadataOfRemoteMembersAsync(Collection<FederatedRepoConfig> federatedRepoConfigs){}
}
