package org.artifactory.model;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.Setter;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.util.LocalDefaultPackageConfigSetter;
import org.artifactory.util.ParamValidationUtils;
import org.jfrog.common.config.validation.NotUnicode;

import javax.validation.Valid;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

/**
 * Local repository configurations
 *
 * @author shayb
 */
@Valid
@Getter
public class LocalRepoConfig implements LocalConfig {

    @NotUnicode
    private final String key;
    @Valid
    @Setter(AccessLevel.PROTECTED)
    private BaseConfig baseConfig;
    @Valid
    @Setter(AccessLevel.PROTECTED)
    private LocalRepositoryRepoTypeConfig repoTypeConfig = new LocalRepositoryRepoTypeConfig();
    private final RepoType packageType;
    @Setter(AccessLevel.PROTECTED)
    private Map<String, String> packageTypeConfig = new HashMap<>();
    @Valid
    @Setter(AccessLevel.PROTECTED)
    private SecurityConfig securityConfig = new SecurityConfig();
    @Setter
    @JsonIgnore
    private long modification;

    @JsonCreator
    LocalRepoConfig(@JsonProperty("key") String repoKey, @JsonProperty("packageType") RepoType packageType) {
        this.key = ParamValidationUtils.requireNonBlank(repoKey, "Repository key cannot be blank");
        this.packageType = Objects.requireNonNull(packageType, "Repository package type cannot be null");
        this.baseConfig = new BaseConfig();
        LocalDefaultPackageConfigSetter.setDefaultValues(packageType, packageTypeConfig);
    }

    /**
     * Factory method enabling exclusive access to instance creation
     *
     * @param repoKey     the repository key
     * @param packageType the repository package type
     * @return Local repo configuration
     * @throws IllegalArgumentException if {@code repoKey} is blank
     * @throws NullPointerException     if {@code packageType} is null
     */
    public static LocalRepoConfig create(String repoKey, RepoType packageType) {
        return new LocalRepoConfig(repoKey, packageType);
    }

    /**
     * Factory method enabling exclusive access to deep copy a local repo config
     *
     * @param localRepoConfig the local repo config to be copied
     * @return a copy of the local repo config
     * @throws NullPointerException if {@code localRepoConfig} is null
     */
    static LocalRepoConfig copy(LocalRepoConfig localRepoConfig) {
        Objects.requireNonNull(localRepoConfig, "Local repo config cannot be null");
        LocalRepoConfig repoCopy = new LocalRepoConfig(localRepoConfig.getKey(), localRepoConfig.getPackageType());
        repoCopy.setBaseConfig(new BaseConfig(localRepoConfig.getBaseConfig()));
        repoCopy.setRepoTypeConfig(new LocalRepositoryRepoTypeConfig(localRepoConfig.getRepoTypeConfig()));
        repoCopy.setPackageTypeConfig(new HashMap<>(localRepoConfig.getPackageTypeConfig()));
        repoCopy.setSecurityConfig(new SecurityConfig(localRepoConfig.getSecurityConfig()));
        repoCopy.setModification(localRepoConfig.getModification());
        return repoCopy;
    }

    @Override
    public boolean isCache() {
        return false;
    }

    @Override
    public RepoDetailsType getRepoType() {
        return RepoDetailsType.LOCAL;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }
        LocalRepoConfig that = (LocalRepoConfig) o;
        return Objects.equals(key, that.key);
    }

    @Override
    public int hashCode() {
        return Objects.hash(key);
    }

    @Override
    public String toString() {
        return key;
    }

}
