package org.artifactory.model;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.util.ParamValidationUtils;
import org.artifactory.util.RemoteDefaultPackageConfigSetter;
import org.jfrog.common.config.validation.NotUnicode;

import javax.validation.Valid;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

/**
 * Remote repository configurations
 *
 * @author shayb
 */
@Slf4j
@Valid
@Getter
public class RemoteRepoConfig implements RealRepoConfig, ProxyRepoConfig {
    @NotUnicode
    private final String key;
    @Valid
    @Setter(AccessLevel.PRIVATE)
    private BaseConfig baseConfig;
    @Valid
    @Setter(AccessLevel.PRIVATE)
    private RemoteRepositoryRepoTypeConfig repoTypeConfig = new RemoteRepositoryRepoTypeConfig();
    private RepoType packageType;
    @Setter(AccessLevel.PRIVATE)
    private Map<String, String> packageTypeConfig = new HashMap<>();
    @Valid
    @Setter(AccessLevel.PRIVATE)
    private SecurityConfig securityConfig = new SecurityConfig();
    @JsonIgnore
    @Setter
    private long modification;

    @JsonCreator
    private RemoteRepoConfig(@JsonProperty("key") String repoKey, @JsonProperty("packageType") RepoType packageType) {
        this.key = ParamValidationUtils.requireNonBlank(repoKey, "Repository key cannot be blank");
        this.packageType = Objects.requireNonNull(packageType, "Repository package type cannot be null");
        this.baseConfig = new BaseConfig();
        RemoteDefaultPackageConfigSetter.setDefaultValues(packageType, packageTypeConfig);
    }

    /**
     * Factory method enabling exclusive access to instance creation
     *
     * @param repoKey     the repository key
     * @param packageType the repository package type
     * @return Remote repo configuration
     * @throws IllegalArgumentException if {@code repoKey} is blank
     * @throws NullPointerException     if {@code packageType} is null
     */
    public static RemoteRepoConfig create(String repoKey, RepoType packageType) {
        return new RemoteRepoConfig(repoKey, packageType);
    }

    /**
     * Factory method enabling exclusive access to deep copy a remote repo config
     *
     * @param remoteRepoConfig the remote repo config to be copied
     * @return a copy of the remote repo config
     * @throws NullPointerException if {@code remoteRepoConfig} is null
     */
    static RemoteRepoConfig copy(RemoteRepoConfig remoteRepoConfig) {
        Objects.requireNonNull(remoteRepoConfig, "Remote repo config cannot be null");
        RemoteRepoConfig repoCopy = new RemoteRepoConfig(remoteRepoConfig.getKey(), remoteRepoConfig.getPackageType());
        repoCopy.setBaseConfig(new BaseConfig(remoteRepoConfig.getBaseConfig()));
        repoCopy.setRepoTypeConfig(new RemoteRepositoryRepoTypeConfig(remoteRepoConfig.getRepoTypeConfig()));
        repoCopy.setPackageTypeConfig(new HashMap<>(remoteRepoConfig.getPackageTypeConfig()));
        repoCopy.setSecurityConfig(new SecurityConfig(remoteRepoConfig.getSecurityConfig()));
        repoCopy.setModification(remoteRepoConfig.getModification());
        return repoCopy;
    }

    @Override
    public boolean isLocal() {
        return false;
    }

    @Override
    public boolean isCache() {
        return false;
    }

    public void processUrl() {
        final String[] credentials = extractCredentialsFromUrl(getRepoTypeConfig().getUrl());
        if (credentials != null) {
            getRepoTypeConfig().setUsername(credentials[0]);
            getRepoTypeConfig().setPassword(credentials[1]);
            getRepoTypeConfig().setUrl(getRepoTypeConfig().getUrl()
                    .replace(getRepoTypeConfig().getUsername() + ":" + getRepoTypeConfig().getPassword() + "@", ""));
        }
    }

    String[] extractCredentialsFromUrl(String url) {
        return Optional.ofNullable(url)
                .map(l -> {
                    try {
                        return new URL(l);
                    } catch (MalformedURLException e) {
                        log.warn("Can not parse URL: {}", url);
                        return null;
                    }
                })
                .map(URL::getUserInfo)
                .map(l -> l.split(":"))
                .filter(l -> l.length > 1)
                .orElse(null);
    }

    @Override
    public void setProxyRef(String proxyKey) {
        getRepoTypeConfig().setProxyRef(proxyKey);
    }

    @Override
    public String getProxyRef() {
        return getRepoTypeConfig().getProxyRef();
    }

    @Override
    public void setDisableProxy(boolean disable) {
        getRepoTypeConfig().setDisableProxy(disable);
    }

    @Override
    public boolean isProxyDisabled() {
        return getRepoTypeConfig().isDisableProxy();
    }

    @Override
    public RepoDetailsType getRepoType() {
        return RepoDetailsType.REMOTE;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }
        RemoteRepoConfig that = (RemoteRepoConfig) o;
        return Objects.equals(key, that.key);
    }

    @Override
    public int hashCode() {
        return Objects.hash(key);
    }

    @Override
    public String toString() {
        return key;
    }

}
