package org.artifactory.storage.db.binstore.service;

import com.google.common.collect.Maps;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.addon.ha.ClusterOperationsService;
import org.artifactory.addon.ha.interceptor.ClusterTopologyListener;
import org.artifactory.api.common.BasicStatusHolder;
import org.artifactory.checksum.ChecksumType;
import org.artifactory.config.CentralConfigKey;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.spring.Reloadable;
import org.artifactory.storage.GCCandidate;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.binstore.service.ArtifactoryBinaryService;
import org.artifactory.storage.db.servers.model.ArtifactoryServer;
import org.jfrog.storage.authentication.AuthenticationTokenProvider;
import org.jfrog.storage.binstore.BinariesInfo;
import org.jfrog.storage.binstore.InternalBinaryService;
import org.jfrog.storage.binstore.common.BinaryElementImpl;
import org.jfrog.storage.binstore.exceptions.BinaryDbException;
import org.jfrog.storage.binstore.exceptions.BinaryNotFoundException;
import org.jfrog.storage.binstore.exceptions.BinaryRejectedException;
import org.jfrog.storage.binstore.exceptions.SignedUrlException;
import org.jfrog.storage.binstore.ifc.BinaryStoreFederatedMember;
import org.jfrog.storage.binstore.ifc.BinaryStream;
import org.jfrog.storage.binstore.ifc.ClusterNode;
import org.jfrog.storage.binstore.ifc.GarbageCollectorInfo;
import org.jfrog.storage.binstore.ifc.ManualRedirectType;
import org.jfrog.storage.binstore.ifc.RedirectType;
import org.jfrog.storage.binstore.ifc.model.BinaryElement;
import org.jfrog.storage.binstore.ifc.model.BinaryElementRequest;
import org.jfrog.storage.binstore.ifc.model.BinaryProvidersInfo;
import org.jfrog.storage.binstore.ifc.model.BinaryProvidersResponseMap;
import org.jfrog.storage.binstore.ifc.model.StorageInfo;
import org.jfrog.storage.binstore.manager.BinaryProviderManager;
import org.jfrog.storage.binstore.providers.base.BinariesTasksInfo;
import org.jfrog.storage.binstore.providers.base.PersistedRemoteTask;
import org.jfrog.storage.common.StatusReporter;
import org.slf4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.annotation.Nullable;
import javax.annotation.PreDestroy;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.sql.SQLException;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

/**
 * @author TamirH
 */
@Slf4j
@Service
@Reloadable(beanClass = ArtifactoryBinaryService.class, initAfter = {ClusterOperationsService.class},
        listenOn = CentralConfigKey.none)
public class ArtifactoryBinaryServiceImpl
        implements ArtifactoryBinaryService, ClusterTopologyListener, ContextCreationListener {

    private InternalBinaryService binaryService;

    @Autowired
    public void setBinaryService(InternalBinaryService binaryService) {
        this.binaryService = binaryService;
    }

    @Override
    @Deprecated
    public InputStream getBinary(String sha1, Map<String, String> headers) throws BinaryNotFoundException {
        return binaryService.getBinary(sha1, headers);
    }

    @Override
    public InputStream getBinary(BinaryElement element) throws BinaryNotFoundException {
        return binaryService.getBinary(element);
    }

    @NonNull
    public InputStream getBinaryForChecksumDeploy(BinaryElement element, boolean checkInFilestore) {
        try {
            return binaryService.getBinaryForChecksumDeploy(element, checkInFilestore);
        } catch (BinaryDbException e) {
            throw new StorageException(e);
        }
    }

    @Override
    public RedirectType calcRedirectType(@NonNull RedirectType redirectType, boolean isAol,
            ManualRedirectType manualRedirectType) {
        return binaryService.calcRedirectType(redirectType, isAol, manualRedirectType);
    }

    @Override
    @NonNull
    public String getSignedUrl(@NonNull BinaryElement binaryElementRequest) throws SignedUrlException {
        return binaryService.getSignedUrl(binaryElementRequest);
    }

    @Override
    @Nullable
    public BinaryElement findBinary(ChecksumType checksumType, String checksum) {
        try {
            return binaryService.findBinary(org.jfrog.storage.common.ChecksumType.valueOf(checksumType.name()),
                    checksum);
        } catch (BinaryDbException e) {
            throw new StorageException(e);
        }
    }

    @Override
    @NonNull
    public Set<BinaryElement> findBinaries(@Nullable Collection<String> checksums) {
        try {
            return binaryService.findBinaries(checksums);
        } catch (BinaryDbException e) {
            throw new StorageException(e);
        }
    }

    @Override
    public BinaryProvidersInfo<Map<String, String>> getBinaryProvidersInfo() {
        return binaryService.getBinaryProvidersInfo();
    }

    @Override
    @NonNull
    public BinaryElement addBinary(BinaryStream in) throws IOException, BinaryRejectedException {
        try {
            return binaryService.addBinary(in);
        } catch (BinaryDbException | org.jfrog.storage.StorageException e) {
            throw new StorageException(e);
        }
    }

    @Override
    public void ping(boolean localOnly) {
        binaryService.ping(localOnly);
    }

    @Override
    public BinariesInfo getBinariesInfo() {
        try {
            return binaryService.getBinariesInfo();
        } catch (BinaryDbException e) {
            throw new StorageException(e);
        }
    }

    @Override
    public long getStorageSize() {
        try {
            return binaryService.getStorageSize();
        } catch (BinaryDbException e) {
            throw new StorageException(e);
        }
    }

    @Override
    public void forceOptimizationOnce() {
        binaryService.forceOptimizationOnce();
    }

    @Override
    public BinaryElement createBinaryElement(String sha1, String sha2, String md5, long length,
            Map<String, String> headers) {
        return binaryService.createBinaryElement(sha1, sha2, md5, length, headers);
    }

    @Override
    public BinaryStream createBinaryStream(InputStream in, Map<String, String> binaryHeaders) {
        Map<String, String> headers = binaryHeaders == null ? Maps.newHashMap() : binaryHeaders;
        return binaryService.createBinaryStream(in, headers);
    }

    @Override
    public boolean executeBinaryCleaner(GCCandidate candidate, GarbageCollectorInfo result) {
        BinaryElementImpl element =
                new BinaryElementImpl(candidate.sha1(), candidate.sha2(), candidate.md5(), candidate.length());
        return binaryService.executeBinaryCleaner(element, result);
    }

    @Override
    public boolean isCdnAllowed() {
        return binaryService.isCdnAllowed();
    }

    @Override
    public void updateFederateRepoBinaryProvider(@NonNull Collection<BinaryStoreFederatedMember> membersToAdd,
            @NonNull Collection<BinaryStoreFederatedMember> membersToDelete,
            AuthenticationTokenProvider authenticationTokenProvider) {
        binaryService.updateFederateRepoBinaryProvider(membersToAdd, membersToDelete, authenticationTokenProvider);
    }

    @Override
    @Nullable
    public BinaryElement addBinaryRecord(Map<String, String> headers, String sha1, String sha2, String md5, long length)
            throws BinaryNotFoundException {
        try {
            return binaryService.addBinaryRecord(headers, sha1, sha2, md5, length);
        } catch (BinaryDbException e) {
            throw new StorageException(e);
        }
    }

    @Override
    public GarbageCollectorInfo runFullGarbageCollect() {
        try {
            return binaryService.runFullGarbageCollect();
        } catch (org.jfrog.storage.StorageException se) {
            throw new StorageException(se);
        }
    }

    @Override
    public void deleteBlobInfos(List<String> deletedBinaries) {
        binaryService.deleteBlobInfos(deletedBinaries);
    }

    @Override
    public void prune(BasicStatusHolder statusHolder) {
        try {
            binaryService.prune(new StatusHolderDelegator(statusHolder));
        } catch (org.jfrog.storage.StorageException se) {
            throw new StorageException(se);
        }
    }

    @Override
    @Deprecated
    @Nullable
    public File getBinariesDir() {
        return binaryService.getBinariesDir();
    }

    @Override
    public boolean isCloudProviderSupportingRedirectionConfigured() {
        return binaryService.isCloudProviderSupportingRedirectionConfigured();
    }

    @Override
    public StorageInfo getStorageInfoSummary() {
        return binaryService.getStorageInfoSummary();
    }

    @Override
    public BinaryProviderManager getBinaryProviderManager() {
        return binaryService.getBinaryProviderManager();
    }

    @Override
    public Collection<BinaryElement> findAllBinaries() {
        try {
            return binaryService.findAllBinaries();
        } catch (BinaryDbException e) {
            throw new StorageException(e);
        }
    }

    @Override
    public int deleteUnusedArchivePaths() {
        try {
            log.debug("Deleting unused archive paths");
            return binaryService.deleteUnusedArchivePaths();
        } catch (StorageException e) {
            log.error("Failed to delete unique paths: {}", e.getMessage());
            log.debug("Failed to delete unique paths", e);
            return 0;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public int deleteUnusedArchiveNames() {
        try {
            return binaryService.deleteUnusedArchiveNames();
        } catch (StorageException e) {
            log.error("Failed to delete unique archive names: {}", e.getMessage());
            log.debug("Failed to delete unique archive paths", e);
            return 0;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public int incrementNoDeleteLock(String sha1) {
        return binaryService.incrementNoDeleteLock(sha1);
    }

    @Override
    public void decrementNoDeleteLock(String sha1) {
        binaryService.decrementNoDeleteLock(sha1);
    }

    @Override
    public boolean updateSha2ForSha1(String targetSha1, String newSha2) throws SQLException {
        try {
            return binaryService.updateSha2ForSha1(targetSha1, newSha2);
        } catch (BinaryDbException e) {
            throw new StorageException(e);
        }
    }

    @Override
    public boolean canCloudProviderGenerateRedirection(String sha1) {
        return binaryService.canCloudProviderGenerateRedirection(sha1);
    }

    @Override
    public BinaryProvidersResponseMap isFileRestoredFromColdStorage(BinaryElementRequest request) {
        return binaryService.isFileRestoredFromColdStorage(request);
    }

    @Override
    public void initiateRestoreFromColdStorage(BinaryElementRequest request) {
        binaryService.initiateRestoreFromColdStorage(request);
    }

    @Override
    public List<PersistedRemoteTask> getNextRemoteTaskFromDataDb(long batchSize, int maxExecutionsCount,
            long retryDelayMs, long maximumExecTimeMs) throws SQLException {
        return binaryService.getNextRemoteTaskFromDataDb(batchSize, maxExecutionsCount, retryDelayMs,
                maximumExecTimeMs);
    }

    @Override
    public boolean isCloudProviderConfigured() {
        return binaryService.isCloudProviderConfigured();
    }

    @Override
    public BinariesTasksInfo getBinariesTasksRepoStatus(String repoKey, long currentTime) {
        return binaryService.getBinariesTasksRepoStatus(repoKey, currentTime);
    }

    @Override
    public BinariesTasksInfo getBinariesTasksStatus() {
        long currentTime = System.currentTimeMillis();
        return binaryService.getBinariesTasksStatus(currentTime);
    }

    @Override
    public List<String> getAndManageErrors() {
        return binaryService.getAndManageErrors();
    }

    @Override
    public void clusterTopologyChanged(List<ArtifactoryServer> activeNodes) {
        List<ClusterNode> active = activeNodes.stream()
                .filter(Objects::nonNull)
                .map(this::clusterNodeFromArtServer)
                .toList();
        binaryService.clusterTopologyChanged(active);
    }

    private ClusterNode clusterNodeFromArtServer(ArtifactoryServer artifactoryServer) {
        return new ClusterNode(artifactoryServer.getContextUrl(), artifactoryServer.getServerId(),
                artifactoryServer.getArtifactoryVersion(), artifactoryServer.getArtifactoryRevision());
    }

    @Override
    public void onContextCreated() {
        this.binaryService.contextCreated();
    }

    @Override
    @PreDestroy
    public void destroy() {
        this.binaryService.contextDestroyed();
    }

    private static class StatusHolderDelegator implements StatusReporter {
        BasicStatusHolder statusHolder;

        public StatusHolderDelegator(BasicStatusHolder statusHolder) {
            this.statusHolder = statusHolder;
        }

        @Override
        public void error(String status, Throwable throwable, @NonNull Logger logger) {
            this.statusHolder.error(status, throwable, logger);
        }

        @Override
        public void error(String status, @NonNull Logger logger) {
            this.statusHolder.error(status, logger);
        }

        @Override
        public void status(String statusMsg, @NonNull Logger logger) {
            this.statusHolder.status(statusMsg, logger);
        }
    }
}
