package org.artifactory.storage.db.binstore.service.cleanup;

import lombok.extern.slf4j.Slf4j;
import org.artifactory.common.ConstantValues;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.storage.binstore.service.ArtifactoryBinaryService;
import org.artifactory.storage.binstore.service.cleanup.BinariesGarbageCollectorService;
import org.artifactory.storage.binstore.service.cleanup.GarbageCollectionDatastoreService;
import org.artifactory.storage.binstore.service.cleanup.GCService;
import org.artifactory.storage.fs.service.ConfigsService;
import org.jfrog.storage.binstore.ifc.GarbageCollectorInfo;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.concurrent.atomic.AtomicInteger;

import static org.artifactory.storage.fs.service.ConfigsServiceKey.ARTIFACTORY_GC_ITERATION;

/**
 * @author TamirH
 */
@Slf4j
@Service
public class GCServiceImpl implements GCService {

    private final AtomicInteger fallbackGcIteration = new AtomicInteger(0);
    private final ArtifactoryBinaryService binaryService;
    private final BinariesGarbageCollectorService minorGcCollectorService;
    private final TrashService trashService;
    private final int skipFullGcCount;
    private final GarbageCollectionDatastoreService garbageCollectionDatastoreService;
    private final ConfigsService configsService;

    @Autowired
    public GCServiceImpl(ArtifactoryBinaryService binaryService,
            BinariesGarbageCollectorService minorGcCollectorService,
            TrashService trashService, GarbageCollectionDatastoreService garbageCollectionDatastoreService,
            ConfigsService configsService) {
        this.binaryService = binaryService;
        this.minorGcCollectorService = minorGcCollectorService;
        this.trashService = trashService;
        this.garbageCollectionDatastoreService = garbageCollectionDatastoreService;
        this.configsService = configsService;
        this.skipFullGcCount = ConstantValues.gcSkipFullGcBetweenMinorIterations.getInt();
    }

    @Override
    public void startGarbageCollect() {
        log.info("Triggering Garbage Collection");
        IterationResult iteration = getGCIteration();
        log.info("Running {} iteration of GC", iteration.value);
        GarbageCollectorInfo garbageCollectorInfo;
        try {
            if (this.trashService.isTrashcanEnabled()) {
                garbageCollectorInfo = executeTrashStrategy(iteration.isFullGc());
            } else {
                garbageCollectorInfo = this.binaryService.runFullGarbageCollect();
            }
        } finally {
            updateConfigIteration(iteration);
        }
        if (garbageCollectorInfo != null) {
            garbageCollectionDatastoreService.saveLastGCRunInfo(garbageCollectorInfo);
        } else {
            log.warn("GC finished without report");
        }
    }

    @Override
    public void forceFull() {
        GarbageCollectorInfo garbageCollectorInfo;

        if (this.trashService.isTrashcanEnabled()) {
            garbageCollectorInfo = executeTrashStrategy(true);
        } else {
            garbageCollectorInfo = this.binaryService.runFullGarbageCollect();
        }

        if (garbageCollectorInfo != null) {
            garbageCollectionDatastoreService.saveLastGCRunInfo(garbageCollectorInfo);
        } else {
            log.warn("Gc finished without report");
        }
    }

    private GarbageCollectorInfo executeTrashStrategy(boolean shouldExecuteFull) {
        if (shouldExecuteFull) {
            this.runTrashAndBinariesGarbageCollect();
            return this.binaryService.runFullGarbageCollect();
        } else {
            return runTrashAndBinariesGarbageCollect();
        }
    }

    private GarbageCollectorInfo runTrashAndBinariesGarbageCollect() {
        return this.minorGcCollectorService.startMinorGC();
    }

    private IterationResult getGCIteration() {
        int nextIteration;
        boolean success;
        try {
            String persistedGcIteration = configsService.getConfig(ARTIFACTORY_GC_ITERATION);
            int prevIteration = persistedGcIteration != null ? Integer.parseInt(persistedGcIteration) : 0;
            nextIteration = prevIteration + 1 >= skipFullGcCount ? 0 : prevIteration + 1;
            success = true;
        } catch (Exception e) {
            log.warn("Couldn't read gc iteration from configs, continue with previous state: {}", e.getMessage());
            log.debug("", e);
            nextIteration = fallbackGcIteration.updateAndGet(n -> (n + 1 >= skipFullGcCount) ? 0 : n + 1);
            success = false;
        }
        fallbackGcIteration.set(nextIteration);
        return new IterationResult(success, nextIteration, System.currentTimeMillis());
    }

    private void updateConfigIteration(IterationResult iteration) {
        try {
            //do not update config if it wasn't retrieved from the database configs
            if (iteration.isSuccessful) {
                configsService.addOrUpdateConfig(ARTIFACTORY_GC_ITERATION, iteration.value + "", iteration.startTime);
            }
        } catch (Exception e) {
            log.warn("GC iteration config wasn't updated: {}", e.getMessage());
            log.debug("", e);
        }
    }


    private record IterationResult(boolean isSuccessful, int value, long startTime) {
        boolean isFullGc() {
            return value == 0;
        }
    }
}
