package org.artifactory.storage.db.repo.service;

import lombok.NonNull;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.db.fs.service.RepositoryPersistingService;
import org.artifactory.storage.db.repo.RepositoryConfigStorageUtils;
import org.artifactory.storage.db.repo.dao.RepositoryConfigDao;
import org.artifactory.storage.db.repo.entity.DbRepositoryConfig;
import org.artifactory.util.bundles.BundleConfigurationUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.support.TransactionSynchronizationManager;
import org.springframework.util.CollectionUtils;

import java.sql.SQLException;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

import static org.artifactory.descriptor.repo.RepoType.ReleaseBundles;
import static org.artifactory.repo.RepoDetailsType.RELEASE_BUNDLE;

/**
 * @author Tamar Rosen
 */
@Service
public class RepositoryConfigStorageServiceImpl implements RepositoryConfigStorageService {
    private static final Logger log = LoggerFactory.getLogger(RepositoryConfigStorageServiceImpl.class);
    private final RepositoryConfigDao repositoryConfigDao;
    private final RepositoryPersistingService repositoryPersistingService;

    @Autowired
    public RepositoryConfigStorageServiceImpl(RepositoryConfigDao repositoryConfigDao,
            RepositoryPersistingService repositoryPersistingService) {
        this.repositoryConfigDao = repositoryConfigDao;
        this.repositoryPersistingService = repositoryPersistingService;
    }

    @Override
    public DbRepositoryConfig findRepositoryByKey(String repoKey) {
        try {
            Optional<DbRepositoryConfig> dbRepositoryConfigOptional = repositoryConfigDao.findRepositoryByKey(repoKey);
            if (dbRepositoryConfigOptional.isPresent()) {
                DbRepositoryConfig dbRepositoryConfig = dbRepositoryConfigOptional.get();
                // in case release bundles handler is not enabled and the repository type is not release bundle v1,
                // then filter out all the repositories with package type "ReleaseBundles".
                // note: this filter is for supporting bundle-handler beta version, once we are moving to GA, this
                // filter will be deleted - [RTDEV-29157].
                if (isReleaseBundlesV2(dbRepositoryConfig.getType(), dbRepositoryConfig.getPackageType())) {
                    if (BundleConfigurationUtil.isReleaseBundleHandlerEnabled()) {
                        return dbRepositoryConfig;
                    }
                    log.info("Unable to find release bundle repository : '{}'", repoKey);
                    return null;
                }
                return dbRepositoryConfig;
            }
        } catch (SQLException e) {
            throw new StorageException("Unable to find repository : " + repoKey);
        }
        log.info("Unable to find repository : '{}'", repoKey);
        return null;
    }

    @Override
    public List<DbRepositoryConfig> getAllRepositories() {
        try {
            boolean isBundlePackageTypeEnabled = BundleConfigurationUtil.isReleaseBundleHandlerEnabled();
            // in case release bundles handler is not enabled and the repository type is not release bundle v1,
            // then filter out all the repositories with package type "ReleaseBundles".
            // note: this filter is for supporting bundle-handler beta version, once we are moving to GA, this filter
            // will be deleted - [RTDEV-29157].
            return isBundlePackageTypeEnabled ? repositoryConfigDao.getAll() :
                    repositoryConfigDao.getAll().stream()
                            .filter(repo -> !isReleaseBundlesV2(repo.getType(), repo.getPackageType()))
                            .toList();
        } catch (SQLException e) {
            throw new StorageException("Unable to fetch all repositories", e);
        }
    }

    @Override
    public Set<DbRepositoryConfig> findRepositoryConfigsByKeys(Set<String> repoKeys) {
        try {
            boolean isBundlePackageTypeEnabled = BundleConfigurationUtil.isReleaseBundleHandlerEnabled();
            // in case release bundles handler is not enabled and the repository type is not release bundle v1,
            // then filter out all the repositories with package type "ReleaseBundles".
            // note: this filter is for supporting bundle-handler beta version, once we are moving to GA, this filter
            // will be deleted - [RTDEV-29157].
            return isBundlePackageTypeEnabled ? repositoryConfigDao.findByKeysWithPagination(repoKeys) :
                    repositoryConfigDao.findByKeysWithPagination(repoKeys).stream()
                            .filter(repo -> !isReleaseBundlesV2(repo.getType(), repo.getPackageType()))
                            .collect(Collectors.toSet());
        } catch (SQLException e) {
            throw new StorageException("Unable to fetch repositories", e);
        }
    }

    @Override
    public void insertRepositoryConfig(DbRepositoryConfig dbRepositoryConfig) {
        try {
            log.info("About to insert new repository configuration '{}'", dbRepositoryConfig.getKey());
            RepositoryConfigStorageUtils.assertNonZeroModificationDates(Set.of(dbRepositoryConfig));
            repositoryConfigDao.insert(dbRepositoryConfig);
            repositoryPersistingService.markDirty();
        } catch (SQLException e) {
            if (findRepositoryByKey(dbRepositoryConfig.getKey()) != null) {
                log.warn("Repository with key '{}' already exists.", dbRepositoryConfig.getKey());
                return;
            }
            throw new StorageException(
                    "Unable to insert new configuration for repository : " + dbRepositoryConfig.getKey());
        }
    }

    @Override
    public void bulkInsertRepositoryConfigs(Set<DbRepositoryConfig> dbRepositoryConfigs) {
        try {
            if (!CollectionUtils.isEmpty(dbRepositoryConfigs)) {
                RepositoryConfigStorageUtils.assertNonZeroModificationDates(dbRepositoryConfigs);
                log.info("About to insert {} new repository configurations", dbRepositoryConfigs.size());
                repositoryConfigDao.bulkInsert(dbRepositoryConfigs.stream()
                        .sorted(Comparator.comparing(DbRepositoryConfig::getKey))
                        .collect(Collectors.toList()));
                repositoryPersistingService.markDirty();
            }
        } catch (SQLException e) {
            throw new StorageException("Failed to insert new repository configurations", e);
        }
    }

    @Override
    public void bulkDeleteRepositoryConfigs(Set<String> repoKeys) {
        try {
            if (!CollectionUtils.isEmpty(repoKeys)) {
                log.debug("About to delete configuration blob for repository: '{}'", repoKeys.size());
                repositoryConfigDao.bulkDelete(repoKeys.stream().sorted(Comparator.comparing(Function.identity()))
                        .collect(Collectors.toList()));
                repositoryPersistingService.markDirty();
            }
        } catch (SQLException e) {
            throw new StorageException("Unable to delete configuration for repository: ", e);
        }
    }

    @Override
    public void updateRepositoryConfig(DbRepositoryConfig dbRepositoryConfig) {
        try {
            log.debug("About to update configuration for repository: '{}'", dbRepositoryConfig.getKey());
            RepositoryConfigStorageUtils.assertNonZeroModificationDates(Set.of(dbRepositoryConfig));
            repositoryConfigDao.update(dbRepositoryConfig);
            repositoryPersistingService.markDirty();
        } catch (SQLException e) {
            throw new StorageException(
                    "Unable to update configuration for repository: " + dbRepositoryConfig.getKey(), e);
        }
    }

    @Override
    public void bulkUpdateRepositoryConfigs(Set<DbRepositoryConfig> dbRepositoryConfigs) {
        try {
            if (!CollectionUtils.isEmpty(dbRepositoryConfigs)) {
                RepositoryConfigStorageUtils.assertNonZeroModificationDates(dbRepositoryConfigs);
                log.info("About to update {} repository configurations", dbRepositoryConfigs.size());
                repositoryConfigDao.bulkUpdate(
                        dbRepositoryConfigs.stream().sorted(Comparator.comparing(DbRepositoryConfig::getKey))
                                .collect(Collectors.toList()));
                repositoryPersistingService.markDirty();
            }
        } catch (SQLException e) {
            throw new StorageException("Failed to update repository configurations", e);
        }
    }

    @Override
    public void delete(String repoKey) {
        try {
            if (repositoryConfigDao.delete(repoKey) > 0) {
                log.debug("Repository '{}' was deleted.", repoKey);
                repositoryPersistingService.markDirty();
            }
        } catch (SQLException e) {
            throw new StorageException("Unable to delete repository " + repoKey, e);
        }
    }

    @Override
    public boolean exists(String repoKey) {
        try {
            return repositoryConfigDao.exists(repoKey);
        } catch (SQLException e) {
            throw new StorageException("Unable to validate repository's " + repoKey + "existence", e);
        }
    }

    @Override
    public int countAll() {
        try {
            // in case release bundles handler is not enabled and the repository type is not release bundle v1,
            // then filter out all the repositories with package type "ReleaseBundles".
            // note: this filter is for supporting bundle-handler beta version, once we are moving to GA, this filter
            // will be deleted - [RTDEV-29157].
            if (BundleConfigurationUtil.isReleaseBundleHandlerEnabled()) {
                return repositoryConfigDao.countAll();
            }
            int releaseBundlesCount = repositoryConfigDao.getAllReleaseBundlesV2RepoKeys().size();
            return repositoryConfigDao.countAll() - releaseBundlesCount;
        } catch (SQLException e) {
            throw new StorageException("Unable to count all repositories", e);
        }
    }

    @Override
    public int deleteAll() {
        try {
            int numDeleted = repositoryConfigDao.deleteAll();
            repositoryPersistingService.markDirty();
            return numDeleted;
        } catch (SQLException e) {
            throw new StorageException("Unable to delete all repositories", e);
        }
    }

    @Override
    public void flushBulkChanges(Set<DbRepositoryConfig> repositoriesToCreate,
            Set<DbRepositoryConfig> repositoriesToUpdate, Set<String> repositoriesToDelete) {
        if (TransactionSynchronizationManager.isSynchronizationActive()) {
            throw new StorageException("Active transaction is not allowed here");
        }
        RepositoryConfigStorageService repositoryConfigStorageService = getAdvisedMe();
        repositoryConfigStorageService.bulkInsertRepositoryConfigs(repositoriesToCreate);
        repositoryConfigStorageService.bulkUpdateRepositoryConfigs(repositoriesToUpdate);
        repositoryConfigStorageService.bulkDeleteRepositoryConfigs(repositoriesToDelete);
        repositoryPersistingService.markDirty();
    }

    @Override
    public Collection<DbRepositoryConfig> findRepositoriesByDetailsType(String repoDetailsType) {
        try {
            return repositoryConfigDao.findRepositoriesByDetailsType(repoDetailsType);
        } catch (SQLException e) {
            throw new StorageException("Unable to fetch repositories", e);
        }
    }

    @Override
    public Set<String> findKeysFromSet(Set<String> repoKeys) {
        try {
            return repositoryConfigDao.findKeysFromSet(repoKeys);
        } catch (SQLException e) {
            throw new StorageException("Unable to check existing repositories", e);
        }
    }

    private RepositoryConfigStorageService getAdvisedMe() {
        return ContextHelper.get().beanForType(RepositoryConfigStorageService.class);
    }

    @Override
    public Set<DbRepositoryConfig> getAllBasicRepoConfigs() {
        try {
            boolean isBundlePackageTypeEnabled = BundleConfigurationUtil.isReleaseBundleHandlerEnabled();
            // in case release bundles handler is not enabled and the repository type is not release bundle v1,
            // then filter out all the repositories with package type "ReleaseBundles".
            // note: this filter is for supporting bundle-handler beta version, once we are moving to GA, this filter
            // will be deleted - [RTDEV-29157].
            return isBundlePackageTypeEnabled ? repositoryConfigDao.getAllBasicRepoConfigs() :
                    repositoryConfigDao.getAllBasicRepoConfigs().stream()
                            .filter(repo -> !isReleaseBundlesV2(repo.getType(), repo.getPackageType()))
                            .collect(Collectors.toSet());
        } catch (SQLException e) {
            throw new StorageException("Unable to get all repositories keys", e);
        }
    }

    @Override
    public Set<String> getAllRepoKeys() {
        try {
            // in case release bundles handler is not enabled and the repository type is not release bundle v1,
            // then filter out all the repositories with package type "ReleaseBundles".
            // note: this filter is for supporting bundle-handler beta version, once we are moving to GA, this filter
            // will be deleted - [RTDEV-29157].
            boolean isBundlePackageTypeEnabled = BundleConfigurationUtil.isReleaseBundleHandlerEnabled();
            if (isBundlePackageTypeEnabled) {
                return repositoryConfigDao.getAllRepoKeys();
            }
            Set<String> releaseBundlesRepoKeys = repositoryConfigDao.getAllReleaseBundlesV2RepoKeys();
            return repositoryConfigDao.getAllRepoKeys().stream()
                    .filter(repoKey -> !releaseBundlesRepoKeys.contains(repoKey))
                    .collect(Collectors.toSet());
        } catch (SQLException e) {
            throw new StorageException("Unable to get all repositories keys", e);
        }
    }


    @NonNull
    @Override
    public Set<String> findAllMissingRepoRootPaths() {
        try {
            // in case release bundles handler is not enabled and the repository type is not release bundle v1,
            // then filter out all the repositories with package type "ReleaseBundles".
            // note: this filter is for supporting bundle-handler beta version, once we are moving to GA, this filter
            // code will be deleted - [RTDEV-29157].
            boolean isBundlePackageTypeEnabled = BundleConfigurationUtil.isReleaseBundleHandlerEnabled();
            if (isBundlePackageTypeEnabled) {
                return repositoryConfigDao.findAllReposThatDoNotExistInNodes();
            }
            Set<String> releaseBundlesRepositories = repositoryConfigDao.getAllReleaseBundlesV2RepoKeys();
            return repositoryConfigDao.findAllReposThatDoNotExistInNodes().stream()
                    .filter(repo -> !releaseBundlesRepositories.contains(repo) ||
                            !releaseBundlesRepositories.contains(repo + "-cache"))
                    .collect(Collectors.toSet());
        } catch (SQLException e) {
            throw new StorageException("Unable to find all repositories that do not exist in the nodes table", e);
        }
    }

    private boolean isReleaseBundlesV2(@NonNull String repoType, @NonNull String packageType) {
        return !RELEASE_BUNDLE.typeNameLowercase().equals(repoType) && ReleaseBundles.getType().equals(packageType);
    }
}