package org.artifactory.storage.db.repo.itest;

import lombok.AllArgsConstructor;
import lombok.Data;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.storage.db.fs.dao.NodesDao;
import org.artifactory.storage.db.fs.entity.Node;
import org.artifactory.storage.db.fs.entity.NodeBuilder;
import org.artifactory.storage.db.itest.DbBaseTest;
import org.artifactory.storage.db.repo.dao.RepositoryConfigDao;
import org.artifactory.storage.db.repo.entity.DbRepositoryConfig;
import org.artifactory.storage.db.repo.service.mapper.RepositoryConfigMapperImpl;
import org.jetbrains.annotations.NotNull;
import org.jfrog.common.JsonUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Optional;
import java.util.Set;

import static org.artifactory.model.LocalCacheRepoConfig.PATH_SUFFIX;
import static org.artifactory.repo.RepoDetailsType.RELEASE_BUNDLE;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertTrue;

/**
 * @author Tamar Rosen
 */
@Test
public class RepositoryConfigDaoTest extends DbBaseTest {
    @Autowired
    private RepositoryConfigDao repositoryConfigDao;
    @Autowired
    private NodesDao nodesDao;

    RepositoryConfigMapperImpl repositoryConfigMapper;

    @BeforeMethod
    public void setUp() throws SQLException {
        repositoryConfigMapper = new RepositoryConfigMapperImpl();
    }

    @AfterMethod
    public void tearDown() throws SQLException {
        repositoryConfigDao.deleteAll();
    }

    public void testBulkDelete() throws SQLException {
        testBulkInsertRepositoryConfig();
        List<String> repoKeysToDelete = new ArrayList<>();
        for (int i = 0; i < 5; i++) {
            repoKeysToDelete.add("key" + i);
        }

        repositoryConfigDao.bulkDelete(repoKeysToDelete);
        List<DbRepositoryConfig> retrievedRepoConfigs = repositoryConfigDao.getAll();
        assertEquals(retrievedRepoConfigs.size(), 5);
    }

    public void testBulkInsertRepositoryConfig() throws SQLException {
        List<DbRepositoryConfig> repoConfigs = getDbRepositoryConfigs();
        repositoryConfigDao.bulkInsert(repoConfigs);
        List<DbRepositoryConfig> retrievedRepoConfigs = repositoryConfigDao.getAll();
        assertEquals(repoConfigs.size(), retrievedRepoConfigs.size());
        for (int i = 0; i < 10; i++) {
            EqualsBuilder.reflectionEquals(repoConfigs.get(i), retrievedRepoConfigs.get(i));
        }
    }

    @NotNull
    private List<DbRepositoryConfig> getDbRepositoryConfigs() {
        List<DbRepositoryConfig> repoConfigs = new ArrayList<>();
        for (int i = 0; i < 10; i++) {
            repoConfigs.add(new DbRepositoryConfig("key" + i, "type" + i,
                    RepoType.values()[i].getType(), i, i, i, null));
        }
        return repoConfigs;
    }

    public void findRepositoryConfigByKey() throws SQLException {
        List<DbRepositoryConfig> repoConfigs = new ArrayList<>();
        DbRepositoryConfig repoToRetrieve = new DbRepositoryConfig("key-to-retrieve", "local",
                RepoType.Gradle.getType(), 1, 2, 3, null);
        repoConfigs.add(repoToRetrieve);
        for (int i = 0; i < 10; i++) {
            repoConfigs.add(new DbRepositoryConfig("key" + i, "type" + i,
                    RepoType.values()[i].getType(), i, i, i, null));
        }
        repositoryConfigDao.bulkInsert(repoConfigs);
        Optional<DbRepositoryConfig> retrievedRepoConfig = repositoryConfigDao.findRepositoryByKey("key-to-retrieve");
        assertTrue(retrievedRepoConfig.isPresent());
        EqualsBuilder.reflectionEquals(retrievedRepoConfig, retrievedRepoConfig);
    }

    public void testInsertRepositoryConfig() throws SQLException {
        LocalRepoConfig repositoryConfig = LocalRepoConfig.create("my-repo", RepoType.Generic);
        repositoryConfig.getBaseConfig().setModelVersion(0);
        repositoryConfig.getPackageTypeConfig().put("modelVersion", "0");
        DbRepositoryConfig insertMe =
                repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);

        DbRepositoryConfig repoConfig = new DbRepositoryConfig("my-repo", "local", RepoType.Maven.getType(),
                1, 2, 3,
                JsonUtils.getInstance().valueToByteArray(new TestBlob(1)));
        repositoryConfigDao.insert(insertMe);
        Optional<DbRepositoryConfig> retrievedRepoConfig = repositoryConfigDao.findRepositoryByKey("my-repo");
        assertTrue(retrievedRepoConfig.isPresent());
        EqualsBuilder.reflectionEquals(retrievedRepoConfig, repoConfig);
    }

    public void testUpdateRepositoryConfigBlob() throws SQLException {
        // insert repo
        DbRepositoryConfig repoConfig =
                new DbRepositoryConfig("my-repo", "local", RepoType.Maven.getType(), 1, 2,
                        3, JsonUtils.getInstance().valueToByteArray(new TestBlob(1)));
        repositoryConfigDao.insert(repoConfig);

        // get repo
        Optional<DbRepositoryConfig> optionalRetrievedRepoConfig = repositoryConfigDao.findRepositoryByKey("my-repo");
        assertTrue(optionalRetrievedRepoConfig.isPresent());
        DbRepositoryConfig retrievedRepoConfig = optionalRetrievedRepoConfig.get();
        RepositoryConfigDaoTest.TestBlob returnedTestBlob =
                JsonUtils.getInstance().readValue(retrievedRepoConfig.getRepositoryInfoBlob(),
                        RepositoryConfigDaoTest.TestBlob.class);
        long num = returnedTestBlob.getNum();
        assertEquals(num, 1L);

        // update config blob
        returnedTestBlob.setNum(2L);
        retrievedRepoConfig.setRepositoryInfoBlob(JsonUtils.getInstance().valueToByteArray(returnedTestBlob));
        repositoryConfigDao.update(retrievedRepoConfig);

        // get repo and validate config blob was updated
        Optional<DbRepositoryConfig> optionalUpdatedRepoConfig = repositoryConfigDao.findRepositoryByKey("my-repo");
        assertTrue(optionalUpdatedRepoConfig.isPresent());
        DbRepositoryConfig updatedRepoConfig = optionalUpdatedRepoConfig.get();
        RepositoryConfigDaoTest.TestBlob updatedTestBlob =
                JsonUtils.getInstance().readValue(updatedRepoConfig.getRepositoryInfoBlob(),
                        RepositoryConfigDaoTest.TestBlob.class);
        assertEquals(updatedTestBlob.getNum(), 2L);
    }

    public void testDeleteByKey() throws SQLException {
        repositoryConfigDao.insert(new DbRepositoryConfig("delete-me", "local", RepoType.YUM.getType(), 1,
                2, 3, null));
        int deleted = repositoryConfigDao.delete("does-not-exist");
        assertEquals(deleted, 0);

        deleted = repositoryConfigDao.delete("delete-me");
        assertEquals(deleted, 1);
    }

    public void testExists() throws SQLException {
        DbRepositoryConfig repoConfig = new DbRepositoryConfig("my-repo", "local", RepoType.Maven.getType(),
                1, 2, 3,
                JsonUtils.getInstance().valueToByteArray(new TestBlob(1)));
        repositoryConfigDao.insert(repoConfig);
        assertTrue(repositoryConfigDao.exists("my-repo"));
        assertFalse(repositoryConfigDao.exists("does-not-exist"));
    }

    public void testBulkUpdate() throws SQLException {
        // insert repos
        List<DbRepositoryConfig> repoConfigs = new ArrayList<>();
        for (int i = 0; i < 10; i++) {
            repoConfigs.add(new DbRepositoryConfig("key" + i, "type" + i,
                    RepoType.values()[i].getType(), i, i, i,
                    JsonUtils.getInstance().valueToByteArray(new TestBlob(i))));
        }
        repositoryConfigDao.bulkInsert(repoConfigs);

        // get repos
        List<DbRepositoryConfig> reposToUpdate = repositoryConfigDao.getAll();
        for (int i = 0; i < reposToUpdate.size(); i++) {
            RepositoryConfigDaoTest.TestBlob returnedTestBlob =
                    JsonUtils.getInstance().readValue(reposToUpdate.get(i).getRepositoryInfoBlob(),
                            RepositoryConfigDaoTest.TestBlob.class);
            assertEquals(returnedTestBlob.getNum(), i);

            // update repos
            returnedTestBlob.setNum(i + 1);
            reposToUpdate.get(i)
                    .setRepositoryInfoBlob(JsonUtils.getInstance().valueToByteArray(returnedTestBlob));
        }

        repositoryConfigDao.bulkUpdate(reposToUpdate);

        // get repos and validate config blob was updated
        List<DbRepositoryConfig> retrievedRepoConfigs = repositoryConfigDao.getAll();
        for (int i = 0; i < retrievedRepoConfigs.size(); i++) {
            RepositoryConfigDaoTest.TestBlob returnedTestBlob =
                    JsonUtils.getInstance().readValue(retrievedRepoConfigs.get(i).getRepositoryInfoBlob(),
                            RepositoryConfigDaoTest.TestBlob.class);
            assertEquals(returnedTestBlob.getNum(), i + 1);
        }
    }

    public void testGetConfigsByKeys() throws SQLException {
        List<DbRepositoryConfig> repoConfigs = new ArrayList<>();
        for (int i = 0; i < 10; i++) {
            repoConfigs.add(new DbRepositoryConfig("key" + i, "type" + i,
                    RepoType.values()[i].getType(), i, i, i,
                    JsonUtils.getInstance().valueToByteArray(new TestBlob(i))));
        }
        repositoryConfigDao.bulkInsert(repoConfigs);

        Set<DbRepositoryConfig> configsByKeys = repositoryConfigDao.findByKeysWithPagination(
                Set.of("key0", "key1", "key2", "key3", "non-existing-repo"));
        Set<DbRepositoryConfig> dbRepositoryConfigs = new HashSet<>(configsByKeys);

        Set<DbRepositoryConfig> expectedRepos = new HashSet<>(repoConfigs.subList(0, 4));
        assertEquals(dbRepositoryConfigs, expectedRepos);
    }

    public void testFindConfigsByDetailsType() throws SQLException {
        List<DbRepositoryConfig> repoConfigs = new ArrayList<>();
        for (int i = 0; i < 3; i++) {
            repoConfigs.add(new DbRepositoryConfig("key" + i, "type" + i,
                    RepoType.values()[i].getType(), i, i, i,
                    JsonUtils.getInstance().valueToByteArray(new TestBlob(i))));
        }
        repositoryConfigDao.bulkInsert(repoConfigs);

        Collection<DbRepositoryConfig> configsByKeys = repositoryConfigDao.findRepositoriesByDetailsType("type1");
        assertEquals(configsByKeys.size(), 1);
        assertEquals(configsByKeys.stream().findFirst().get().getType(), "type1");
    }

    public void testCountAll() throws SQLException {
        assertEquals(repositoryConfigDao.countAll(), 0);

        DbRepositoryConfig repoConfig = new DbRepositoryConfig("my-repo", "local", RepoType.Maven.getType(),
                1L, 2L, 3L,
                JsonUtils.getInstance().valueToByteArray(new TestBlob(1)));
        DbRepositoryConfig repoConfig1 = new DbRepositoryConfig("my-repo1", "remote", RepoType.Conan.getType(),
                10, 20, 30,
                JsonUtils.getInstance().valueToByteArray(new TestBlob(2)));

        repositoryConfigDao.insert(repoConfig);
        repositoryConfigDao.insert(repoConfig1);
        assertEquals(repositoryConfigDao.countAll(), 2);
    }

    public void testGetAllRepoKeys() throws SQLException {
        DbRepositoryConfig repoConfig = new DbRepositoryConfig("all-repo-1", "local", RepoType.Maven.getType(),
                1L, 2L, 3L,
                JsonUtils.getInstance().valueToByteArray(new TestBlob(1)));
        DbRepositoryConfig repoConfig1 = new DbRepositoryConfig("all-repo-2", "remote", RepoType.Conan.getType(),
                10, 20, 30,
                JsonUtils.getInstance().valueToByteArray(new TestBlob(2)));

        repositoryConfigDao.insert(repoConfig);
        repositoryConfigDao.insert(repoConfig1);
        Set<String> repoKeys = repositoryConfigDao.getAllRepoKeys();
        assertEquals(repoKeys.size(), 2);
        assertTrue(CollectionUtils.isEqualCollection(repoKeys, Set.of("all-repo-1", "all-repo-2")));

    }

    public void getAllReleaseBundlesV2RepoKeys() throws SQLException {
        DbRepositoryConfig repoConfig =
                new DbRepositoryConfig("release-bundles-v2", "local", RepoType.ReleaseBundles.getType(),
                        1L, 2L, 3L, JsonUtils.getInstance().valueToByteArray(new TestBlob(1)));
        DbRepositoryConfig repoConfig1 =
                new DbRepositoryConfig("project-release-bundles-v2", "local", RepoType.ReleaseBundles.getType(),
                        10, 20, 30, JsonUtils.getInstance().valueToByteArray(new TestBlob(2)));
        DbRepositoryConfig repoConfig2 =
                new DbRepositoryConfig("release-bundles-v1", RELEASE_BUNDLE.typeNameLowercase(),
                        RepoType.ReleaseBundles.getType(),
                        10, 20, 30, JsonUtils.getInstance().valueToByteArray(new TestBlob(2)));
        DbRepositoryConfig repoConfig3 =
                new DbRepositoryConfig("any-repo", "local", RepoType.Maven.getType(),
                        10, 20, 30, JsonUtils.getInstance().valueToByteArray(new TestBlob(2)));
        DbRepositoryConfig invalidType =
                new DbRepositoryConfig("projectb-release-bundles-v2", RELEASE_BUNDLE.typeNameLowercase(),
                        RepoType.Maven.getType(),
                        10, 20, 30, JsonUtils.getInstance().valueToByteArray(new TestBlob(2)));

        repositoryConfigDao.insert(repoConfig);
        repositoryConfigDao.insert(repoConfig1);
        repositoryConfigDao.insert(repoConfig2);
        repositoryConfigDao.insert(invalidType);
        Set<String> repoKeys = repositoryConfigDao.getAllReleaseBundlesV2RepoKeys();
        assertEquals(repoKeys.size(), 2);
        assertTrue(CollectionUtils.isEqualCollection(repoKeys,
                Set.of("release-bundles-v2", "project-release-bundles-v2")));

    }

    @Test(dataProvider = "existingRepos")
    public void testFindReposThatDontExistInNodes(Set<String> repoConfigs, Set<String> repoNodes, boolean isRemote,
            Collection<String> expectedMissingRepos)
            throws SQLException {
        try {
            // setup
            List<DbRepositoryConfig> dbRepositoryConfigs = new ArrayList<>();
            Iterator<String> repoConfigIter = repoConfigs.iterator();
            for (int i = 0; i < repoConfigs.size(); i++) {
                dbRepositoryConfigs.add(new DbRepositoryConfig(repoConfigIter.next(), isRemote ? "remote" : "local",
                        RepoType.values()[i].getType(), i, i, i, null));
            }
            repositoryConfigDao.bulkInsert(dbRepositoryConfigs);

            ArrayList<Node> nodesToCreate = new ArrayList<>();
            Iterator<String> repoNodesIter = repoNodes.iterator();
            for (int i = 0; i < repoNodes.size(); i++) {
                nodesToCreate.add(new NodeBuilder().nodeId(i + 1).repo(repoNodesIter.next()).build());
            }
            nodesDao.createBulk(nodesToCreate);

            Set<String> nonExistingRepos = repositoryConfigDao.findAllReposThatDoNotExistInNodes();
            assertTrue(nonExistingRepos.containsAll(expectedMissingRepos));
            assertEquals(expectedMissingRepos.size(), nonExistingRepos.size());
        } finally {
            ArtifactoryHome.get().getArtifactoryProperties()
                    .setProperty(ConstantValues.existsRepoChunkSize.getPropertyName(),
                            ConstantValues.existsRepoChunkSize.getDefValue());
            nodesDao.deleteBulk(List.of(1L, 2L, 3L));
            repositoryConfigDao.bulkDelete(new ArrayList<>(repoConfigs));
        }
    }

    @DataProvider
    public Object[][] existingRepos() {
        return new Object[][] {
                {Set.of("notExist", "notExist2"), Set.of(), false, Set.of("notExist", "notExist2")},
                {Set.of("repo1", "repo2", "repo3"), Set.of("repo1", "repo2", "repo3"), false, Collections.emptySet()},
                {Set.of("repo1", "repo2", "repo4"), Set.of("repo4"), false, Set.of("repo1", "repo2")},
                {Set.of("notExistRemote", "remote"), Set.of("remote", "remote" + PATH_SUFFIX), true, Set.of(
                        "notExistRemote", "notExistRemote" + PATH_SUFFIX)},
                {Set.of("notExistRemote1", "notExistRemote2"), Set.of(), true, Set.of("notExistRemote1",
                        "notExistRemote2", "notExistRemote1" + PATH_SUFFIX, "notExistRemote2" + PATH_SUFFIX)},
                {Set.of("remote1", "remote2"), Set.of("remote1", "remote2", "remote1" + PATH_SUFFIX,
                        "remote2" + PATH_SUFFIX), true, Set.of()}
        };
    }

    public void testFindKeysFromSet() throws SQLException {
        assertEquals(repositoryConfigDao.countAll(), 0);

        DbRepositoryConfig repoConfig = new DbRepositoryConfig("my-repo", "local", RepoType.Maven.getType(),
                1L, 2L, 3L,
                JsonUtils.getInstance().valueToByteArray(new TestBlob(1)));
        DbRepositoryConfig repoConfig1 = new DbRepositoryConfig("my-repo1", "remote", RepoType.Conan.getType(),
                10, 20, 30,
                JsonUtils.getInstance().valueToByteArray(new TestBlob(2)));

        repositoryConfigDao.insert(repoConfig);
        repositoryConfigDao.insert(repoConfig1);

        Set<String> keys = repositoryConfigDao.findKeysFromSet(Set.of("my-repo", "my-repo1", "my-repo2"));
        assertEquals(keys.size(), 2);
        assertTrue(keys.contains("my-repo"));
        assertTrue(keys.contains("my-repo1"));
    }

    @Data
    @AllArgsConstructor
    public static class TestBlob {
        private long num;
    }
}