package org.artifactory.webapp.servlet;

import org.apache.http.HttpStatus;
import org.artifactory.api.build.BuildService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.exception.FileExpectedException;
import org.artifactory.api.repo.exception.RepoRejectException;
import org.artifactory.api.request.ArtifactoryResponse;
import org.artifactory.api.request.UploadService;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.request.ArtifactoryRequest;
import org.jfrog.client.util.PathUtils;
import org.junit.jupiter.api.Test;
import org.mockito.MockedStatic;

import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;

import static org.mockito.Mockito.RETURNS_DEEP_STUBS;
import static org.mockito.Mockito.any;
import static org.mockito.Mockito.anyBoolean;
import static org.mockito.Mockito.anyString;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.mockStatic;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author Michael Akushsky
 */
public class RepoFilterTest {

    @Test
    void testFilterWithFileExpectedExceptionOccurs() throws IOException, RepoRejectException, ServletException {
        // RepoFilter object to test
        RepoFilter repoFilterSpy = spy(RepoFilter.class);

        ArtifactoryContext artifactoryContextMock = mock(ArtifactoryContext.class, RETURNS_DEEP_STUBS);
        UploadService uploadServiceMock = mock(UploadService.class);
        BuildService buildServiceMock = mock(BuildService.class);
        MockedStatic<ContextHelper> contextHelperMock = mockStatic(ContextHelper.class);
        contextHelperMock.when(ContextHelper::get).thenReturn(artifactoryContextMock);
        when(artifactoryContextMock.beanForType(UploadService.class)).thenReturn(uploadServiceMock);
        when(artifactoryContextMock.beanForType(BuildService.class)).thenReturn(buildServiceMock);

        // RTFACT-22667 situation preparing
        doThrow(FileExpectedException.class).when(uploadServiceMock)
                .upload(any(ArtifactoryRequest.class), any(ArtifactoryResponse.class));

        // Construct different fake objects
        MockedStatic<RequestUtils> requestUtilsMock = mockStatic(RequestUtils.class);
        requestUtilsMock.when(() -> RequestUtils.isRepoRequest(any(), anyBoolean())).thenReturn(true);
        requestUtilsMock.when(() -> RequestUtils.getServletPathFromRequest(any())).thenReturn("/");

        ArtifactoryHome artifactoryHomeObjectMock = mock(ArtifactoryHome.class, RETURNS_DEEP_STUBS);
        MockedStatic<ArtifactoryHome> artifactoryHomeMock = mockStatic(ArtifactoryHome.class);
        artifactoryHomeMock.when(ArtifactoryHome::get).thenReturn(artifactoryHomeObjectMock);

        MockedStatic<PathUtils> pathUtilsMock = mockStatic(PathUtils.class);
        pathUtilsMock.when(() -> PathUtils.getFirstPathElement(anyString())).thenReturn("/");
        pathUtilsMock.when(() -> PathUtils.splitZipResourcePathIfExist(anyString(), anyBoolean()))
                .thenReturn(new String[]{"_","_"});

        // Construct fake http request and response objects
        HttpServletRequest servletRequestMock = mock(HttpServletRequest.class);
        when(servletRequestMock.getAttribute(anyString())).thenReturn("FAKE");
        when(servletRequestMock.getMethod()).thenReturn("put");
        when(servletRequestMock.getRequestURI()).thenReturn("FAKE");
        HttpServletResponse servletResponseMock = mock(HttpServletResponse.class, RETURNS_DEEP_STUBS);
        FilterChain filterChainMock = mock(FilterChain.class);

        // Wrap them to artifactory ones
        ArtifactoryResponse artifactoryResponse = new HttpArtifactoryResponse(servletResponseMock);
        ArtifactoryResponse artifactoryResponseSpy = spy(artifactoryResponse);
        doReturn(artifactoryResponseSpy).when(repoFilterSpy).createArtifactoryResponce(any());

        // And the rest - fake input for filter
        FilterConfig filterConfigMock = mock(FilterConfig.class);
        ServletContext servletContextMock = mock(ServletContext.class);
        when(filterConfigMock.getServletContext()).thenReturn(servletContextMock);
        when(filterConfigMock.getFilterName()).thenReturn("FAKE");
        when(servletContextMock.getAttribute(anyString())).thenReturn(null);

        // Now let's execute
        repoFilterSpy.init(filterConfigMock);
        repoFilterSpy.doFilter(servletRequestMock, servletResponseMock, filterChainMock);

        // And verify HttpStatus
        verify(artifactoryResponseSpy).sendError(eq(HttpStatus.SC_CONFLICT), any(), any());
    }
}