package org.artifactory.ui.rest.service.builds.buildsinfo.tabs.permissions;

import lombok.AllArgsConstructor;
import lombok.Data;
import lombok.NonNull;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.api.build.BuildService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;

import javax.annotation.Nullable;
import javax.servlet.http.HttpServletResponse;

import static org.artifactory.build.BuildInfoUtils.formatBuildTime;

/**
 * @author Yevdo Abramov
 * Created on 28/10/2020
 */
abstract class GetBuildEffectivePermissionServiceBase implements RestService {
    protected final AuthorizationService authService;
    protected final UserGroupService userGroupService;
    protected final BuildService buildService;

    GetBuildEffectivePermissionServiceBase(@NonNull AuthorizationService authService,
            @NonNull UserGroupService userGroupService, @NonNull BuildService buildService) {
        this.authService = authService;
        this.userGroupService = userGroupService;
        this.buildService = buildService;
    }

    @Nullable
    BuildInfo getAndValidateRequestDataByPathParams(@NonNull ArtifactoryRestRequest request,
            @NonNull RestResponse response) {
        String buildName = request.getPathParamByKey("name");
        String buildNumber = request.getPathParamByKey("number");
        String buildStarted = request.getPathParamByKey("date");
        String buildRepo = buildService.normaliseBuildRepoName(request.getQueryParamByKey("buildRepo"));
        return getAndValidateRequestData(response, buildName, buildNumber, buildStarted, buildRepo);
    }

    @Nullable
    BuildInfo getAndValidateRequestDataByQueryParams(@NonNull ArtifactoryRestRequest request,
            @NonNull RestResponse response) {
        String buildName = request.getQueryParamByKey("buildName");
        String buildNumber = request.getQueryParamByKey("buildNumber");
        String buildStarted = request.getQueryParamByKey("buildDate");
        String buildRepo = buildService.normaliseBuildRepoName(request.getQueryParamByKey("buildRepo"));
        return getAndValidateRequestData(response, buildName, buildNumber, buildStarted, buildRepo);
    }

    @Nullable
    private BuildInfo getAndValidateRequestData(@NonNull RestResponse response, @Nullable String buildName,
            @Nullable String buildNumber, @Nullable String buildStarted, @NonNull String buildRepo) {
        if (StringUtils.isEmpty(buildName) || StringUtils.isEmpty(buildNumber) || StringUtils.isEmpty(buildStarted)) {
            response.error("One or more of the following parameters is missing: buildName, buildNumber, buildDate")
                    .buildResponse();
            return null;
        }

        String formattedBuildStarted;
        try {
            formattedBuildStarted = formatBuildTime(buildStarted);
        } catch (NumberFormatException e) {
            response.error("Invalid buildDate format").buildResponse();
            return null;
        }
        if (!authService.canManageBuild(buildName, buildNumber, formattedBuildStarted, buildRepo)) {
            response.responseCode(HttpServletResponse.SC_FORBIDDEN).buildResponse();
            return null;
        }

        return new BuildInfo(buildName, buildNumber, formattedBuildStarted, buildRepo);
    }

    @Data
    @AllArgsConstructor
    static class BuildInfo {
        private String name;
        private String number;
        private String started;
        private String buildRepo;
    }
}
