/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.builds.search;

import org.apache.commons.lang3.StringUtils;
import org.artifactory.api.build.BuildService;
import org.artifactory.api.rest.build.BuildIdInfo;
import org.artifactory.api.rest.build.BuildIdResponse;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.build.BuildId;
import org.artifactory.common.ConstantValues;
import org.artifactory.projects.ProjectsService;
import org.artifactory.projects.ProjectsUtils;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.joda.time.format.ISODateTimeFormat;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.annotation.Nullable;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;

/**
 * @author Rotem Kfir
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class SearchBuildIDsByNameService implements RestService {

    private BuildService buildService;
    private AuthorizationService authService;

    @Autowired
    public void setBuildService(BuildService buildService) {
        this.buildService = buildService;
    }

    @Autowired
    public void setAuthService(AuthorizationService authService) {
        this.authService = authService;
    }

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        BuildsSearchFilter searchBuildFilter = createBuildsSearchFilter(request);
        List<BuildId> builds = buildService.findBuildIDsByNameCaseInsensitive(searchBuildFilter);
        BuildIdResponse buildIdResponse = getBuildIdResponse(builds);
        response.iModel(buildIdResponse);
    }

    private BuildsSearchFilter createBuildsSearchFilter(ArtifactoryRestRequest request) {
        String buildName = request.getQueryParamByKey("name");
        String numOfRows = request.getQueryParamByKey("num_of_rows");
        String buildRepo = request.getQueryParamByKey("buildRepo");
        String orderBy = request.getQueryParamByKey("order_by");
        String direction = request.getQueryParamByKey("direction");
        String beforeStr = request.getQueryParamByKey("before");
        String afterStr = request.getQueryParamByKey("after");
        long before = StringUtils.isNotBlank(beforeStr) ? Long.parseLong(beforeStr) : 0;
        long after = StringUtils.isNotBlank(afterStr) ? Long.parseLong(afterStr) : 0;
        long limit = StringUtils.isNotBlank(numOfRows) ? Long.parseLong(numOfRows) : Long.MAX_VALUE;
        String projectKey = request.getQueryParamByKey(ProjectsService.PROJECT_KEY);
        String finalBuildRepo = getBuildRepoForSearch(buildRepo, projectKey);
        return BuildsSearchFilter.builder()
                .name(buildName)
                .buildRepo(finalBuildRepo)
                .projectKey(projectKey)
                .before(before)
                .after(after)
                .limit(limit)
                .orderBy(orderBy)
                .sortOrder(direction)
                .build();
    }

    private BuildIdResponse getBuildIdResponse(List<BuildId> builds) {
        List<BuildIdInfo> generalBuildInfoList = builds.stream()
                .map(this::getBuildIdInfo)
                .collect(Collectors.toList());
        BuildIdResponse buildIdResponse = new BuildIdResponse();
        generalBuildInfoList.forEach(buildIdResponse::add);
        return buildIdResponse;
    }

    private BuildIdInfo getBuildIdInfo(BuildId build) {
        Date startDate = build.getStartedDate();
        long startTime = startDate != null ? startDate.getTime() : 0;
        return BuildIdInfo.builder()
                .buildName(build.getName())
                .lastBuildTime(ISODateTimeFormat.dateTime().print(startTime))
                .buildRepo(build.getBuildRepo())
                .time(startTime)
                .lastBuildNumber(build.getNumber())
                .canDelete(ConstantValues.buildUiSkipDeletePermissionCheck.getBoolean() ||
                        authService.canDeleteBuild(build.getName(), build.getNumber(), build.getStarted(),
                                build.getBuildRepo())).build();
    }

    /**
     * @return Null - for searching in all build-repos
     * Default build-info repo - For Unassigned \ Default projectKey
     * Any other buildRepo - For explicit non-null buildRepo or non default projectKey
     */
    @Nullable
    String getBuildRepoForSearch(@Nullable String buildRepo, @Nullable String projectKey) {
        if (StringUtils.isNotEmpty(buildRepo)) {
            return buildService.normaliseBuildRepoName(buildRepo);
        } else if (StringUtils.isNotEmpty(projectKey)) {
            if (ProjectsUtils.isUnassignedContext(projectKey)) {
                return buildService.getBuildInfoRepoForDefaultProject();
            } else {
                return buildService.getBuildInfoRepoKeyByProject(projectKey);
            }
        }
        return null;
    }
}
