/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.utils.repositories;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.Lists;
import lombok.NonNull;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.LocalCacheRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RealRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.descriptor.repo.RepoLayout;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.model.FederatedRepoConfig;
import org.artifactory.model.LocalConfig;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.RepoConfig;
import org.artifactory.model.VirtualRepoConfig;
import org.artifactory.projects.ProjectsService;
import org.artifactory.projects.ProjectsUtils;
import org.artifactory.repo.InternalRepoPathFactory;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.system.SystemRepos;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.security.ArtifactoryPermission;
import org.artifactory.ui.rest.model.artifacts.search.packagesearch.criteria.PackageSearchCriteria;
import org.artifactory.ui.rest.model.utils.repositories.RepoKeyType;
import org.artifactory.ui.rest.model.utils.repositories.RepositoriesData;
import org.artifactory.ui.utils.RequestUtils;
import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

import static org.artifactory.descriptor.repo.RepoType.ReleaseBundles;
import static org.artifactory.repo.RepoDetailsType.FEDERATED;
import static org.artifactory.repo.RepoDetailsType.LOCAL;
import static org.artifactory.repo.RepoDetailsType.REMOTE;
import static org.artifactory.repo.system.SystemRepos.SUPPORT_BUNDLE_REPO_KEY;
import static org.artifactory.repo.system.SystemRepos.USAGE_REPO_KEY;
import static org.artifactory.repo.system.SystemRepos.isInTransitRepo;
import static org.artifactory.repo.system.SystemRepos.isObservabilityRepo;

/**
 * @author Chen Keinan
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class GetAllRepositoriesService implements RestService {

    private AuthorizationService authorizationService;
    private RepositoryService repoService;
    private CentralConfigService centralConfigService;
    private ProjectsService projectsService;

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        String projectKey = RequestUtils.getProjectKeyFromRequest(request);
        boolean userOnly = Boolean.parseBoolean(request.getQueryParamByKey("user"));
        boolean deploy = Boolean.parseBoolean(request.getQueryParamByKey("deploy"));
        boolean localOnly = Boolean.parseBoolean(request.getQueryParamByKey("local"));
        boolean searchOnly = Boolean.parseBoolean(request.getQueryParamByKey("search"));
        boolean backupOnly = Boolean.parseBoolean(request.getQueryParamByKey("backup"));
        boolean packageSearchOnly = Boolean.parseBoolean(request.getQueryParamByKey("packageSearch"));
        boolean localRemoteOnly = Boolean.parseBoolean(request.getQueryParamByKey("all"));
        boolean permission = Boolean.parseBoolean(request.getQueryParamByKey("permission"));

        RepositoriesData repositoriesData = new RepositoriesData();
        if (searchOnly) { //Called from the search screen
            searchRepoData(response, repositoriesData, projectKey);
        } else if (packageSearchOnly) {
            packageSearchRepoData(response, repositoriesData, projectKey);
        } else if (backupOnly) { //Called from the Backups and Repositories Import/Export screens
            repositoriesData.setRepoList(getRepoData(localOnly, false, projectKey));
            response.iModel(repositoriesData);
        } else if (localRemoteOnly) {
            List<RepoKeyType> allRepositoriesData = getAllRepositoriesData(projectKey);
            repositoriesData.setRepoTypesList(allRepositoriesData);
            response.iModelList(allRepositoriesData);
        } else if (permission) { //Called from the Permissions screen
            List<RepoKeyType> allRepositoriesData = getPermissionRepoData(projectKey);
            repositoriesData.setRepoTypesList(allRepositoriesData);
            response.iModel(repositoriesData);
        } else if (deploy) {
            getRepositoriesDataForDeploy(repositoriesData, projectKey);
            response.iModel(repositoriesData);
        } else {
            /// get repo data for local or user
            getRepositoriesData(userOnly, localOnly, repositoriesData, projectKey);
            response.iModel(repositoriesData);
        }
    }

    /**
     * get repository list for search screen
     *
     * @param artifactoryResponse - encapsulated data related to response
     * @param repositoriesData    - repository data
     */
    private void searchRepoData(RestResponse artifactoryResponse, RepositoriesData repositoriesData,
            @NonNull String projectKey) {
        List<RepoKeyType> orderedRepoKeys = getOrderedRepoKeys(projectKey);
        repositoriesData.setRepoTypesList(orderedRepoKeys);
        artifactoryResponse.iModel(repositoriesData);
    }

    /**
     * get repository list for package search screen
     *
     * @param artifactoryResponse - encapsulated data related to response
     * @param repositoriesData    - repository data
     */
    private void packageSearchRepoData(RestResponse artifactoryResponse, RepositoriesData repositoriesData,
            @NonNull String projectKey) {
        List<RepoKeyType> orderedRepoKeys = getPackageSearchRepos(projectKey);
        repositoriesData.setRepoTypesList(orderedRepoKeys);
        artifactoryResponse.iModel(repositoriesData);
    }

    /**
     * return repositories data for user / local / all
     *
     * @param userOnly  - if true return only user repositories
     * @param localOnly - if true return only local repositories
     */
    private void getRepositoriesData(boolean userOnly, boolean localOnly, RepositoriesData repositoriesData,
            @NonNull String projectKey) {
        if (userOnly) {
            getRepoDataForUser(repositoriesData, projectKey);
        } else {
            List<String> repoData = getRepoData(localOnly, false, projectKey);
            repositoriesData.setRepoList(repoData);
        }
    }

    @VisibleForTesting
    void getRepoDataForUser(RepositoriesData repositoriesData, String projectKey) {
        List<String> repoData = getRepoData(true, true, projectKey);
        List<RepoKeyType> deployRepos = new ArrayList<>();
        // below filters the allowed deploy repository types
        repoData.forEach(repoKey -> {
            if (repoKey.equals(SUPPORT_BUNDLE_REPO_KEY)) {
                deployRepos.add(new RepoKeyType(RepoType.Support, repoKey));
            } else {
                LocalConfig localConfig = repoService.localOrFederatedRepoConfigByKey(repoKey);
                if (localConfig != null && !isObservabilityRepo(repoKey) && !isInTransitRepo(repoKey)) {
                    deployRepos.add(new RepoKeyType(localConfig.getPackageType(), repoKey));
                }
                FederatedRepoConfig federatedRepoConfig = repoService.federatedRepoConfigByKey(repoKey);
                if (federatedRepoConfig != null) {
                    deployRepos.add(new RepoKeyType(federatedRepoConfig.getPackageType(), repoKey));
                }
            }
        });
        if (authorizationService.isAdmin()) {
            repositoriesData.setRepoTypesList(deployRepos);
        } else {
            repositoriesData.setRepoTypesList(getUserRepoForDeploy(deployRepos));
        }
    }

    /**
     * return repositories data for deploy (like for user, but include virtual repos with configured default
     * deployment local repo)
     */
    private void getRepositoriesDataForDeploy(RepositoriesData repositoriesData, @NonNull String projectKey) {
        MutableCentralConfigDescriptor mutableDescriptor = centralConfigService.getMutableDescriptor();
        repositoriesData.setFileUploadMaxSizeMb(mutableDescriptor.getFileUploadMaxSizeMb());

        List<String> repoData;
        repoData = getRepoData(true, false, projectKey);
        Set<String> virtualRepoData = extractRepoKeys(repoService.getVirtualRepoConfigs());

        List<RepoKeyType> deployReposes = new ArrayList<>();
        repoData.forEach(repoKey -> {
            if (!SystemRepos.isInTransitRepo(repoKey)) {
                LocalConfig localConfig = repoService.localOrFederatedRepoConfigByKey(repoKey);
                if (localConfig == null) {
                    localConfig = repoService.federatedRepoConfigByKey(repoKey);
                }
                RepoKeyType repoKeyType = new RepoKeyType(localConfig.getPackageType(), repoKey);
                String repoLayoutRef = localConfig.getBaseConfig().getRepoLayoutRef();
                RepoLayout layout = centralConfigService.getDescriptor().getRepoLayout(repoLayoutRef);
                if (layout != null) {
                    repoKeyType.setLayoutPattern(layout.getArtifactPathPattern());
                    repoKeyType.setLayoutFolderItegRevRegex(layout.getFolderIntegrationRevisionRegExp());
                    repoKeyType.setLayoutFileItegRevRegex(layout.getFileIntegrationRevisionRegExp());
                }
                deployReposes.add(repoKeyType);
            }
        });
        virtualRepoData.forEach(repoKey -> {
            VirtualRepoConfig virtualRepoConfig = repoService.virtualRepoConfigByKey(repoKey);
            if (virtualRepoConfig.getRepoTypeConfig().getDefaultDeploymentRepoRef() != null) {
                RepoKeyType repoKeyType = new RepoKeyType(virtualRepoConfig.getPackageType(), repoKey);
                String repoLayoutRef = virtualRepoConfig.getBaseConfig().getRepoLayoutRef();
                RepoLayout layout = centralConfigService.getDescriptor().getRepoLayout(repoLayoutRef);
                if (layout != null) {
                    repoKeyType.setLayoutPattern(layout.getArtifactPathPattern());
                    repoKeyType.setLayoutFolderItegRevRegex(layout.getFolderIntegrationRevisionRegExp());
                    repoKeyType.setLayoutFileItegRevRegex(layout.getFileIntegrationRevisionRegExp());
                }
                deployReposes.add(repoKeyType);
            }
        });
        if (authorizationService.isAdmin()) {
            repositoriesData.setRepoTypesList(deployReposes);
        } else {
            repositoriesData.setRepoTypesList(getUserRepoForDeploy(deployReposes));
        }
    }

    /**
     * return repo data list from config descriptor
     *
     * @return repo data list
     */
    @VisibleForTesting
    List<String> getRepoData(boolean localOnly, boolean includeSupportBundleRepo, @NonNull String projectKey) {
        List<String> repos = new ArrayList<>();
        List<LocalRepoConfig> localRepoConfigs = repoService.getLocalRepoConfigsIncludingInfoRepos();
        // add remote repositories
        if (!localOnly) {
            List<RemoteRepoConfig> remoteRepoConfigs = repoService.getRemoteRepoConfigs();
            repos.addAll(extractRepoKeys(remoteRepoConfigs));
        }
        if (includeSupportBundleRepo) {
            repos.add(SUPPORT_BUNDLE_REPO_KEY);
        }
        repos.addAll(extractRepoKeys(repoService.getFederatedRepoConfigs()));
        repos.addAll(extractRepoKeys(localRepoConfigs));
        Function<String, Boolean> projectReposFilter = createProjectReposFilter(projectKey);

        return repos.stream()
                .filter(projectReposFilter::apply)
                .filter(r -> authorizationService.canRead(RepoPathFactory.create(r)) ||
                        authorizationService.canDeploy(RepoPathFactory.create(r)))
                .collect(Collectors.toList());
    }

    @NotNull
    private Set<String> extractRepoKeys(List<? extends RepoConfig> repoConfigs) {
        return repoConfigs.stream().map(repoConfig -> repoConfig.getKey()).collect(
                Collectors.toSet());
    }

    /**
     * get list of local repository allowed for this user to deploy
     *
     * @return - list of repositories
     */
    private List<RepoKeyType> getUserRepoForDeploy(List<RepoKeyType> localRepos) {
        List<RepoKeyType> userAuthorizedRepos = Lists.newArrayList();
        localRepos.forEach(repoType -> {
            RepoPath repoPath = InternalRepoPathFactory.create(repoType.getRepoKey(), "");
            if (RepoType.BuildInfo.equals(repoType.getRepoType())) {
                //Special case for build repo since it moves files to the correct path automatically its ok to show
                //deployment anywhere
                if (authorizationService.hasBuildPermission(repoType.getRepoKey(), ArtifactoryPermission.DEPLOY)) {
                    userAuthorizedRepos.add(repoType);
                }
            } else if (!repoType.getRepoKey().endsWith("-cached") && authorizationService.canDeploy(repoPath)) {
                userAuthorizedRepos.add(repoType);
            } else {
                getVirtualDeployRepoIfAvailable(userAuthorizedRepos, repoType);
            }
        });
        return userAuthorizedRepos;
    }

    private void getVirtualDeployRepoIfAvailable(List<RepoKeyType> userAuthorizedRepos, RepoKeyType repoType) {
        VirtualRepoDescriptor virtualRepoDescriptor = repoService.virtualRepoDescriptorByKey(repoType.getRepoKey());
        if (virtualRepoDescriptor != null) {
            LocalRepoDescriptor defaultDeploymentRepo = virtualRepoDescriptor.getDefaultDeploymentRepo();
            if (defaultDeploymentRepo != null) {
                RepoPath repoPath = InternalRepoPathFactory.create(defaultDeploymentRepo.getKey(), "");
                if (authorizationService.canDeploy(repoPath)) {
                    userAuthorizedRepos.add(repoType);
                }
            }
        }
    }

    private List<RepoKeyType> getOrderedRepoKeys(@NonNull String projectKey) {
        List<LocalRepoDescriptor> repoSet = Lists.newArrayList();
        List<LocalRepoDescriptor> localFederatedAndCachedRepoDescriptors =
                repoService.getLocalFederatedAndCachedRepoDescriptorsByProject(projectKey);
        localFederatedAndCachedRepoDescriptors.sort(new LocalAndCachedDescriptorsComparator());
        localFederatedAndCachedRepoDescriptors.stream()
                .filter(descriptor -> authorizationService
                        .canRead(InternalRepoPathFactory.repoRootPath(descriptor.getKey())) &&
                        !SystemRepos.isInTransitRepo(descriptor.getKey()) &&
                        !descriptor.getKey().equalsIgnoreCase(USAGE_REPO_KEY))
                .forEach(repoSet::add);
        List<RepoKeyType> repoKeys = Lists.newArrayList();
        for (LocalRepoDescriptor descriptor : repoSet) {
            String type;
            if (descriptor instanceof LocalCacheRepoDescriptor) {
                type = REMOTE.typeNameLowercase();
            } else if (descriptor instanceof FederatedRepoDescriptor) {
                type = FEDERATED.typeNameLowercase();
            } else {
                type = LOCAL.typeNameLowercase();
            }
            repoKeys.add(new RepoKeyType(type, descriptor.getKey()));
        }
        return repoKeys;
    }

    private List<RepoKeyType> getPackageSearchRepos(@NonNull String projectKey) {
        List<RepoKeyType> repoKeys = Lists.newArrayList();
        List<LocalRepoDescriptor> localCacheAndFederated =
                repoService.getLocalFederatedAndCachedRepoDescriptorsByProject(projectKey);
        List<LocalRepoDescriptor> localAndFederatedOnly =
                repoService.getLocalRepoDescriptorsExcludingInfoReposByProject(projectKey);
        localAndFederatedOnly.addAll(repoService.getFederatedRepoDescriptorsByProject(projectKey));
        localCacheAndFederated.sort(new LocalAndCachedDescriptorsComparator());
        for (PackageSearchCriteria.PackageSearchType type : PackageSearchCriteria.PackageSearchType.values()) {
            //Get Docker and Terraform repos only once
            if (type.equals(PackageSearchCriteria.PackageSearchType.dockerV1) ||
                    type.equals(PackageSearchCriteria.PackageSearchType.oci) ||
                    type.equals(PackageSearchCriteria.PackageSearchType.terraformProvider)) {
                continue;
            }
            List<LocalRepoDescriptor> repoList =
                    type.isRemoteCachesProps() ? localCacheAndFederated : localAndFederatedOnly;
            repoList.stream()
                    .filter(descriptor -> descriptor.getType().equals(type.getRepoType()) ||
                            (descriptor.getType().isMavenGroup() && type.getRepoType().isMavenGroup())
                    )
                    .filter(descriptor ->
                            authorizationService.canRead(InternalRepoPathFactory.repoRootPath(descriptor.getKey())))
                    .forEach(repoDescriptor -> {
                        String localOrRemote =
                                repoDescriptor.isCache() ? REMOTE.typeNameLowercase() : LOCAL.typeNameLowercase();
                        if (repoDescriptor.getType().equals(RepoType.Docker)) {
                            repoKeys.add(new RepoKeyType(localOrRemote, repoDescriptor.getType(),
                                    repoDescriptor.getKey(), repoDescriptor.getDockerApiVersion()));
                        } else {
                            repoKeys.add(new RepoKeyType(localOrRemote, repoDescriptor.getType(),
                                    repoDescriptor.getKey()));
                        }
                    });
        }
        return repoKeys;
    }

    /**
     * Comparator that compares local and cached repositories according to their type (local or cached local) and then
     * internally sorting them by their key.
     */
    private static class LocalAndCachedDescriptorsComparator implements Comparator<RepoDescriptor> {
        @Override
        public int compare(RepoDescriptor o1, RepoDescriptor o2) {
            if (o1 instanceof LocalCacheRepoDescriptor && !(o2 instanceof LocalCacheRepoDescriptor)) {
                return 1;
            } else if (!(o1 instanceof LocalCacheRepoDescriptor) && (o2 instanceof LocalCacheRepoDescriptor)) {
                return -1;
            } else {
                return o1.getKey().toLowerCase().compareTo(o2.getKey().toLowerCase());
            }
        }
    }

    /**
     * return remote and local repository data
     *
     * @return list of repositories repo keys
     */
    private List<RepoKeyType> getAllRepositoriesData(@NonNull String projectKey) {
        List<RepoKeyType> repos = getLocalRepoKeyTypes();
        List<RepoKeyType> remoteRepos = getRemoteRepoKeyTypes();
        List<RepoKeyType> releaseBundleRepos = getReleaseBundlesRepoKeyTypes();
        repos.addAll(remoteRepos);
        repos.addAll(releaseBundleRepos);

        Function<String, Boolean> projectReposFilter = createProjectReposFilter(projectKey);
        return repos.stream()
                .filter(r -> projectReposFilter.apply(r.getRepoKey()))
                .filter(r -> authorizationService.canRead(RepoPathFactory.create(r.getRepoKey())))
                .collect(Collectors.toList());
    }

    private List<RepoKeyType> getPermissionRepoData(@NonNull String projectKey) {
        List<RepoKeyType> repos = new ArrayList<>();
        List<RealRepoDescriptor> localAndRemoteRepoDescriptors =
                repoService.getLocalFederatedAndRemoteRepoDescriptorsByProject(
                        projectKey);
        localAndRemoteRepoDescriptors.stream()
                .filter(repo -> !repoService.isInfoRepo(repo.getType()) &&
                        !ReleaseBundles.equals(repo.getType()) &&
                        !SystemRepos.isInTransitRepo(repo.getKey()) &&
                        !isObservabilityRepo(repo.getKey()))
                .forEach(repoDesc -> {
                    if (repoDesc instanceof RemoteRepoDescriptor) {
                        repos.add(new RepoKeyType(REMOTE.typeNameLowercase(), repoDesc.getKey()));
                    } else {
                        repos.add(new RepoKeyType(LOCAL.typeNameLowercase(), repoDesc.getKey()));
                    }
                });
        return repos.stream()
                .filter(r -> authorizationService.canRead(RepoPathFactory.create(r.getRepoKey())))
                .collect(Collectors.toList());
    }

    private List<RepoKeyType> getLocalRepoKeyTypes() {
        return repoService.getLocalRepoConfigsIncludingInfoRepos().stream()
                .map(repoConfig -> new RepoKeyType(repoConfig.getPackageType(),
                        repoConfig.getKey())).collect(Collectors.toList());
    }

    private List<RepoKeyType> getRemoteRepoKeyTypes() {
        return repoService.getRemoteRepoConfigs().stream()
                .map(remoteRepoConfig -> new RepoKeyType(REMOTE.typeNameLowercase(),
                        remoteRepoConfig.getKey())).collect(
                        Collectors.toList());
    }

    private List<RepoKeyType> getReleaseBundlesRepoKeyTypes() {
        return repoService.getReleaseBundleRepoConfigs().stream()
                .map(remoteRepoConfig -> new RepoKeyType("releaseBundles",
                        remoteRepoConfig.getKey())).collect(
                        Collectors.toList());
    }

    /**
     * If providing a null projectKey name or {@link ProjectsService#DEFAULT_PROJECT} then an empty filter will return
     */
    Function<String, Boolean> createProjectReposFilter(@Nullable String projectKey) {
        if (ProjectsUtils.isDefaultProject(projectKey)) {
            return repoKey -> true;
        }
        Set<String> projectRepoKeys = projectsService.getProjectRepositoriesKeys(projectKey, false);
        return projectRepoKeys::contains;
    }

    @Autowired
    public void setAuthorizationService(AuthorizationService authorizationService) {
        this.authorizationService = authorizationService;
    }

    @Autowired
    public void setRepoService(RepositoryService repoService) {
        this.repoService = repoService;
    }

    @Autowired
    public void setCentralConfigService(CentralConfigService centralConfigService) {
        this.centralConfigService = centralConfigService;
    }

    @Autowired
    public void setProjectsService(ProjectsService projectsService) {
        this.projectsService = projectsService;
    }
}
