package org.artifactory.ui.rest.service.artifacts.browse.treebrowser.tabs.permission;

import org.artifactory.addon.AddonsManager;
import org.artifactory.aol.AolService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.build.InternalBuildService;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.security.ArtifactoryGroupCachingRepoImpl;
import org.artifactory.security.SecurityServiceImpl;
import org.artifactory.security.access.AccessServiceImpl;
import org.artifactory.spring.ArtifactoryApplicationContext;
import org.artifactory.storage.db.security.service.access.AccessUserGroupStoreService;
import org.artifactory.storage.db.security.service.access.GroupMapper;
import org.artifactory.storage.fs.service.ConfigsService;
import org.artifactory.storage.security.service.AclStoreService;
import org.artifactory.test.ArtifactoryHomeBoundJUnitTest;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.tabs.permission.RepoPermissionsModel;
import org.artifactory.ui.rest.service.builds.buildsinfo.tabs.permissions.GetBuildEffectivePermissionService;
import org.artifactory.ui.rest.service.distribution.GetReleaseBundleEffectivePermissionService;
import org.jfrog.access.client.AccessClient;
import org.jfrog.access.client.group.GroupsClient;
import org.jfrog.access.client.user.UsersClient;
import org.jfrog.access.model.Realm;
import org.jfrog.access.rest.group.Group;
import org.jfrog.access.rest.group.Groups;
import org.jfrog.access.rest.user.Users;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.context.ApplicationEventPublisher;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import static org.artifactory.repo.system.SystemRepos.LOG_SHIPPING_REPO_KEY;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;

/**
 * @author Yevdo Abramov
 * Created on 27/10/2020
 */
class GetRepoEffectivePermissionServiceTest extends ArtifactoryHomeBoundJUnitTest {
    @Mock
    private AuthorizationService authService;
    @Mock
    private RepositoryService repoService;
    @Mock
    private GetBuildEffectivePermissionService buildEffectivePermissionService;
    @Mock
    private GetReleaseBundleEffectivePermissionService releaseBundleEffectivePermissionService;
    @Mock
    private UserGroupService userGroupService;

    private GetRepoEffectivePermissionService service;

    @BeforeEach
    public void setUp() {
        MockitoAnnotations.openMocks(this);
        service = new GetRepoEffectivePermissionService(authService, repoService, userGroupService,
                buildEffectivePermissionService, releaseBundleEffectivePermissionService);
    }

    @Test
    void testExecuteWhenNoPermissions() {
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        RestResponse response = mock(RestResponse.class);

        when(request.getQueryParamByKey("repoKey")).thenReturn("my_repo");
        when(request.getQueryParamByKey("path")).thenReturn("my_path");
        when(authService.canManage(any(RepoPath.class))).thenReturn(false);

        service.execute(request, response);

        verify(request).getQueryParamByKey("repoKey");
        verify(request).getQueryParamByKey("path");
        verify(authService).canManage(any());
        verifyNoMoreInteractions(userGroupService);
    }

    @Test
    void testExecuteIsAcceptingManagePermissions() {
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        RestResponse response = mock(RestResponse.class);

        when(request.getQueryParamByKey("repoKey")).thenReturn("my_repo");
        when(request.getQueryParamByKey("path")).thenReturn("my_path");
        when(authService.canManage(any(RepoPath.class))).thenReturn(true);

        service.execute(request, response);

        verify(request).getQueryParamByKey("repoKey");
        verify(request).getQueryParamByKey("path");
        verify(authService).canManage(any());
        verify(userGroupService).getRepoPathAcls(any());
    }

    @Test
    void testLogShippingRepo() {
        ArtifactoryRestRequest<?> request = mock(ArtifactoryRestRequest.class);
        RestResponse<?> response = mock(RestResponse.class);
        doAnswer(invocationOnMock -> {
            assertEquals(invocationOnMock.getArguments().length, 1);
            RepoPermissionsModel repoPermissionsModel = invocationOnMock.getArgument(0, RepoPermissionsModel.class);
            assertNotNull(repoPermissionsModel);
            assertEquals(repoPermissionsModel.getPermissionTargets().size(), 0);
            assertEquals(repoPermissionsModel.getGroupEffectivePermissions().size(), 1);
            assertEquals(repoPermissionsModel.getUserEffectivePermissions().size(), 1);
            repoPermissionsModel.getUserEffectivePermissions().forEach(curr -> {
                assertTrue(curr.getPermission().isRead());
                assertFalse(curr.getPermission().isAnnotate());
                assertTrue(curr.getPermission().isDelete());
                assertFalse(curr.getPermission().isDeploy());
                assertFalse(curr.getPermission().isDistribute());
            });
            repoPermissionsModel.getGroupEffectivePermissions().forEach(curr -> {
                assertTrue(curr.getPermission().isRead());
                assertFalse(curr.getPermission().isAnnotate());
                assertTrue(curr.getPermission().isDelete());
                assertFalse(curr.getPermission().isDeploy());
                assertFalse(curr.getPermission().isDistribute());
            });
            return null;
        }).when(response).iModel(any());

        when(request.getQueryParamByKey("repoKey")).thenReturn(LOG_SHIPPING_REPO_KEY);
        when(request.getQueryParamByKey("path")).thenReturn("my_path");

        // Repo type of pipe info
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setKey(LOG_SHIPPING_REPO_KEY);
        localRepoDescriptor.setType(RepoType.Generic);
        when(repoService.localOrFederatedRepoDescriptorByKey(LOG_SHIPPING_REPO_KEY)).thenReturn(localRepoDescriptor);

        // Security mocking
        when(authService.canManage(any(RepoPath.class))).thenReturn(true);
        HashMap<String, Boolean> adminUserMap = new HashMap<>();
        adminUserMap.put("test", true);
        when(userGroupService.getAllUsersAndAdminStatus(true)).thenReturn(adminUserMap);
        ArrayList<String> adminGroups = new ArrayList<>();
        adminGroups.add("groupAdmin");
        when(userGroupService.getAllAdminGroupsNames()).thenReturn(adminGroups);

        // Context creation
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        AolService aolService = mock(AolService.class);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
        ArtifactoryContextThreadBinder.bind(context);

        service.execute(request, response);

        verify(request).getQueryParamByKey("repoKey");
        verify(request).getQueryParamByKey("path");
        verify(authService).canManage(any());
    }

    @Test
    void testExecutePipelineInfoRepo() {
        ArtifactoryRestRequest<?> request = mock(ArtifactoryRestRequest.class);
        RestResponse<?> response = mock(RestResponse.class);
        doAnswer(invocationOnMock -> {
            assertEquals(invocationOnMock.getArguments().length, 1);
            RepoPermissionsModel repoPermissionsModel = invocationOnMock.getArgument(0, RepoPermissionsModel.class);
            assertNotNull(repoPermissionsModel);
            assertEquals(repoPermissionsModel.getPermissionTargets().size(), 0);
            assertEquals(repoPermissionsModel.getGroupEffectivePermissions().size(), 1);
            assertEquals(repoPermissionsModel.getUserEffectivePermissions().size(), 1);
            return null;
        }).when(response).iModel(any());

        when(request.getQueryParamByKey("repoKey")).thenReturn("my_repo");
        when(request.getQueryParamByKey("path")).thenReturn("my_path");
        // Repo type of pipe info
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setKey("my_repo");
        localRepoDescriptor.setType(RepoType.PipeInfo);
        when(repoService.localOrFederatedRepoDescriptorByKey("my_repo")).thenReturn(localRepoDescriptor);

        // Security mocking
        when(authService.canManage(any(RepoPath.class))).thenReturn(true);
        HashMap<String, Boolean> adminUserMap = new HashMap<>();
        adminUserMap.put("test", true);
        when(userGroupService.getAllUsersAndAdminStatus(true)).thenReturn(adminUserMap);
        ArrayList<String> adminGroups = new ArrayList<>();
        adminGroups.add("groupAdmin");
        when(userGroupService.getAllAdminGroupsNames()).thenReturn(adminGroups);

        // Context creation
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        AolService aolService = mock(AolService.class);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
        ArtifactoryContextThreadBinder.bind(context);

        service.execute(request, response);

        verify(request).getQueryParamByKey("repoKey");
        verify(request).getQueryParamByKey("path");
        verify(authService).canManage(any());
    }

    @AfterEach
    public void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    void testBuildEffectivePermissionsExcludingInternals() {
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        RestResponse response = mock(RestResponse.class);
        String buildRepoKey = "proj-build-info";
        RepoPath repoPath = RepoPathFactory.create(buildRepoKey, "my_path");
        when(request.getQueryParamByKey("repoKey")).thenReturn(buildRepoKey);
        when(request.getQueryParamByKey("path")).thenReturn(repoPath.getPath());
        when(authService.canManage(any(RepoPath.class))).thenReturn(true);
        LocalRepoDescriptor buildRepo = new LocalRepoDescriptor();
        buildRepo.setKey(buildRepoKey);
        buildRepo.setType(RepoType.BuildInfo);
        when(repoService.localOrFederatedRepoDescriptorByKey(buildRepoKey)).thenReturn(buildRepo);
        service.execute(request, response);

        verify(request).getQueryParamByKey("repoKey");
        verify(request).getQueryParamByKey("path");
        verify(authService).canManage(any());
        verify(buildEffectivePermissionService).getPermissionModel(repoPath);
    }

    @Test
    void effectivePermissionExecutesGetGroupAndUsersEffectively() {
        // given
        ArtifactoryApplicationContext context = mock(ArtifactoryApplicationContext.class);
        when(context.beanForType(AolService.class)).thenReturn(mock(AolService.class));
        ArtifactoryContextThreadBinder.bind(context);

        AccessServiceImpl accessService = new AccessServiceImpl(
                mock(ApplicationEventPublisher.class),
                mock(ConfigsService.class),
                context);

        AccessClient accessClient = mock(AccessClient.class);
        accessService.updateAccessClient(accessClient);

        GroupsClient groupsClient = mock(GroupsClient.class);
        when(accessClient.groups()).thenReturn(groupsClient);

        Realm realm = mock(Realm.class);
        when(realm.getName()).thenReturn("realName");

        String groupName = "groupName";
        Group group = mock(Group.class);
        when(group.getRealm()).thenReturn(realm);
        when(group.getName()).thenReturn(groupName);
        when(group.getCustomData(GroupMapper.ArtifactoryBuiltInGroupProperty.artifactory_admin.name()))
                .thenReturn(Boolean.TRUE.toString());

        Groups groups = mock(Groups.class);
        when(groups.getGroups()).thenReturn(List.of(group));
        when(groupsClient.findGroups()).thenReturn(groups);

        UsersClient usersClient = mock(UsersClient.class);
        when(accessClient.users()).thenReturn(usersClient);

        Users users = mock(Users.class);
        when(usersClient.findUsers()).thenReturn(users);

        AccessUserGroupStoreService userGroupStoreService = new AccessUserGroupStoreService();
        userGroupStoreService.setAccessService(accessService);

        InternalRepositoryService repositoryService = mock(InternalRepositoryService.class);
        when(repositoryService.localFederatedOrCachedRepoDescriptorByKey(any())).thenReturn(
                mock(LocalRepoDescriptor.class));

        SecurityServiceImpl userGroupService = spy(new SecurityServiceImpl());
        userGroupService.setUserGroupStoreService(userGroupStoreService);
        userGroupService.setRepositoryService(repositoryService);
        userGroupService.setBuildService(mock(InternalBuildService.class));
        userGroupService.setAclStoreService(mock(AclStoreService.class));

        ArtifactoryGroupCachingRepoImpl artifactoryGroupCache =
                new ArtifactoryGroupCachingRepoImpl(userGroupStoreService, mock(AddonsManager.class));

        userGroupService.setGroupCachingRepo(artifactoryGroupCache);

        ArtifactoryRestRequest artifactoryRestRequest = mock(ArtifactoryRestRequest.class);
        when(artifactoryRestRequest.getQueryParamByKey("repoKey")).thenReturn("repoKey");

        AuthorizationService authorizationService = mock(AuthorizationService.class);
        when(authorizationService.canManage(any())).thenReturn(true);

        GetRepoEffectivePermissionService getRepoEffectivePermissionService =
                new GetRepoEffectivePermissionService(authorizationService,
                        repositoryService, userGroupService, mock(GetBuildEffectivePermissionService.class),
                        mock(GetReleaseBundleEffectivePermissionService.class));

        // when
        getRepoEffectivePermissionService.execute(artifactoryRestRequest, mock(RestResponse.class));

        // then
        verify(groupsClient, times(1)).findGroups();
        verify(usersClient, times(1)).findUsers();

        // second (and other) call gets only users since there are no users cache
        // when
        getRepoEffectivePermissionService.execute(artifactoryRestRequest, mock(RestResponse.class));

        // then
        verify(groupsClient, times(1)).findGroups();
        verify(usersClient, times(2)).findUsers();

        // third call
        // when
        getRepoEffectivePermissionService.execute(artifactoryRestRequest, mock(RestResponse.class));

        // then
        verify(groupsClient, times(1)).findGroups();
        verify(usersClient, times(3)).findUsers();
    }
}